#include <getopt.h>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <memory>
#include <sstream>
using namespace std;

#include "Histogram.h"
#include "AddressTranslator.h"

#include "prsd_traverse.h"
#include "rsd_queue.h"
#include "rank_recursive.h"
#include "rsd_reader.h"
#include "utils.h"
#include "umpi_internal.h"
#include "VParam.h"

static const char *program_name;               /// Name of this program (from args)
static bool expand_ranks = false;              /// Whether to expand recusive rank lists
static bool strip_wrapping = true;             /// Strip mpi wrappers, or not
static bool show_parameters = false;           /// Show op parameters along w/op
static bool only_src_dest = false;             /// Only show src/dest parameters
static bool expand_trace = false;              /// View trace as prsd or expanded
static bool indent_trace = false;              /// View trace with indentation
static bool timing_data = false;               /// Show timing data
static bool only_op = false;                   /// Print only MPI op

static auto_ptr<AddressTranslator> a2l;        /// Address translator (if exe provided)


static struct option long_options[] = {
  {"full-paths",       no_argument,       NULL, 'f'},
  {"no-demangle",      no_argument,       NULL, 'c'},
  {"exe",              required_argument, NULL, 'e'},
  {"hide-functions",   no_argument,       NULL, 'H'},
  {"help",             no_argument,       NULL, 'h'},
  {"inlines",          no_argument,       NULL, 'i'},
  {"expand-ranklists", no_argument,       NULL, 'x'},
  {"expand-trace",     no_argument,       NULL, 'X'},
  {"target",           required_argument, NULL, 't'},
  {"rank",             required_argument, NULL, 'r'},
  {"no-mpi-strip",     no_argument,       NULL, 'm'},
  {"show-parameters",  no_argument,       NULL, 'p'},
  {"only-src-dest",    no_argument,       NULL, 'P'},
  {"indent",           no_argument,       NULL, 'I'},
  {"walk",             no_argument,       NULL, 'w'},
  {"timing-data",      no_argument,       NULL, 'd'},
  {"site",             no_argument,       NULL, 's'},
  {"op",               no_argument,       NULL, 'o'},
  {0,                  no_argument,       0,     0 }
};

/* Print a usage message to STREAM and exit with STATUS.  */
static void usage (FILE *stream, int status) {
  fprintf (stream, "Usage: %s [option(s)] tracefile(s)\n", program_name);
  fprintf (stream, " Command-line viewer for compressed MPI Traces.\n");
  fprintf (stream, " Formatting options:\n\
  -e           --exe=<executable>    Executable for reading callpaths (prints raw addresses if not supplied)\n\
  -X <rank>    --expand-rank=<rank>  Show fully expanded trace for a particular rank (can result in LONG output).\n\
  -I           --indent              Indent PRSD nesting levels in output expanded with -X.\n\
  -r <rank>    --rank=<rank>         Show only trace for a particular rank, without expanding.\n\
  -x           --expand-ranklists    Expand rank lists so they're human-readable.\n\
  -p           --show-parameters     Show human-readable parameter values for trace elements.\n\
  -P           --only-src-dest       Only show src and destination parameters (implies -p).\n\
  -d           --timing-data         Show timing data (histograms)\n\
  -w           --walk                Stack walk is done \"outwards\"\n\
\n\
 Address translation options (for callpaths):\n\
  -t <format>  --target=<format>     Specify target binary format name\n\
  -m           --no-strip            Do not strip mpi wrapper functions rom callpaths\n\
  -s           --site                Show only callsite + MPI call (not full path)\n\
  -o           --op                  Show only MPI call (no addr translation)\n\
  -i           --inlines             Unwind inlined functions (only with recent binutils))\n\
  -f           --full-paths          Use full directory paths (instead of only base names)\n\
  -H           --hide-functions      Hide function names (show only file/line)\n\
  -c           --no-demangle         Don't demangle function names (default demangles)\n\
\n\
 Other options:\n\
  -h           --help                Print this message\n\
\n");
  exit (status);
}

/// Used to store indentation for below routines.
static ostringstream indent;
static inline void print_indent(ostream& out) {
  if (indent_trace) out << indent.str();
}


static void field_printer(replay_op *op, op_field_desc_t *field) {
  char buf[4096];
  char *pos = buf;
  pos += sprintf(pos, "%-16s", field->name);
  pos += soutput_field_value(pos, op, field);

  print_indent(cout);
  cout << buf << endl;
}


void printTranslated(stack_sig *signature, ostream& out) {
  if (a2l.get()) {
    vector<string> path;
    a2l->translate(sig_len(signature->sig), signature->sig, path);
    
    if (path.size()) out << path[0];
    for (size_t i=1; i < path.size(); i++) {
      out << " > " << path[i];
    }
      
  } else {
    int len = sig_len(signature->sig);
    out << hex;
    if (len) out << signature->sig[0];
    for (int i=0; i < len; i++) {
      out << " > " << signature->sig[i];
    }
    out << dec;
  }
}


/**
 * Visitor routine for executing replay operations.  This is called
 * by the traversal routines to translate ops to actual MPI function 
 * calls.
 */
static int replay_visitor(rsd_node *node, int rank, int depth, int iter, int last) {
  if (indent_trace) {
    indent.str("");
    for (int i=0; i < depth; i++) {
      indent << "    ";
    }
  }

  print_indent(cout);
  if (expand_ranks) {
    foutput_expanded_list(stdout, node->rlist);
  } else {
    foutput_recursive_list(stdout, node->rlist);
  }

  print_indent(cout);
  if (!only_op) {
    printTranslated(&node->signature, cout);
    cout << " > ";
  }

  if (strip_wrapping) {
    // if the wrappers are stripped, elide them to the mpi opcode.
    cout << get_struct_by_opcode(node->data.op)->name 
         << endl;
  }

  if (!expand_trace && node->prsd_info.head) {
    prsd_node *ptr = node->prsd_info.head;
    while(ptr) {
      print_indent(cout);
      cout << "PRSD: " << ptr->loop_iters << " x " << ptr->loop_mems << " elts";
      
      if (!only_op) {
        cout << " to [ ";
        printTranslated(&ptr->tail_sig, cout);
        cout << " ]";
      }

      cout << endl;
      ptr = ptr->next; 
    }
  }

  if (show_parameters) {
    // print parameters designated in the fields.
    if (only_src_dest) {
      node->data.fields &= (SOURCE | DEST);
      node->data.arrays = 0;
    }
    map_over_fields(&node->data, field_printer);
    
    // output rlparams if there are any in the trace.
    if (node->data.pvectors != 0) {
      cout << "Untranslated rank list params:" << endl;
      
      char buf[4096];
      if (node->data.pvectors & RECVCOUNTS) {
        rlparam_output (node->data.rlparam, RECVCOUNTS, buf);
        cout << buf;
      }
      if (node->data.pvectors & DISPLS) {
        rlparam_output (node->data.rlparam, DISPLS, buf);
        cout << buf;
      }
      if (node->data.pvectors & COUNTS) {
        rlparam_output (node->data.rlparam, COUNTS, buf);
        cout << buf;
      }
      if (node->data.pvectors & RDISPLS) { 
        rlparam_output (node->data.rlparam, RDISPLS, buf);
        cout << buf;
      }
    }

    // output arrays if there are any.
    if (node->data.arrays) {
      field_mask_t arr;
      char buf[4096];

      if ((arr = array_type1(node->data.arrays))) {
        soutput_array(buf, arr, node->data.array_size1, node->data.array1);
        cout << buf;
      }
      
      if ((arr = array_type2(node->data.arrays))) {
        soutput_array(buf, arr, node->data.array_size2, node->data.array2);
        cout << buf;
      }
      
      if ((arr = array_type3(node->data.arrays))) {
        soutput_array(buf, arr, node->data.array_size3, node->data.array3);
        cout << buf;
      }
      
      if ((arr = array_type4(node->data.arrays))) {
        soutput_array(buf, arr, node->data.array_size4, node->data.array4);
        cout << buf;
      }
    }
  }

#ifdef FEATURE_TIMING_HISTO
  // output histograms
  if (timing_data && node->data.histograms[0].histo) {
    char buf[8192];
    cout << "TIME0: " << endl;
    static_cast<Histogram*>(node->data.histograms[0].histo)->output(buf);
    cout << buf;
  }
  if (timing_data && node->data.histograms[1].histo) {
    char buf[8192];
    cout << "TIME0: " << endl;
    static_cast<Histogram*>(node->data.histograms[1].histo)->output(buf);
    cout << buf;
  }
#endif /* FEATURE_TIMING_HISTO */

  cout << endl;

  return 0;
}


int main(int argc, char *argv[]) {
  program_name = argv[0];

  char *target_format = NULL;
  char *exe_name = NULL;
  int process_rank = -1;
  bool base_names = true;
  bool do_demangle = true;
  bool with_functions = true;
  bool unwind_inlines = false;
  bool only_callsite = false;
  bool walk_outward = false;

  /* Parse options from command line. */
  int longIndex;
  int c;
  while ((c = getopt_long (argc, argv, "Lldhe:r:t:fcHixX:pPtIwso", long_options, &longIndex)) != EOF)
  {
    switch (c)
    {
    case 'h':
      usage(stderr, 1);
      break;
    case 'X': 
      expand_trace    = true;  // fallthru
    case 'r':
      char *endptr;
      process_rank = (int)strtol(optarg, &endptr, 10);
      if (*endptr) usage(stderr, 1);
      break;
    case 'e': exe_name        = optarg;  break;
    case 't': target_format   = optarg;  break;
    case 'f': base_names      = false;   break;
    case 'c': do_demangle     = false;   break;
    case 'H': with_functions  = false;   break;
    case 'i': unwind_inlines  = true;    break;
    case 'x': expand_ranks    = true;    break;
    case 'I': indent_trace    = true;    break;
    case 's': only_callsite   = true;    break;
    case 'o': only_op         = true;    break;
    case 'd': timing_data     = true;    break;
    case 'p': show_parameters = true;    break;
    case 'P': show_parameters = only_src_dest = true; break;
    case 'w': walk_outward   = true;    break;
    case 0:   /* We've been given a long option that has no short arg.  */
      if (strcmp("no-strip", long_options[longIndex].name ) == 0) {
        strip_wrapping = false;
      }
      break;
    default:
      usage (stderr, 1);
      break;
    }
  }

  argv += optind;
  argc -= optind;
  
  if (argc < 1) usage(stderr, 1);

  if (exe_name) {
    a2l.reset(new AddressTranslator(exe_name, target_format));
    if (strip_wrapping) a2l->setStripPrefix("MPI_");      // strip mpi routine from the callpath
    a2l->setBaseNames(base_names);
    a2l->setDemangle(do_demangle);
    a2l->setWithFunctions(with_functions);
    a2l->setUnwindInlines(unwind_inlines);
    a2l->setOnlyCallsite(only_callsite);
    a2l->setWalkOutward(walk_outward);
  }

  char **tracefile = argv;
  while (argc--) {
    rsd_queue *queue = read_rsd_file_rank(*tracefile++, process_rank);

    if (expand_trace) {
      replay_queue(queue->head, 0, replay_visitor);
    } else {
      traverse_queue(queue->head, replay_visitor);
    }

    free_rsd_nodes(queue);    
  }

  exit(0);
}
#if 0
#include "VParam.h"
void vparam_init (VParamList *v) {
}
void vparam_free (VParamList *v) {
}
int vparam_output (VParamList vparam, int64_t param, char *buf){
  return 0;
}

int64_t vparam_get_value (char * buf, int64_t param, int rank){
  return 0;
}
#endif
