/* addr2line.c -- convert addresses to line number and function name
   Copyright 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2006
   Free Software Foundation, Inc.

   Author: Todd Gamblin
   This file is a C++ port of the Addr2Line utility, part of GNU Binutils.
   For more info on GNU Binutils (which this requires to compile), see
   http://www.gnu.org/software/binutils/.

   This is intended to allow C++ programs an easy interface to address
   translation.  This file needs to be compiled with libbfd and libiberty,
   and it requires that the libiberty headers also be accessible.  Some
   versions of binutils do not install these by default.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.  */
#ifndef ADDRESS_TRANSLATOR_H
#define ADDRESS_TRANSLATOR_H

#include <string>
#include <memory>
#include <vector>
#include <ostream>

extern "C" {
#include "bfd.h"
}

class AddressTranslator {
private:
  // These variables are options for parsing output.
  bool unwind_inlines;	 /// unwind inlined functions (only with later binutils versions!) 
  bool with_functions;	 /// show function names.  
  bool do_demangle;	 /// demangle names.  
  bool base_names;	 /// strip directory names.  
  bool only_site;        /// shows only last callsite
  bool walk_outward;    /// which way is the stack sig

  /// If not null string, this will stop expanding when it a routine with this prefix is found. 
  std::string strip_prefix;
  
  // These variables are used to pass information between
  // translate_addresses and find_address_in_section.  
  bfd_vma pc;
  const char *filename;
  const char *functionname;
  unsigned int line;
  bool found;

  // data read from file.
  asymbol **syms;		/* Symbol table.  */  
  bfd *abfd;
  
  std::auto_ptr<char> target;  // target format
  
  /// converts string to lowercase
  static void lowercase(std::string& s);

  // utility functions used by translate.
  void slurp_symtab (bfd *);
  void find_address_in_section (bfd *, asection *);
  static off_t get_file_size (const char * file_name);

  /// Inits bfd library once and only once.
  /// Note: NOT thread safe, so disregard above if you use threads.
  static void initBfd();

public:
  /**
   * Construct a new address translator to translate addresses for 
   * the executable file specified by filename.
   */
  AddressTranslator(const char *filename, const char *target = NULL);

  /**
   * Destroys this class, its symbol table, etc.
   */
  ~AddressTranslator();

  /**
   * Translates addressses as character hex strings to strings representing lines in
   * the executable.
   */
  void translate(int count, const char **addrs, std::vector<std::string>& lines);

  /**
   * Translates raw addresses to strings representing lines in the executable.
   */
  void translate(int count, const intptr_t *addrs, std::vector<std::string>& lines);
  
  
   /// Utility function wraps libiberty's cplus_demangle.
  static char * demangle(bfd *abfd, const char *name);


  // These adjust options for generation of output
  void setUnwindInlines(bool value) { unwind_inlines = value; }
  void setWithFunctions(bool value) { with_functions = value; }
  void setDemangle(bool value) { do_demangle = value; }
  void setBaseNames(bool value) { base_names = value; }
  void setOnlyCallsite(bool site) { only_site = site; }
  void setStripPrefix(const std::string& value);
  void setWalkOutward(bool value) { walk_outward = value; }

};

#endif /* ADDRESS_TRANSLATOR_H */
