#ifndef BIT_VECTOR_H
#define BIT_VECTOR_H

#include <cmath>
#include <cassert>
#include <iostream>
#include <stdint.h>

#include "RankSet.h"

/**
 * This is a bit vector whose size is fixed upon construction.  
 * It replaces vector<bool> and bitset, as the former behaves 
 * inconsistently across C++ implementations and the latter must
 * have a fixed size at compile time.
 */
class BitVector : public RankSet {
public:
  typedef uint64_t elt_t;         /** type of elements in the bitvector. */
  static const int LOG2_INT = 6;  /** log2 number of bits in a uint64_t */
  static const int INT_REM = 63;  /** bit mask for last 6 bits */

private:

  const int numElts;   /** Number of unsigned ints in this BitVector. */
  const int realSize;  /** Actual size of bit vector, in bits. */
  elt_t *const bits;        /** The actual bits in the BitVector. */

  /** Disallow copying. */
  BitVector(const BitVector& other);

  /** Disallow assignment. */
  BitVector& operator=(const BitVector& other); 

public:

  /**
   * Constructs a new BitVector with at least the 
   * specified number of bits, all initialized to 0.
   */
  BitVector(int size, bool initialValue = false);

        
  /** Deallocates space for bits. */
  virtual ~BitVector();

  /**
   * @param i the index of the bit to test.
   * @return whether the bit at index i is set.
   */
  virtual bool member(int i) const {
    return bits[i >> LOG2_INT] & (((unsigned) 1) << (i & INT_REM));
  }

        
  /** 
   * Returns true if the bit at index i is set, false if not.
   * @warning this does NOT return a reference to the ith bit,
   * it returns a copy.  You cannot use this as you would the 
   * vector subscript operator. 
   * <p>
   * @param i the index of the bit to fetch.
   */
  bool operator[](int i) const {
    return member(i);
  }


  /**
   * Sets the bit at index i high.
   * @param i the index of the bit to set.
   */
  void set(int i) {
    bits[i >> LOG2_INT] |= (((unsigned) 1) << (i & INT_REM)); 
  }

  /** Synonym for set() for RankList interface. */
  virtual void add(int rank) {
    set(rank);
  }


  /**
   * Sets the bit at index i low.
   * @param i the index of the bit to set.
   */
  void unset(int i) {
    bits[i >> LOG2_INT] &= ~(((unsigned) 1) << (i & INT_REM));
  }


  /**
   * Sets the bit at index i to the value of the supplied status.
   * <p>
   * @param i the index of the bit to set.
   * @param status the value to set the bit to.
   */
  void set(int i, bool status) {
    if (status) {
      set(i);
    } else {
      unset(i);
    }
  }


  unsigned getNumElts() const {
    return numElts;
  }

  
  elt_t* getData() const {
    return bits;
  }


  /**
   * @return the capacity (in bits) of this vector.
   */
  virtual int capacity() const {
    return realSize;
  }


  /**
   * Sets all bits in the vector to false.
   */
  void clear();


  /**
   * Sets all bits in the vector to false.
   */
  void setAll();

  /**
   * Gives the number of bits set in the bitvector.
   */
  virtual int size() const;


  /**
   * Ands all values in this BitVector with those in the other.
   * PRE: vecors are the same size.
   */
  BitVector& operator&=(const BitVector& other);


  virtual void merge(const RankSet& other) {
    (*this) |= dynamic_cast<const BitVector&>(other);
  }

  /**
   * Ors all values in this BitVector with those in the other.
   * PRE: vecors are the same size.
   */
  BitVector& operator|=(const BitVector& other);


  /**
   * Outputs hex version of bit vector to provided output stream.
   */
  virtual int soutput(char *buf) const;

  
  /**
   * Reads a string formatted as per output of soutput and creates a 
   * corresponding BitVector.
   * Optionally gives the number of characters read from the input, if a non-null
   * int pointer is provided in count.
   */
  static BitVector *input(const char *buf, int *count = NULL);


  /** Prints out set nodes in the bit vector. */
  friend std::ostream& operator<<(std::ostream& out, const BitVector& vec);


  // methods for MPI packing of bitvectors.
  virtual int packed_size(MPI_Comm comm) const;
  virtual void pack(void *buf, int bufsize, int *position, MPI_Comm comm) const;
  static BitVector *unpack(void *buf, int bufsize, int *position, MPI_Comm comm);

  /**
   * Creates a duplicate of this BitVector. 
   */
  virtual BitVector *clone() const;
};

#endif //BIT_VECTOR_H

