/*
 * Filename: Event.h
 * Author: Karthik
 * This class holds one MPI Event with parameters.
 * This also holds timing histogram and other user defined
 * histograms for any general purpose like using Hardware counter
 * results.
 */

#ifndef __EVENT_H__
#define __EVENT_H__

#include <inttypes.h>
#include "Histogram.h"
#include "Param.h"
#include "ScalarParam.h"
#include "VectorParam.h"
#include "opstruct_macro.h"

#ifdef __cplusplus
#include <map>

//#define FORMAT_OUT(x) (log2pow2((field_mask_t)x))
#define FORMAT_OUT(x) ((field_mask_t)x)

class Event {

private:
  std::map<int64_t,GenericParam<MPIParam>* > params;
  //std::map<int64_t,GenericScalarParam * > params;
  int event_id;
  int64_t param_list;
  int seq_num;
  //Histogram * timing_histo;
  struct time_data time_info[TIMING_MAX];

  Histogram * user_def_histo_1;
  Histogram * user_def_histo_2;

  int checkCompareException(int64_t val);

public:

  static const int64_t param_special_case = 
       (1ll << TYPE_COUNT) | (1ll << TYPE_ROOT) | (1ll << TYPE_DEST) | (1ll << TYPE_REQUEST) | (1ll << TYPE_COLOR) | (1ll << TYPE_KEY) | (1ll << TYPE_SOURCE);

  Event():event_id(0),param_list(0),user_def_histo_1(0),user_def_histo_2(0){ 
    for(int i = 0; i < TIMING_MAX; i++)
      time_info[i].time_list = 0;
  }

  ~Event();

  Event& operator=(const Event&);

  void init();

  void setParamList(int64_t val);

  void setEventId(int val);

  void setSeqNum (int val);

  int getEventId() {
    return event_id;
  }

  int getSeqNum() {
    return seq_num;
  }

  int64_t getParamList() {
    return param_list;
  }

  std::map<int64_t,GenericParam<MPIParam>* > getParamMap() {
    return params;
  }

  GenericParam<MPIParam>* getParam(int64_t param_id);
  
  void addParam (int64_t param_id, int val);

  void addParam (int64_t param_id, float val);

  void addParam (int64_t param_id, double val);

  void addParam (int64_t param_id, char* val);

  void addVectorParam (int64_t param_id, int *arr_param, int size);

  void printParams (); 

  int compareParams(Event* e);

  int soutput_params(char *buf);

  int soutput_timing(char *buf);

  void setTime(long long int time, int index);

  TimeList ** getTimeList(int index);

  long long int getTime(int index);

  int event_packed_size(MPI_Comm comm);

  void unpack(void *buf, int bufsize, int *position, MPI_Comm comm);

  void pack(void *buf, int bufsize, int *position, MPI_Comm comm);
};

extern "C" {
#endif /*end of __cplusplus*/
  typedef struct Event Event;
//  typedef struct ScalarParam ScalarParam;
//  typedef struct VectorParam VectorParam;

  void init_event(Event *event);
  
  void create_event(Event **event);

  void add_scalar_param_int(Event *event, int64_t param_id, int val);
  
  void add_scalar_param_float(Event *event, int64_t param_id, float val);

  void add_scalar_param_double(Event *event, int64_t param_id, double val);

  void add_scalar_param_string(Event *event, int64_t param_id, char *val);

  void add_vector_param(Event *event, int64_t param_id, int *arr_param, int size);

  void event_set_param_list(Event *event, int op, int seqnum, int64_t value);

  void event_print(Event *event);

  void event_copy(Event **dest, Event **src);

  void free_event(Event **e);

  int compare_event_params(Event *e1, Event *e2);

  int soutput_event_params(char *pos, Event *e);

  int soutput_event_timing(char *buf, Event *e);

  void set_event_time(Event *e, long long int time, int index);

  void event_time_merge(int index, Event *dest, const stack_sig *dest_sig,
    int desk_rank, Event *src, const stack_sig* src_sig, int src_rank, int B);

  int event_merge_vectors(Event *merge, Event *target, int64_t param);

  void event_vector_reset_undo(Event *e, int64_t param);

  void event_vector_undo_addition(Event *e, int64_t param);

  int event_packed_size(Event *e, MPI_Comm comm);

  void event_pack(Event *e, void *buf, int bufsize, int *position, MPI_Comm comm);  

  void event_unpack(Event **e, void *buf, int bufsize, int *position, MPI_Comm comm);  

  void event_merge_param_value_list(Event *master, Event *slave, int64_t param,
          int *m_list, int *s_list);

  
void event_get_param (Event *event, int64_t param_id, void *param);

#ifdef __cplusplus
}
#endif /*__cplusplus*/

#endif /*__EVENT_H__*/
