/*
 * Filename: histo.h
 * Author:   tgamblin, adapted from histo.c by pratn
 * Created:  Fri Aug 31 10:08:27 2007
 * $Id: Histogram.h,v 1.3 2008/12/26 20:44:48 gamlin Exp $
 */
#ifndef __HISTOGRAM_H__
#define __HISTOGRAM_H__


#ifdef __cplusplus
#include <memory>
#include <ostream>
#include <mpi.h>
#include "rank_recursive.h"
#include "RankSet.h"
#include "stack_sig.h"
#include "RandomVector.h"

#define HISTO_MAGIC 0xaabbccdd
#define BIN_MAGIC   0xdeadd00d

/**
 * This is the element type for the histogram's list of bins.
 */
struct histogram_bin {
  int    magic;        /* Magic number for mem debugging. */
  double start;        /* bin start */
  double min;          /* bin start */
  double max;          /* bin end */
  double end;          /* bin end */
  //int    frequency;    /* #items */
  long long frequency;    /* #items */
  double average;      /* Average */
  double variance;     /* Variance --- (n-1)*(sigma^2) to be precise */

  /** 
   * Set of ranks that contributed to the values in this bin.
   * If this is not set then ranks in the bin are considered to be local,
   * and do not represent participants from multiple tasks.
   */
  std::auto_ptr<RankSet> participants;

  struct histogram_bin * next;   /* linked list next ptr */
  
  /** This constructor sets only magic.  Used for inputting histo bins. */
  histogram_bin();

  /** Destructor -- destroys participant list if it's there. */
  ~histogram_bin();

  /** This constructor sets magic, start, and end and sets everything else to zero. */
  histogram_bin(double _start, double _end);

  /** Adds a value to the bin and updates values. */
  void add(double v, long long freq = 1);

  /** 
   * Gives the frequency divided by number of participants.  Use to estimate
   * the frequency of a value for a particular process.
   */
  double singleRankFrequency() {
    double result = frequency;
    if (participants->size()) result /= (double)participants->size();
    return result;
  }

  void addParticipant(int rank);
};


/**
 * Actual class for histograms.  
 */
class Histogram {
public:
  // defaults for various histogram parameters
  static int ranks_in_histograms;           // rank-preserving or non-rank-preserving histos
  static int default_buckets;               // buckets in default histograms
  static int default_fixwidth;              // fix width of floats in histograms.
  static int default_smooth_threshold;
  static int default_merge_limit;
  static int default_merge_ratio;
  
private:
  unsigned int magic;   /* magic number */

  //int num_elems;        /* num of values added to the histogram */
  long long num_elems;        /* num of values added to the histogram */
  int num_bins;         /* num of bins */
  int smooth_threshold; /* 'smoothen' the histogram after every smooth_threshold additions */
  int merge_limit;      /* maximum difference between two bin sizes which can be merged */
  int merge_ratio;      /* merge limit expressed as a % diff */

  /* Below instance variables are for tracking outliers (max & min) */
  int creator_rank;     /* Rank of the node the histogram was created on. */
  int max_rank;         /* recorded rank of node with max value */
  int min_rank;         /* recorded rank of node with min value */
  double max_value;
  double min_value;

  stack_sig signature;  /* signature of the previous op. */
  
  histogram_bin **list;   /* sorted list of bins */
  struct histogram_bin * bins;   /* actual bin data */


  /**
   * Private no-arg constructor, used only for reading in histograms from input.
   */
  Histogram();

  void init ();

  /**
   * 'smoothen' the histogram - finds the largest and smallest bins and then
   * splits the large bin only when two bins can be merged. This ensures that the
   * number of bins is maintained constant.
   */
  void smooth();


  /**
   * Construct a sorted list of bins. The bins are sorted by the frequency field.
   * This is used to determine the largest (split-able) and smallest (merge-able)
   * bins.
   */
  void sort();

  /**
   * Find the lowest frequency bin which can be merged with the next bin. The attr
   * fields define the threshold (absolute as well as percentage) which is used to
   * determine if two bins can be merged.
   */
  histogram_bin *low_bin();
  

  /**
   * Merges a single bin into this histogram.  Helper for merge(Histogram*).
   */
  void merge(histogram_bin *bin);

  void skyline_merge(histogram_bin *bin);

  /**
   * Given a value, find the bin to which it falls into in this histogram.
   */
  histogram_bin *find_bin_for_value(double value);


  /**
   * Add a value v to an existing histogram. This method will take a value and add
   * it into the appropriate histogram while updating the average and standard
   * deviation.
   *
   * This will return a pointer to the bin in which the value was inserted, 
   * UNLESS smoothing was done.  If smoothing was done, this returns NULL,
   * and routines wanting to know where the value was added will need to search.
   *
   * This is a private helper method.  It does not update min/max values.
   */
  histogram_bin *insert(double v);

  
  /**
   * Updates min value according to new value + rank.
   */
  void update_min(double v, int rank) {
    if (v < min_value) {
      min_value = v;
      min_rank = rank;
    }
  }


  /**
   * Updates the max value and max rank.
   */
  void update_max(double v, int rank) {
    if (v > max_value) {
      max_value = v;
      max_rank = rank;
    }
  }

public:
  /**
   *  Creates an empty histogram with numBins bins.
   */
  Histogram(int numBins, 
	    int smooth_threshold = default_smooth_threshold,
	    int merge_limit = default_merge_limit,
	    int merge_ratio = default_merge_ratio,
	    int creator_rank = getCommRank() );
  

  /**
   * Constructor for TimeList.  Todo: extract this interface into some class
   * hierarchy... it's a little messy.
   */
  Histogram(int value, int B, 
	    const stack_sig *sig, int my_rank, int min_rank, int max_rank);


  /**
   * free all allocated memory for histogram
   */
  ~Histogram();
  
  /**
   * Add a value v to an existing histogram. This method will take a value and add
   * it into the appropriate histogram while updating the average and standard
   * deviation.
   *
   * If this is the first add, bin sizes are set so that the histogram's range is
   * 2v with value at the center of the range.
   */
  void add(double value, int rank = -1);

  /**
   * Changes the default rank of this histogram.
   */
  void set_rank(int rank);

  /**
   * explicitly init min/max ranks
   */
  void set_min_max_ranks (int min, int max) ;

  /**
   * Does this histogram have ranks other than creator's rank as member?
   */
  bool has_other (int rank);

  /**
   * print the histogram for debugging purposes
   */
  void print(unsigned char, unsigned char);
  
  /**
   * print out histogram to trace buffer
   * buf is pointer to trace buffer
   * h   is histogram handle
   * Returns number of characters written.
   */
  int output (char *buf, bool fix_width = default_fixwidth) const;
  
  /**
   * Merges other histogram into this one.  
   * Note: this will alter this histogram.
   */
  void merge(const Histogram *other);

  void skyline_merge(const Histogram *other);

  /**
   * Get the (expected) value of the ith element
   * Use unique k to get an interpolated value
   */
  double value (int k, int j);

  double value_from_vector ( RandomVector *rv);

  /**
   * Get the expected value for a particular rank out of this 
   * histogram.
   */
  double value_for_rank(int rank);
  

  /**
   * average value of this histogram
   */
  double average();

  
  /**
   * minimum value in this histogram
   */
  double min();

  
  /**
   * find out the min and max value of this histogram
   */
  int min_max (double * min, double * max, int * min_rank, int * max_rank);

  /**
   * signature match
   */
  int sig_match (const Histogram * other) const {
    return SIG_MATCH(signature, other->signature);
  }

  int sig_match (const stack_sig *sig) const {
    return SIG_MATCH(signature, *sig);
  }


  /**
   * True if any bins contain values contributed by the specified rank.
   */
  bool isParticipant(int rank);


  /** Gets the rank of the histogram's creator. */
  int getCreatorRank() const {
    return creator_rank;
  }

  /**
   * Returns an upper bound on the size of this histogram if it were packed w/MPI_Pack.
   */
  int get_packed_size(MPI_Comm comm);

  void pack(void *buf, int bufsize, int *position, MPI_Comm comm);

  static Histogram *unpack(void *buf, int bufsize, int *position, MPI_Comm comm);

  /**
   * Inputs a recursive rank list with brackets around it.
   * TODO: move to rank_recursive.[hc]?
   */
  static int *input_rlist(char *buf);
  
  /*
   * construct a histogram from string buffer
   * buf is pointer to trace buffer
   * N   is numbe of buckets to be read
   *
   * h   is histogram handle
   */
  static Histogram *input(char *buf);


  RankSet *getParticipants();

  /**
   * Gets rank of current node.  Helper for constructor.
   */
  static int getCommRank();
  
  void sig_print();
};

std::ostream& operator<<(std::ostream& out, const Histogram& histo);


extern "C" {
#endif /* __cplusplus */

#include "RandomVector.h"
  typedef struct Histogram Histogram;
  //TODO: eliminate  all these functions as we move towards compiling w/C++ compiler.

  void histogram_free(Histogram *histo);
  int histogram_output(char *buf, Histogram *histo);
  Histogram* histogram_input(char *buf);
  int histogram_packed_size(Histogram *histo, MPI_Comm comm);
  Histogram * histogram_unpack(void *buf, int bufsize, int *position, MPI_Comm comm);
  void histogram_pack(Histogram *histo, void *buf, int bufsize, int *position, MPI_Comm comm);

  void histogram_merge(Histogram **m, long int mtime, Histogram **s, long int stime, int srank, int B);
  void histogram_init(Histogram **histo, int B);

  void histogram_add (Histogram *h, int v);

  double histogram_value (Histogram *h, int k, int j);
  double histogram_value_from_vector (Histogram *h, RandomVector *rv);
  int histogram_min_max (Histogram *histo, double * min, double * max, int * min_rank, int * max_rank);
  void histogram_average (Histogram *histo, double *avg);
  int histogram_has_other (Histogram *histo, int rank);
  
  
#ifdef __cplusplus
}
#endif /* __cplusplus */

  
#endif /* __HISTOGRAM_H__ */
/* vim: set tw=80 sw=2 expandtab: */
  
