#ifndef RLIST_H
#define RLIST_H

#include <cassert>
#include <string>
#include <mpi.h>
#include "RankSet.h"
#include "rank_recursive.h"

/**
 * Rank set implementation based on recursive lists (see rank_recursive.h)
 */
class RList : public RankSet {
  int *rlist;
public:  
  /** Constructs an empty rlist. */
  RList();

  /** Constructs an rlist with one member: the provided rank. */
  RList(int rank);

  /** Construct an rlist out of a recursive rank list. */
  RList(int *rlist);


  /** destroys an rlist. */
  virtual ~RList() {
    if (rlist) destroy_recursive_list(rlist);
  }

  /** Adds a rank to the set. */
  virtual void add(int rank) {
    if (rlist) {
      merge_single(rank, &rlist);
    } else {
      create_recursive_list(rank, &rlist);
    }
  }

  /** Returns 'rlist' */
  virtual int * get_rlist () {
    return rlist;
  }

  /** Returns the number of bits that are set in this rank set. */
  virtual int size() const {
    return member_count(rlist);
  }

  /** Returns the number of bytes required to hold this */
  virtual int list_length() const {
    return size_recursive_list(rlist);
  }

  /** 
   * Total number of bits that this can hold, or negative for no limit.
   */
  virtual int capacity() const {
    return -1;
  }

  /** 
   * Merges contents of other rank set with this one.
   * This will modify this rank set.
   */
  virtual void merge(const RankSet& olist); 

  /** Use an integer list instead of a RankSet */
  virtual void merge(const int * rl);

  /**
   * Returns true if rank is a member of this set, false otherwise.
   */
  virtual bool member(int rank) const {
    return rlist ? member_rank(rank, rlist) : false;
  }


  /** Get the packed size of the rank list. */
  virtual int packed_size(MPI_Comm comm) const;

  /** Packs the rank list into a buffer for MPI transport. */
  virtual void pack(void *buf, int bufsize, int *position, MPI_Comm comm) const;
  

  /** Unpacks an RList from a buffer. */
  static RList *unpack(void *buf, int bufsize, int *position, MPI_Comm comm);


  /** Output to char buffer. */
  virtual int soutput(char *buf) const;


  /** Input an rlist from a buffer, in the format produced by soutput. */
  static RList *input(const char *buf);

  /** Convenience overload of input(); */
  static RList *input(const std::string& s) {
    return input(s.c_str());
  }

  
  /** Creates a duplicate RList. */
  virtual RList *clone() const;

  /** clones a int * list instead of itself */
  static RList *clone (int *rlist) ;

  /**
   * Calls the provided callable entity for each rank in the rlist.
   * PRE: T can be called with one parameter, an int representing a 
   * member rank in the rlist.
   */
  template<typename T>
  void traverse(T visit) {
    int pos, cur_prsd, depth, start;
    
    if(!rlist) return;
    
    pos = 1;
    cur_prsd = 0;
    
    while(cur_prsd < rlist[0]) {
      depth = rlist[pos];
      start = rlist[pos + 1];
      pos += 2;
      
      traversal_helper(rlist + pos, start, depth, visit);
      if(cur_prsd + 1 == rlist[0]) break;     // find next prsd 

      pos++;
      while(rlist[pos] > 0) {
	pos++;
      }
      cur_prsd++;
    }
  }


  /** 
   * Recursive helper for traverse().  See traverse above.
   */
  template <typename T>
  static void traversal_helper(int *list, int start, int depth, T visit) {
    int i, rank, iters, stride;
    
    if(depth > -1)
      return;
    
    iters = list[0];
    stride = list[1];
  
    for(i = 0, rank = start; i < iters; i++, rank += stride) {
      visit(rank);
      if(depth != -1) {
	traversal_helper(list + 2, rank, depth + 1, visit);
      }
    }
  }
};

#endif //RLIST_H
