#ifndef RANK_SET_H
#define RANK_SET_H

#include <mpi.h>


/**
 * Interface for sets of ranks, 
 */
class RankSet {
public:
  
  /** Adds a rank to the set. */
  virtual void add(int rank) = 0;

  /** 
   * Merges contents of other rank set with this one.
   * This will modify this rank set.
   */
  virtual void merge(const RankSet& other) = 0;

  /** Returns the number of members currently in this rank set. */
  virtual int size() const = 0;

  /** 
   * Total number of bits that this can hold, or negative for no limit.
   */
  virtual int capacity() const = 0;

  /**
   * Returns true if rank is a member of this set, false otherwise.
   */
  virtual bool member(int rank) const = 0;

  /**
   * True if this set is empty.  Default implementatino returns size() == 0.
   */
  virtual bool empty() const;

  /** 
   * Packs this rank set into the buffer provided. 
   * Currently, user is responsible for knowing which rank set he is using
   * and for using a specific unpack method.
   */
  virtual void pack(void *buf, int bufsize, int *position, MPI_Comm comm) const = 0;


  /** Get the packed size of the rank set. */
  virtual int packed_size(MPI_Comm comm) const = 0;


  /** 
   * Outputs the rank set to the provided stream. 
   * Returns number of characters output.
   */
  virtual int soutput(char *buf) const = 0;


  /**
   * Clones this rank set.
   */
  virtual RankSet *clone() const = 0;


  /* Empty (but virtual) destructor. */
  virtual ~RankSet() {}
};


#endif //RANK_SET
