/*
 * Filename: VParam.h
 * Author:   ratn1
 * Created:  Wed Oct 31 14:20:05 2007
 * $Id: VParam.h,v 1.5 2009/01/13 23:26:16 gamlin Exp $
 */
#ifndef __VPARAM_H__
#define __VPARAM_H__

#include <stdint.h>

typedef enum param_type {
  Param_undef,
  Param_scalar,
  Param_array
} param_type_t;
#ifdef __cplusplus
#include <list>
#include "RList.h"
#include "mpi_util.h"

using namespace std;

class Param {
  public:
  //  virtual int64_t getValue () = 0;
  /* get the size of the parameter */
  virtual int getSize() = 0;
  
  /* write param to buffer buf */
  virtual int output (char *buf) = 0;
  
//  virtual Param& operator= (const Param& p) = 0;
//  virtual bool   operator==(const Param& p) = 0;
//  virtual bool   operator!=(const Param& p) = 0;
  virtual int packed_size (MPI_Comm comm) = 0;
  virtual void pack(void *buf, int bufsize, int *position, MPI_Comm comm) = 0;

  virtual ~Param() {}
};

class ScalarParam : public Param {
  private:
    int64_t value;
  public:
    ScalarParam () : value (0) { }
    ScalarParam (int64_t val) : value (val) { }
    int64_t getValue () { return value; }
    int getSize() {
      return (int)(sizeof(int64_t));
    }
    /*
    ScalarParam& operator=(const ScalarParam& p) {
      value = p.value;
      return *this;
    }
    ScalarParam& operator=(int64_t p) {
      value = p;
      return *this;
    }*/
    bool operator==(const ScalarParam& p) {
      return (value == p.value);
    }
    bool operator==(int64_t p) {
      return (value == p);
    }
    bool operator!=(const ScalarParam& p) {
      return (value != p.value);
    }
    int packed_size (MPI_Comm comm) {
      return ::packed_size (sizeof(int64_t), MPI_BYTE, comm);
    }
    void pack(void *buf, int bufsize, int *position, MPI_Comm comm) {
      PMPI_Pack(&value, sizeof(int64_t), MPI_BYTE, buf, bufsize, position, comm);
    }
    static ScalarParam* unpack (void *buf, int bufsize, int *position, MPI_Comm
        comm) {
      ScalarParam * param = new ScalarParam ();
      PMPI_Unpack(buf, bufsize, position, &param->value, sizeof (int64_t), MPI_BYTE, comm);
      return param;
    }
    int output (char *buf);
};

class ArrayParam : public Param {
  private:
    int size;
    int * entries;
  public:
    ArrayParam () : size (0), entries (NULL) { }
    ArrayParam (int sz, int * list): size (sz) {
      entries = new int[sz];
      for (int i = 0; i  < sz; i++)
        entries [i] = list [i];
    }
    virtual ~ArrayParam() { delete[] entries; }
    int getSize () {
      return sizeof (int) + size * sizeof (int);
    }
    int getArraySize () {
      return size;
    }
    int * getArray () {
      return entries;
    }
    bool operator==(ArrayParam *a) {
      if (size != a->size) return false;
      for (int i = 0; i < size; i++) {
        if (entries[i] != a->entries[i]) return false;
      }
      return true;
    }
    int packed_size (MPI_Comm comm) {
      int sz = 0;
      sz = ::packed_size (1, MPI_INT, comm);
      sz+= ::packed_size (size, MPI_INT, comm);
      return sz;
    }
    void pack(void *buf, int bufsize, int *position, MPI_Comm comm) {
      PMPI_Pack(&size, 1, MPI_INT, buf, bufsize, position, comm);
      for (int i = 0; i < size; i++)
        PMPI_Pack(&entries[i], 1, MPI_INT, buf, bufsize, position, comm);
      return;
    }
    static ArrayParam* unpack (void *buf, int bufsize, int * position, MPI_Comm
        comm) {
      ArrayParam * param = new ArrayParam ();
      PMPI_Unpack (buf, bufsize, position, &param->size, 1, MPI_INT, comm);
      param->entries = new int[param->size];
      for (int i = 0; i < param->size; i++)
        PMPI_Unpack (buf, bufsize, position, &param->entries[i], 1, MPI_INT, comm);
      return param;
    }
    int output (char * buf);
};

typedef struct value_list {
  param_type_t type;
  Param * value;  // all values can be stored in 64 bits hopefully
  RList * rlist;  // list of participants with this value 
} value_list;

class RLParam {
  private:
    int64_t param;          // param being stored, such as SOURCE, DEST etc
    list<value_list*> vlist;  // list of (value, rank-list) tuples
  public:

    RLParam ();
   ~RLParam ();

    RLParam (int64_t p);

    int64_t get_param () { return param; }
    param_type_t param_type ();

    int add (int64_t val, int * rlist);
    int add (const pair< int , int *> & arr, int* rlist);

    void merge (RLParam * other);

    int output (char *);

    int packed_size (MPI_Comm comm);

    void     pack   (void *buf, int bufsize, int *position, MPI_Comm comm);

    static RLParam * unpack (void *buf, int bufsize, int *position, MPI_Comm comm);
};

extern "C" {
#endif

typedef struct Param Param;  
typedef struct RLParam RLParam;  
typedef void * RLParamList;

void rlparam_merge (int64_t p,
    RLParamList* rlparam1,
    RLParamList* rlparam2,
    int64_t val1, int * rlist1,
    int64_t val2, int * rlist2);

void rlparam_merge_arr (int64_t p,
    RLParamList* rlparam1,
    RLParamList* rlparam2,
    int arr1size, int * arr1, int * rlist1,
    int arr2size, int * arr2, int * rlist2);

void rlparam_init (RLParamList* v);
void rlparam_free (RLParamList* v);

int rlparam_packed_size (RLParamList rlparam, MPI_Comm comm);

void rlparam_pack (RLParamList rlparam, void * buf, int bufsize, int * position, MPI_Comm comm);

RLParamList rlparam_unpack (void * buf, int bufsize, int * position, MPI_Comm comm);

int rlparam_output (RLParamList rlparam, int64_t param, char *buf);

int64_t * rlparam_get_int64 (char * buf, int rank, int * size);

param_type_t rlparam_type (int64_t param, RLParamList rlparam);
void str2array (char * buf, char * end, size_t * size, int64_t **arr);
#ifdef __cplusplus
}
#endif

#endif 
/* vim: set tw=80 sw=2 expandtab: */
