#ifndef BIN_OP_H
#define BIN_OP_H

#include <mpi.h>
#include <stdio.h>

#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <stdint.h>
#include <inttypes.h>
#include <limits.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "mpi_util.h"
#include "TimeList.h"
#include "VParam.h"
//#include "umpi_internal.h"
#include "umpi_mpi_ops.h"
#include "cvector.h"
#include "opstruct_macro.h"
#include "Event.h"

#ifdef __cplusplus
extern "C" {
#endif

// all datatypes, communicators are converted to integers
// so the trace records can contain integers
typedef int OP_Datatype;
typedef int OP_Comm;
typedef int OP_Group;
typedef int OP_Op;



typedef struct {
  field_mask_t field;   // bitmask value for the field
  const char *format;   // format string for the field's value
  const char *name;     // name of the field (as per replay_op definition)
} op_field_desc_t;

#ifdef FEATURE_FILE_IO
//Added by Karthik - Start
typedef struct file_offset_info {
  int startpos;		//Start position of the file view
  int stride;		//Number of steps for next position
  int totalcount;	//Total count
} file_offset_info_t;

typedef struct __posix_args {
  int flags;
  mode_t mode;
} posix_args;
//Added by Karthik - End
#endif

extern op_field_desc_t op_field_table[];

/** Looks up a field in the descriptor table by value. Returns NULL if not found. */
op_field_desc_t *get_field_descriptor(field_mask_t value);

/* ==================================================================================================== */

/* binary data structure for ops */
typedef struct {
  /* meta info to parse the op */
  int sync_bytes;
  int seq_num;
  MPI_Op_t op;
  int arrays;
  int array_size1;
  int array_size2;
  int array_size3;
  int array_size4;
  int array_size_displs;

  /* This bitfield designates which param fields below are used. */
  field_mask_t fields;

  /*Added by Karthik - Start*/
  #ifdef FEATURE_FILE_IO
  /*This is extension of above bitfield to accomodate new parameters*/
  field_mask_t fields_extn;
  int arrays_extn;
  #endif
  /*Added by Karthik - End*/

  /* which fields are stored as vectors? */  
  field_mask_t pvectors;   
  /* vector of params */
  RLParamList rlparam;

  /* Which fields are stored as offset from rank? */
  field_mask_t rank_offsets;

  /* params */
  OP_Comm global_world;
  int null_req;
  OP_Comm comm;
  int count;
  cvector * count_vector;
  OP_Datatype datatype;
  int recvcount;
  OP_Datatype recvtype;
  OP_Op mpi_op;
  int root;
  cvector * root_vector;
  int tag;
  int dest;
  cvector * dest_vector;
  OP_Group group;
  int request;         /* Offset of request?  Not actual request? */
  cvector * request_vector;
  int color;
  int key;
  OP_Group group1;
  OP_Group group2;
  int local_leader;
  OP_Comm peer_comm;
  int remote_leader;
  int high;
  int source;
  cvector * source_vector;
  int recvtag;
  OP_Datatype newtype;
  OP_Datatype oldtype;
  int blocklen;
  int stride;
  int reorder;
  int newrank;
  int outcount;
#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
  //                             |__|-|__|-|__|-|__|-.. time_list
  // time_info[compute]  |__|---
  //                             time

  //                         |__|-|__|-|__|-|__|-.. time_list
  // time_info[mpi]  |__|---
  //                         time
  // time_list is a list of (histogram, histogram-limits) tuple
  struct time_data time_info[TIMING_MAX];

  void * replay_prev_op;
  void * replay_histo;
#endif

  /* array ptrs */
  array_elt_t *array1;
  array_elt_t *array2;
  array_elt_t *array3;
  array_elt_t *array4;
  MPI_Aint    *array_of_displacements;

#ifdef FEATURE_FILE_IO
  /* MPI_FILE_xxx*/
  char *filename;
  char *datarep;
  int amode;
  int fh;
  int offset;
  int disp;
  int etype;
  int filetype;
  int whence;
  /*Added by Karthik - Start*/
  file_offset_info_t offsetinfo;
  int is_mpi_io;
  char *filepath;
  int rank;
  int comm_size;
  int order;
  cvector * disp_vector;
  posix_args posixargs;
  /*Added by Karthik - End*/
#endif
} replay_op;

extern replay_op operation;
extern Event* event;
void init_op(replay_op *operation);

/**
 * Below code disables/enables timing depending on if we need it.
 * This avoids unnecessary gettimeofday calls for plain traces.
 */
#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
void initTime ();
void recordTime(replay_op *operation, int index);
#else
#define initTime()
#define recordTime(op, index)
#endif

/* ==================================================================================================== */
/* These macros will tell you if a particular array field is used and, if so, which one it is.          */
/* Looking at these, you can also see which fields are stored in which array in the opstruct.           */
/* ==================================================================================================== */
#define array_type1(arraymask) (                                     \
    (arraymask & RECVCOUNTS)         ? RECVCOUNTS :                  \
    (arraymask & ARRAY_OF_DIMS)      ? ARRAY_OF_DIMS :	             \
    (arraymask & DIMSP)              ? DIMSP :		             \
    (arraymask & ARRAY_OF_DEGREES)   ? ARRAY_OF_DEGREES :            \
    (arraymask & RANGES)             ? RANGES :		             \
    (arraymask & ARRAY_OF_RANKS)     ? ARRAY_OF_RANKS :	             \
    (arraymask & ARRAY_OF_REQUESTS)  ? ARRAY_OF_REQUESTS :           \
    (arraymask & ARRAY_OF_BLOCKLENS) ? ARRAY_OF_BLOCKLENS :          \
    (arraymask & ARRAY_OF_GSIZES)    ? ARRAY_OF_GSIZES :             \
    0)

#define array_type2(arraymask) (                                     \
    (arraymask & DISPLS)                 ? DISPLS :		     \
    (arraymask & ARRAY_OF_PERIODS)       ? ARRAY_OF_PERIODS :	     \
    (arraymask & ARRAY_OF_EDGES)         ? ARRAY_OF_EDGES :	     \
    (arraymask & ARRAY_OF_DISPLACEMENTS) ? ARRAY_OF_DISPLACEMENTS :  \
    (arraymask & ARRAY_OF_DISTRIBS)      ? ARRAY_OF_DISTRIBS :       \
    0)

#define array_type3(arraymask) (                                     \
    (arraymask & COUNTS)         ? COUNTS :                          \
    (arraymask & ARRAY_OF_TYPES) ? ARRAY_OF_TYPES :                  \
    (arraymask & ARRAY_OF_DARGS) ? ARRAY_OF_DARGS :                  \
    0)

#define array_type4(arraymask) (                                     \
    (arraymask & RDISPLS)         ? RDISPLS :                        \
    (arraymask & ARRAY_OF_PSIZES) ? ARRAY_OF_PSIZES :                \
    0)

/**
 * This will print out the name and contents of an array, along with a label indicating
 * what type of array it is.
 */
int soutput_array(char *buf, field_mask_t arr_type, int size, array_elt_t *array);

/*Added by Karthik - Start
#ifdef FEATURE_FILE_IO
int soutput_array_extn(char *buf, field_mask_t arr_type, int size, array_elt_t *array);
#endif
Added by Karthik - End*/

/* ==================================================================================================== */
/* These functions can be used to easily map over all the fields in an op descriptor.                   */
/* ==================================================================================================== */
/** Outputs the properly formatted value of the field in the op specified. */
int soutput_field_value(char *buf, replay_op *op, op_field_desc_t *field);

/** Function type field handlers.  Define  these and pass to map_over_fields. */
typedef void (field_handler_t)(replay_op *op, op_field_desc_t *field);

/** Dispatches a function for each field in the bitmask and passes it the corresponding descriptor. */
void map_over_fields(replay_op *op, field_handler_t func);

/* ==================================================================================================== */

#ifdef __cplusplus
}
#endif

#endif /* BIN_OP_H */
