#ifndef RANK_RECURSIVE_H
#define RANK_RECURSIVE_H
/**
 * This file contains routines for dealing with recursive rank lists.
 * Rank lists are int arrays formatted as follows:
 * 
 * [ num_lists 
 *   [depth startrank [ iters stride ] [ iters stride ] ... ] 
 *   ...
 * ]
 * 
 * The first element is the number of lists contained in the array.
 * 
 * If the first element is nonzero, The second element is the start of the header of
 * the first recursive list.  The header is two ints:  The first is the depth, or the
 * number of (iters, stride) pairs in the entire list.  The second is the starting rank
 * for the list.
 * 
 * In this implementation, the depths of lists are stored as negative numbers.  I assume
 * this was done initially for readability in lists printed out to the command line.
 *
 * To unwind a recursive list, rank is set to the start rank, and for each (iter, stride) pair
 * in the list rank is incremented by (stride) (iters) times.  For each iteration, subsequent
 * (iter, stride) pairs are processed recursively in the same way.  The set of all of rank's values 
 * at leaves in this recursive process specifies the full participant list.
 *
 * It may help to think of (iter, stride) pairs as a range in one dimension in a space of ranks. 
 * each of these dimensions is crossed with subsequent (iter, stride) pairs to define the 
 * full participant list.  e.g.:
 *
 *  0   1   2   3   4
 *    +----------+
 *  5 | 6   7   8|  9
 *    |          |   
 * 10 |11  12  13| 14
 *    |          |   
 * 15 |16  17  18| 19
 *    +----------+   
 * 20  21  22  23  24
 * 
 * 25  26  27  28  29
 * 
 * The ranks 6,7,8,11,12,13,16,17,18 can be specified with the recursive rank list:
 * 
 *    1 2 6 3 5 3 1
 * 
 * This specifies that the list contains 1 list, with depth of 2 pairs and a start rank of 6.
 * The first pair is 3 iterations of stride 5.  This covers 6, 11, and 16.  Crossing each of these
 * with 3 iterations of stride 1, we have (6,7,8), (11, 12, 13), and (16, 17, 18): the full 
 * set of participating ranks.
 * 
 * Note that since depths are stored negative (for readability?), this list will appear in 
 * output as:
 * 
 *    1 -2 6 3 5 3 1
 * 
 * Documentation by Todd Gamblin.
 */
#include <limits.h>
#include <stdio.h>
#include <stdint.h>
#include <mpi.h>
#include "mpi_util.h"

#ifdef __cplusplus
extern "C" {
#endif


#define UNROLL_BUFSIZE 4096
#define RANK_END (INT_MAX)

/** 
 * Visitor function type for running through recursive lists.
 * Last is the result of the last call to an visitor, rank is
 * the rank being visited.  Use this to implement accumulation/reduction
 * functions on rank lists.
 */
typedef void (rank_visitor_fn_t)(void *data, int rank);

/** 
 * Calls visitor with each rank in a recursive list. 
 * Returns result of last call to the visitor.
 */
void traverse_rlist(int *list, void *data, rank_visitor_fn_t visit);


void foutput_recursive_list(FILE *out, int *list);
void foutput_expanded_list(FILE *out, int *list);
int soutput_range_list(int *rlist, char *buf);

// parses a recursive rank list from a buffer.
int *parse_recursive_list(const char *buf);

int size_recursive_list(int *list);
int member_rank(int rank, int *list);

// returns a count of members in the recursive list.
int member_count(int *list);

int create_empty_rlist(int **list);
int   create_recursive_list     (int rank, int **list);
void  merge_recursive_lists     (int **range1, int *range2);
int   diff_recursive_lists      (int **range1, int *range2);
int * intersect_recursive_lists (int *list1, int *list2);
void  dup_recursive_list        (int **dest, int *src);

// creates space for an rlist but does not init.  Use for data transfer
int *alloc_rlist(int len); 
void destroy_recursive_list(int *list);

// helper functions
void merge_single(int rank, int **list);
int merge_compress(int **list);
void merge_unroll(int *list, int *unrolled);

void recursive_list_pack (int * rlist, void * buf, int bufsize, int * position, MPI_Comm comm);
int  recursive_list_packed_size (int * rlist, MPI_Comm comm);
int *recursive_list_unpack (void * buf, int bufsize, int * position, MPI_Comm comm);

#ifdef __cplusplus
}
#endif

#endif /* RANK_RECURSIVE_H */
