#include "AddressTranslator.h"

#include <sys/stat.h>
#include <libiberty.h>
#ifdef HAVE_DEMANGLE
#include <demangle.h>
#endif

#include <iostream>
#include <cstdio>
#include <cerrno>
#include <cctype>
#include <sstream>
using namespace std;


static void kill_prefix(vector<string>& lines) {
  vector<string>::iterator next_to_last = lines.end();
  next_to_last--;
  lines.erase(lines.begin(), next_to_last);
}


/* Wrapper around cplus_demangle.  Strips leading underscores and
   other such chars that would otherwise confuse the demangler.  */
char * AddressTranslator::demangle (bfd *abfd, const char *name) {
#ifdef HAVE_DEMANGLE
  char *res, *alloc;
  const char *pre, *suf;
  size_t pre_len;

  if (abfd != NULL && bfd_get_symbol_leading_char (abfd) == name[0])
    ++name;

  /* This is a hack for better error reporting on XCOFF, PowerPC64-ELF
     or the MS PE format.  These formats have a number of leading '.'s
     on at least some symbols, so we remove all dots to avoid
     confusing the demangler.  */
  pre = name;
  while (*name == '.')
    ++name;
  pre_len = name - pre;

  alloc = NULL;
  suf = strchr (name, '@');
  if (suf != NULL)
  {
    alloc = (char*)malloc (suf - name + 1);
    memcpy (alloc, name, suf - name);
    alloc[suf - name] = '\0';
    name = alloc;
  }

  res = cplus_demangle (name, DMGL_ANSI | DMGL_PARAMS);
  if (res != NULL)
  {
    /* Now put back any suffix, or stripped dots.  */
    if (pre_len != 0 || suf != NULL)
    {
      size_t len;
      size_t suf_len;
      char *final;

      if (alloc != NULL)
        free (alloc);

      len = strlen (res);
      if (suf == NULL)
        suf = res + len;
      suf_len = strlen (suf) + 1;
      final = (char*)malloc (pre_len + len + suf_len);

      memcpy (final, pre, pre_len);
      memcpy (final + pre_len, res, len);
      memcpy (final + pre_len + len, suf, suf_len);
      free (res);
      res = final;
    }

    return res;
  }

  if (alloc != NULL)
    free (alloc);

  return xstrdup (pre);
#else
  return (char*)name;
#endif
}


void AddressTranslator::lowercase(std::string& s) {
  for (unsigned i=0; i < s.size(); i++) {
    s[i] = tolower(s[i]);
  }
}

/* Returns the size of the named file.  If the file does not
   exist, or if it is not a real file, then a suitable non-fatal
   error message is printed and zero is returned.  */
off_t AddressTranslator::get_file_size (const char * file_name)
{
  struct stat statbuf;
  
  if (stat (file_name, &statbuf) < 0)
  {
    if (errno == ENOENT)
      cerr << "'" << file_name <<"': No such file" << endl;
    else
      cerr << "Warning: could not locate '" << file_name << "'.  reason: " 
           << strerror (errno) << endl;
  }  
  else if (! S_ISREG (statbuf.st_mode))
    cerr << "Warning: '" << file_name << "' is not an ordinary file" << endl;
  else
    return statbuf.st_size;

  return 0;
}


void AddressTranslator::slurp_symtab (bfd *abfd)
{
  long symcount;
  unsigned int size;

  if ((bfd_get_file_flags (abfd) & HAS_SYMS) == 0)
    return;

  symcount = bfd_read_minisymbols (abfd, false, (void **) &syms, &size);
  if (symcount == 0)
    symcount = bfd_read_minisymbols (abfd, true /* dynamic */, (void **) &syms, &size);

  if (symcount < 0) {
    cerr << bfd_get_filename (abfd) << endl;
  }
}

/* Look for an address in a section.  This is called via
   bfd_map_over_sections.  */
void AddressTranslator::find_address_in_section(bfd *abfd, asection *section) {
  bfd_vma vma;
  bfd_size_type size;

  if (found)
    return;

  if ((bfd_get_section_flags (abfd, section) & SEC_ALLOC) == 0)
    return;

  vma = bfd_get_section_vma (abfd, section);
  if (pc < vma)
    return;

#ifdef bfd_get_section_size_before_reloc
  size = bfd_get_section_size_before_reloc (section);
#else /* Newer version of binutils */
  size = bfd_get_section_size (section);
#endif /* size incompatibility crap */
  if (pc >= vma + size)
    return;

  found = bfd_find_nearest_line (abfd, section, syms, pc - vma,
                                 &filename, &functionname, &line);
}


void AddressTranslator::translate(int count, const intptr_t *addrs, vector<string>& lines) {
  for (int i=0; i < count; i++) {
    if (walk_outward)
      pc = addrs[count - i - 1];
    else
      pc = addrs[i];

    found = false;
    asection *sec;
    for (sec = abfd->sections; sec != NULL; sec = sec->next) {
      find_address_in_section(abfd, sec);
    }

    ostringstream name_stream;
    if (! found) {
      if (with_functions) name_stream << "?? (";
      name_stream << "??:0";
      if (with_functions) name_stream << ")";
      return;
      //continue;
    } 

    do {
      const char *name;
      char *alloc;

      name = functionname;
      if (name == NULL || *name == '\0') {
        name = "??";

      } else if (do_demangle) {
        alloc = demangle (abfd, name);
        name = alloc;
      }

      if (strip_prefix.size() != 0) {
        string lowername(name);
        lowercase(lowername);
        if (strstr(lowername.c_str(), strip_prefix.c_str()) == lowername.c_str()) {
          if (only_site) kill_prefix(lines);
          return;	      // stop processing if we found the bailout prefix
        }
      }

      if (with_functions) {
        name_stream << name << " (";
      }

      if (base_names && filename != NULL) {
        char *h;
        h = strrchr (filename, '/');
        if (h != NULL) {
          filename = h + 1;
        }
      }

      name_stream << (filename ? filename : "??") << ":" << line;

      if (with_functions) name_stream << ")";


      found = false;
#ifdef bfd_find_inliner_info
      // Do this if we have a sufficiently late binutils version to support it.
      if (unwind_inlines) {
        found = bfd_find_inliner_info (abfd, &filename, &functionname, &line);
      }
#endif /* bfd_find_inliner_info */

      lines.push_back(name_stream.str());
    } while (found);
    //} while (1);
  }

  if (only_site) kill_prefix(lines);
}

void AddressTranslator::translate(int count, const char **addrs, vector<string>& lines) {
  intptr_t pointers[count];
  for (int i=0; i < count; i++) {
    pointers[i] = bfd_scan_vma (addrs[i], NULL, 16);  // convert hex string -> bfd address type
  }
  translate(count, pointers, lines);
}

// Do library initialization once.  Called on construction to make sure lib is inited.
void AddressTranslator::initBfd() {
  static bool inited = false;
  if (inited) return;
  
  bfd_init ();
  inited = true;
}


void AddressTranslator::setStripPrefix(const std::string& value) {
  strip_prefix = value; 
  lowercase(strip_prefix);
}



AddressTranslator::AddressTranslator(const char *file_name, const char *_target)
  : unwind_inlines(false), with_functions(true), do_demangle(true), base_names(true), only_site(false),
    syms(NULL), target(_target ? strdup(_target) : NULL)
{
  initBfd(); // make sure lib is inited

  char **matching;
  if (get_file_size (file_name) < 1) {
    cerr << "Fatal error: Couldn't open " << file_name << endl;
    exit(1);
  }

  abfd = bfd_openr (file_name, target.get());
  if (abfd == NULL) {
    cerr << "Fatal error: Couldn't open " << file_name << endl;
    exit(1);
  }

  if (bfd_check_format (abfd, bfd_archive)) {
    cerr << file_name << ": cannot get addresses from archive" << endl;
    exit(1);
  }

  if (! bfd_check_format_matches (abfd, bfd_object, &matching)) {
    cerr << bfd_get_filename (abfd) << ":" << endl;
    if (bfd_get_error () == bfd_error_file_ambiguously_recognized) {
      cerr << "Matching formats:" << endl;
      char **p = matching;
      while (*p) {
        cerr << " " << *p++; 
      }
      cerr << endl;
      free (matching);
    }
    exit (1);
  }

  slurp_symtab (abfd);
}


AddressTranslator::~AddressTranslator() {
  if (syms != NULL)
  {
    free (syms);
    syms = NULL;
  }

  bfd_close (abfd);
}


