#include "BitVector.h"

#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <memory>
#include <set>
using namespace std;

#include "mpi_util.h"

/** 
 * Utility function to print out ids in compressed form, in that
 * Strings of consecutive ids are printed as ranges.
 * e.g., if the ids are 0 1 2 3 4 5, then this will print out 0-5.
 * <p>
 * NOTE: ids between begin and end MUST be sorted.
 */
template <class FwdIterator> 
void printIds(FwdIterator begin, FwdIterator end, std::ostream& out = std::cout) {
  if (begin == end) return;
  
  int first = *begin;
  int expected = first;
  
  for (FwdIterator i = begin; i != end; ++i) {
    int cur = *i;
    if (cur == expected) {
      expected++;
    } else {
      int last = expected - 1;
      if (first == last) {
	out << first << " ";
      } else {
	out << first << "-" << last << " ";
      }
      first = cur;
      expected = cur + 1;
    }
  }
  
  int last = expected - 1;
  if (first == last) {
    out << first;
  } else {
    out << first << "-" << last;
  }
}



BitVector::BitVector(int _size, bool initialValue) :
  numElts((_size >> LOG2_INT) + (_size & INT_REM ? 1 : 0)),
  realSize(_size),
  bits(new elt_t[numElts])
{
  //start vector out with initial value
  if (initialValue) {
    setAll();
  } else {
    clear();
  }
}


BitVector::~BitVector() {
  delete [] bits;
}


void BitVector::clear() {
  memset(bits, 0, sizeof(elt_t) * numElts);
}


void BitVector::setAll() {
  memset(bits, ~0, sizeof(elt_t) * numElts);
}


BitVector& BitVector::operator&=(const BitVector& other) {
  assert(other.realSize == realSize);
  
  for (int i=0; i < numElts; i++) {
    bits[i] &= other.bits[i];
  }

  return *this;
}


BitVector& BitVector::operator|=(const BitVector& other) {
  if (other.realSize != realSize) {
    cerr << "Error: bitvectors not the same size!" << endl;
    exit(1);
  }
  
  for (int i=0; i < numElts; i++) {
    bits[i] |= other.bits[i];
  }

  return *this;
}


/**
 * Returns the number of hex digits required to represent the first field
 * of this BitVector.  Enables us to output these w/o a lot of padding zeros
 * for small clusters.
 *
 * Optionall can supply a format, to which a proper format string will be
 * printed.
 */ 
static int firstFieldWidth(int realSize, char *format = NULL) {
  int digits = realSize >> 2;  // total number of hex digits
  if (realSize % 4) digits++;

  // figure out width of first element in bit vector based on size
  int initial_width = (digits % 16);
  if (!initial_width) initial_width = 16;

  if (format) {
    sprintf(format, "%%0%dlX", initial_width);
  }

  return initial_width;
}




int BitVector::soutput(char *buf) const {
  char *pos = buf;

  // output size first.
  pos += sprintf(pos, "[%d|", realSize);

  // first digit is unpadded.
  char format[8];
  int width = firstFieldWidth(realSize, format);
  pos += sprintf(pos, format, bits[0] & ( (1 << width*4) - 1) );

  // output fixed-width element fields
  for (int i=1; i < numElts; i++) {
    pos += sprintf(pos, "%016lX", bits[i]);
  }

  pos += sprintf(pos, "]");

  return (pos - buf);
}


BitVector *BitVector::input(const char *buf, int *count) {
  const char *pos = buf;

  // read in size
  int size;
  if (sscanf(pos, "[%d|", &size) != 1) {
    cerr << "Malformed BitVector: " << buf << endl;
    exit(1);
  }
  while (*pos++ != '|'); // skip size and move to start of bits
  assert(size);

  // create the bit vector.
  auto_ptr<BitVector> bv(new BitVector(size));

  // figure out width of first element in bit vector based on size
  char format[8];
  int width = firstFieldWidth(size, format);
  if (sscanf(pos, format, &bv->bits[0]) != 1) {
    cerr << "Malformed BitVector: couldn't read first digit." << buf << endl;
    exit(1);
  }
  pos += width;
  
  // read in remainder of bits in 16-digit pieces
  for (int i=1; i < bv->numElts; i++) {
    if (sscanf(pos, "%016lX", &bv->bits[i]) != 1) {
      cerr << "Malformed BitVector: couldn't read digit." << buf << endl;
      exit(1);
    }
    pos += 16;
  }
  
  if (*pos != ']') {
    cerr << "Malformed BitVector." << buf << endl; 
    exit(1);
  }
  
  if (count) *count = (pos + 1 - buf);
  
  return bv.release();
}


ostream& operator<<(ostream& out, const BitVector& vec) {
  set<unsigned> members;
  for (int i=0; i < vec.capacity(); i++) {
    if (vec.member(i)) members.insert(i);
  }
  
  out << "<";
  printIds(members.begin(), members.end(), out);
  out << ">";
  return out;
}


int BitVector::size() const {
  int count = 0;
  for (int i=0; i < numElts; i++) {
    int n = bits[i];
    while (n) {    // This loop is O(set bits)
      count++;
      n &= (n-1);  // this will unset the lowest set bit.
    }
  }
  return count;
}


int BitVector::packed_size(MPI_Comm comm) const {
  int size = ::packed_size(1, MPI_INT, comm);
  size += ::packed_size(numElts, MPI_LONG_LONG_INT, comm);
  return size;
}


void BitVector::pack(void *buf, int bufsize, int *position, MPI_Comm comm) const {
  int size = realSize;
  PMPI_Pack(&size, 1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(bits, numElts, MPI_LONG_LONG_INT, buf, bufsize, position, comm);
}


BitVector *BitVector::unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  int realSize;
  PMPI_Unpack(buf, bufsize, position, &realSize, 1, MPI_INT, comm);

  auto_ptr<BitVector> result(new BitVector(realSize));
  PMPI_Unpack(buf, bufsize, position, result->getData(), result->getNumElts(), MPI_LONG_LONG_INT, comm);

  return result.release();
}


BitVector *BitVector::clone() const {
  BitVector *copy = new BitVector(realSize);
  memcpy(copy->bits, bits, sizeof(elt_t) * numElts);
  return copy;
}


#if 0
int main ()
{
  BitVector * b = new BitVector (144, true);
  char buf [512];
  int len;
  //char inputbuf[] = "[144|000000000000000000000000000000000000]";
  char inputbuf[] = "[144|FFFFFFFF00000000FFFFFFFF00000000FFFFFFFF]";
  len = b->soutput (buf);

  cout << "len: " << len << endl;
  cout << "buf: " << buf << endl;

  b = BitVector::input (inputbuf, &len);
  len = b->soutput (buf);

  cout << "len: " << len << endl;
  cout << "buf: " << buf << endl;

  return 0;
}
#endif
