/*
 * Filename: Histogram.C
 * Author:   pratn
 *   tgamblin     C++ class-ified
 *   tgamblin     added value_for_rank
 * 
 * Created:  Fri Aug 31 10:08:27 2007
 * $Id: Histogram.C,v 1.5 2008/12/26 20:44:49 gamlin Exp $
 */
#include "Histogram.h"

#include <sys/time.h>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <cassert>
#include <cfloat>
#include <cstring>
#include <sstream>
using namespace std;

#include "trace.h"
#include "opstruct.h"
#include "rand_max.h"
#include "utils.h"
#include "RList.h"
#include "BitVector.h"
#include "mpi_util.h"

#define weighted_avg(m, a, n, b) (((a)*(m)+(b)*(n))/((m)+(n)))

// The combined sigma of two samples is :
// S^2=((N1-1)*S1^2+(N2-1)*S2^2)/(N1+N2-2)
// we store the value (n-1)sigma^2, hence this :-
#define combined_sigma(s, m, t, n) \
  ((((s) + (t))*((m) + (n) - 1))/((m) + (n) - 2))


// defaults for histo params.
// TODO: get rid of ranks in histograms.
int Histogram::ranks_in_histograms = 1;     // rank-preserving or non-rank-preserving histos
int Histogram::default_buckets = 5;         // buckets in default histograms
int Histogram::default_fixwidth = 1;        // fix width of floats in histograms.
int Histogram::default_smooth_threshold = 1;
int Histogram::default_merge_limit = 3;
int Histogram::default_merge_ratio = 10;

// This determins what sort of RankSet we use in the histogram.
// Note: MUST extend RankSet.
// TODO: roll these into a policy and make Histogram a template w/rank policy?
/*
typedef RList histo_rank_set_t;
static histo_rank_set_t *create_rank_set() {
  return new RList();
}
*/

typedef BitVector histo_rank_set_t;
static histo_rank_set_t *create_rank_set() {
  int size = get_my_size ();
  return new BitVector(size);
}


/** This constructor sets only magic.  Used for inputting histo bins. */
histogram_bin::histogram_bin() 
  : magic(BIN_MAGIC), 
    start(0), min(0), max(0), end(0),
    frequency(0), average(0), variance(0), 
    participants(create_rank_set()),
    next(NULL)
{ }


/** This constructor sets magic, start, and end and sets everything else to zero. */
histogram_bin::histogram_bin(double _start, double _end)
  : magic(BIN_MAGIC), 
    start(_start), min(0.0), max(0.0), end(_end),
    frequency(0), average(0.0), variance(0.0), 
    participants(create_rank_set()), 
    next(NULL)
{ }


histogram_bin::~histogram_bin() { }

#if 0
void histogram_bin::add(double value) {
  if (frequency == 0) {
    average = value;
  }

  min = (value < min || min == 0) ? value : min;
  max = (value > max) ? value : max;
  frequency++;
  
  double old_avg = average;
  average += (((double)value - average))/frequency;
  //assert(average >= 0);

  variance += ((double)value - old_avg) * ((double)value - average);
  //assert(variance >= 0);

  TRACE ( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, variance));
}
#endif

void histogram_bin::add(double value, long long freq) {
  assert(freq >= 0);
  if (frequency == 0) {
    average = value;
  }
  if((frequency + freq) < 0) {
    printf("Frequency %d, freq %d\n", frequency, freq);
    return;
  }
  min = (value < min || min == 0) ? value : min;
  max = (value > max) ? value : max;
  //frequency++;

  double old_avg = average;
  long long old_freq = frequency;
  //average += (((double)value - average))/frequency;
  average = ((double)frequency * average + value*(double)freq)/(frequency+freq);
  //assert(average >= 0);
  frequency += freq;

  variance += ((double)value - old_avg) * ((double)value - average);
  //assert(variance >= 0);
  if(average < 0 || frequency < 0) {
    //assert(average > 0);
    average = old_avg;
    frequency = old_freq;
  }
  TRACE ( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, variance));
}


void histogram_bin::addParticipant(int rank) {
  if (Histogram::ranks_in_histograms) {
    participants->add(rank);
  }
}


/**
 * Initialize histogram with just the magic number.  This is the 
 * private constructor.
 */
Histogram::Histogram(): magic(HISTO_MAGIC), list(NULL), bins(NULL) { }


/**
 * Initialize histogram with valid bins.  This is the public constructor.
 */
Histogram::Histogram(int B, int smooth_thresh, int merge_lim, int merge_rat, int c_rank) :
  magic(HISTO_MAGIC),
  num_elems(0),
  num_bins(B),
  smooth_threshold(smooth_thresh),
  merge_limit(merge_lim),
  merge_ratio(merge_rat),

  // these are values for rank tracking.  TODO: orthogolinalize this stuff somehow.
  // Use policy class?  Add rank tracking policies?  argh... clutter.
  creator_rank(c_rank), max_rank(c_rank), min_rank(c_rank),
  max_value(DBL_MIN),  min_value(DBL_MAX),

  list(NULL),
  bins(NULL)
{
  init ();
}

Histogram::Histogram(int value, int B, 
		     const stack_sig *sig, int my_rank, int _max_rank, int _min_rank) :
  magic(HISTO_MAGIC),
  num_elems(0),
  num_bins(B),
  smooth_threshold(default_smooth_threshold),
  merge_limit(default_merge_limit),
  merge_ratio(default_merge_ratio),

  creator_rank(my_rank), 
  max_rank(_max_rank), min_rank(_min_rank), 
  max_value(value),    min_value(value),

  list(NULL),
  bins(NULL)
{
  init ();
  sig_assign(&signature, sig);
  add((double)value);
}

void Histogram::init ()
{
  sig_init(&signature);

  // we create bins here but leave ranges uninited.
  // ranges are set on first add.
  histogram_bin *tail = NULL;
  for (int i = 0; i < num_bins; i++) {
    histogram_bin *bin = new histogram_bin();

    if (tail) {
      tail->next = bin;  // attach to tail
    } else {
      bins = bin;      // attach head
    }
    tail = bin;
  }

}

#if 0
bool Histogram::has_other (int rank) {
  for (histogram_bin* bin = bins; bin; bin = bin->next) {
    int numElts = bin->participants->size();
    if (numElts > 1)
      return true;
    else if (numElts == 1) {
      if (!bin->participants->member(rank))
        return true;
    }
  }
  return false;
}
#endif

bool Histogram::has_other (int rank) {
  if(max_rank == min_rank && max_rank == rank)
    return false;
  else
    return true;
}

void Histogram::set_rank(int rank) {
  creator_rank = rank;
}


void Histogram::set_min_max_ranks (int min, int max) {
  min_rank = min;
  max_rank = max;
}


/*
 * print out histogram to a buffer
 * used to output histogram to trace
 */
int Histogram::output(char *buf, bool fix_width) const
{
  char *pos = buf;

  assert (magic == HISTO_MAGIC);

  pos += sig_soutput(pos, &signature);

  const char *format = (fix_width) ? 
    "(%d,%d,%d,%d,%d,%d,%d,%.2f,%.2f)" : "(%d,%d,%d,%d,%d,%d,%d,%.10g,%.10g)";
  pos += sprintf(pos, format,
		 creator_rank, max_rank, min_rank, 
		 num_bins, smooth_threshold, merge_limit, merge_ratio,
		 max_value, min_value);

  for (histogram_bin* bin = bins; bin; bin = bin->next) {
    double stddev = 0.0;

    if (bin->frequency > 1) {
      stddev = sqrt (bin->variance/(bin->frequency - 1));
    }

    //pos += sprintf (pos, "{%.10g,%.10g,%.10g,%.10g,%d,%.10g,%.10g,", 
    pos += sprintf (pos, "{%.2f,%.2f,%.2f,%.2f,%lld,%.2f,%.2f", 
		    bin->start, bin->min, bin->max, bin->end, bin->frequency, bin->average, stddev); 
    #if 0
    pos += sprintf (pos, "{%.2f,%.2f,%.2f,%.2f,%d,%.2f,%.2f,", 
		    bin->start, bin->min, bin->max, bin->end, bin->frequency, bin->average, stddev); 
    //Karthik - Participant list is not really necessary in trace output to infer statistical data
    //Can reenable it if any need arises.
    pos += bin->participants->soutput(pos);
    #endif
    pos += sprintf (pos, "}");

  }
  //pos += sprintf (pos, "\n");

  return (pos - buf);
}

/*
 * Scan buffer to reconstruct histogram.
 * used to input histogram from trace
 */
Histogram* Histogram::input (char * buf) {
  //TODO: this is getting pretty nasty.  
  //      needs some cleanup or maybe we need a parser?

  histogram_bin* bin = 0;
  double stddev = 0.0;
  char *c2, tmp[512];
  int f;
  double start, min, max, end, avg;

  
  auto_ptr<Histogram> h(new Histogram());
  
  char *ch = buf;

  if (*ch == '(') {
    sig_init (&h->signature);
  } else {
    sig_input(&h->signature, ch);
  }
  
  // rank and limits
  ch = strchr(ch, '(');
  if (9 != sscanf(ch, "(%d,%d,%d,%d,%d,%d,%d,%lf,%lf)", 
		  &h->creator_rank, &h->max_rank, &h->min_rank, 
		  &h->num_bins, &h->smooth_threshold, &h->merge_limit, &h->merge_ratio,
		  &h->max_value, &h->min_value))
  {
    cerr << "Malformed histogram (no limits)!" << endl;
    exit(1);
  }
  
  ch = strchr(buf, '{');
  if (!ch) {
    cerr << "Malformed histogram (no bins)!" << endl;
    exit(1);
  }

  bin = h->bins = new histogram_bin();
  if (7 != sscanf (ch, "{%lf,%lf,%lf,%lf,%lld,%lf,%lf,", 
		   &bin->start, &bin->min, &bin->max, &bin->end, 
		   &bin->frequency, &bin->average, &stddev)) {
    cerr << "Malformed histogram (bad bin)!" << endl;
    exit(1);    
  }
  bin->variance = stddev*stddev*(bin->frequency-1);

#if 0
  ch = strchr(ch, '[');

//Karthik - Commented out mandatory participant list chk
  if (!ch) {
    cerr << "Malformed histogram (no participants)!" << endl;
    exit(1);
  }
  bin->participants.reset(histo_rank_set_t::input(ch));
#endif
  c2 = strchr(ch, '[');
  if (c2) {
    ch = c2;
    c2=NULL;
    bin->participants.reset(histo_rank_set_t::input(ch));
  }
  
  long long num_elems = bin->frequency;
  int n = 1;

  ch = strchr (ch + 1, '{');
  while (ch && n < h->num_bins)
  {
    c2 = strchr (ch, '}');
    strncpy (tmp, ch, c2-ch+1);
    if (7 != sscanf (ch, "{%lf,%lf,%lf,%lf,%d,%lf,%lf,", &start, &min,
	    &max, &end, &f, &avg, &stddev))    {
      cerr << "Malformed histogram (bad bin)!" << endl;
      exit(1);    
    }
    bin->next = new histogram_bin(start, end);
    bin = bin->next;

  #if 0
    ch = strchr(ch, '[');
  //Karthik - Commented out mandatory participant list chk
    if (!ch) {
      cerr << "Malformed histogram (no participants)!" << endl;
      exit(1);
    }
    bin->participants.reset(histo_rank_set_t::input(ch));
  #endif
    c2 = NULL;
    c2 = strchr(ch, '[');
    if (c2) {
      ch=c2;
      c2=NULL;
      bin->participants.reset(histo_rank_set_t::input(ch));
    }
    if (bin == 0) abort ();
    bin->min = min;
    bin->max = max;
    bin->average = avg;
    bin->frequency = f;
    bin->variance = stddev*stddev*(bin->frequency-1);
    num_elems += bin->frequency;
    n++;
    ch = strchr (ch + 1, '{');
  }
  h->num_elems = num_elems;
  h->list = NULL;

  return h.release();
}


Histogram::~Histogram () {
  assert (magic == HISTO_MAGIC); 

  histogram_bin *b1 = 0;
  histogram_bin *b2 = 0;
  b1 = bins;

  while (b1) {
    b2 = b1->next;
    delete b1;
    b1 = b2;
  }

  // if the list was created, delete it.
  if (list) delete [] list;
}


/*
 * Construct a sorted list of bins. The bins are sorted by the frequency field.
 * This is used to determine the largest (split-able) and smallest (merge-able)
 * bins.
 */
void Histogram::sort() {
  assert (magic == HISTO_MAGIC);
  if (!list) {
    list = new histogram_bin*[num_bins];
  }

  histogram_bin* tmp = bins;
  for (int i = 0; i < num_bins && tmp; i++) {
    list[i] = tmp;
    tmp = tmp->next;
  }
  // insertion sort
  for (int i = 1; i < num_bins; i++)
  {
    int j = i;
    tmp = list[j];
    while (j > 0 && list[j-1]->frequency > tmp->frequency)
    {
      list[j] = list[j-1];
      j--;
    }
    list[j] = tmp;
  }

  return;
}

/*
 * Find the lowest frequency bin which can be merged with the next bin. The attr
 * fields define the threshold (absolute as well as percentage) which is used to
 * determine if two bins can be merged.
 * 
 * PRE: bins are sorted by frequency using sort.
 */
histogram_bin* Histogram::low_bin () {
  histogram_bin* bin = NULL;
  int i = 0;
  int merge = 0;

  assert (magic == HISTO_MAGIC);
  for (merge = 0; merge < merge_limit; merge++)
    for (i = 0; i < (num_bins - 1); i++)
    {
      bin = list[i];

      if (bin->next)
      {
        if (abs(bin->frequency - bin->next->frequency) <= merge) {
          return bin;
	}
      }
    }

  /* check for ratio difference */
  for (i = 0; i < (num_bins - 1); i++)
  {
    bin = list[i];
    if (bin && bin->next &&
        (abs(bin->frequency - bin->next->frequency)*(100/merge_ratio) < bin->frequency))
    {
      TRACE( ( "b =%d, b->next = %d\n", bin->frequency,
          bin->next->frequency));
      return bin;
    }
  }

  return NULL;
}

/*
 * 'smoothen' the histogram - finds the largest and smallest bins and then
 * splits the large bin only when two bins can be merged. This ensures that the
 * number of bins is maintained constant.
 */
void Histogram::smooth()
{
  double sigma = 0.0;
  double max, min;

  assert (magic == HISTO_MAGIC);

  sort();  // Sort and store the histogram separately

  // find the largest bin
  histogram_bin* high = list[num_bins - 1];

  // split if the largest bin has more than 1 element
  if (high->frequency <= 1)
  {
    TRACE ( ( "largest bin is too small\n"));
    return;
  }

  TRACE( ("{[%.2f, %.2f] %d %.2f %.2f}\n", high->start, high->end,
      high->frequency, high->average, high->variance));

  // Find the two smallest adjacent bins ....
  histogram_bin* low = low_bin();

  // .... do nothing if no low is found ....
  if (!(low && low->next)) {
    TRACE ( ( "No low found\n"));
    return ;
  }

  // ... or if high and low are same ...
  if (high == low->next || high == low) {
    TRACE ( ( "high and low coincide\n"));
    TRACE ( ( "high = %d, low = %d, low->next = %d\n", high->frequency, low->frequency, low->next->frequency));
    return;
  }

  // ... or if the merge would create a very large bin ...
  else if ((low->frequency + low->next->frequency) >= (high->frequency/2)) {
    TRACE ( ( "too large to merge\n"));
    TRACE ( ( "low = %d low->next = %d, high = %d\n", low->frequency, low->next->frequency, high->frequency));
    return;
  }


  // .... merge them if we haven't bailed yet.
  double old_avg[2];
  old_avg[0] = low->average;        // save old averages
  old_avg[1] = low->next->average;  // for later

  // adjust end point
  low->end = low->next->end;

  int freq = low->frequency + low->next->frequency;
  if (freq > 0)
  {
    low->average = 
      weighted_avg(low->frequency, low->average,
		   low->next->frequency, low->next->average);
    //assert (low->average >= 0);
  }

  if (freq == 0)
  {
    low->min = 0;
    low->max = 0;
    low->average = 0;
    low->variance = 0;
  }
  else if (freq == 1)
  {
    low->variance = 0;
    low->min = low->average;
    low->max = low->average;
  }
  else if (freq == 2)
  {
    // 1 element in each bin
    if (low->frequency == low->next->frequency)
    {
      low->variance = (old_avg[0] - low->average)*(old_avg[0] - low->average) + 
                      (old_avg[1] - low->average)*(old_avg[1] - low->average) ;
      //assert (low->variance >= 0);
      TRACE( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, low->variance));

      sigma = sqrt (low->variance/(freq - 1));
      //FIXME: change this to utilise sigma
      if (!(sigma <= (low->end - low->start)))
      {
        sigma = low->end - low->start;
      }
    }
    // both elements in one bin
    else
    {
      low->variance = (low->variance > low->next->variance) ? \
                      low->variance : low->next->variance;
      //assert (low->variance >= 0);
      TRACE ( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, low->variance));
    }
  }
  else
  {
    low->variance   = combined_sigma (low->variance, low->frequency, \
        low->next->variance, low->next->frequency);
    //assert (low->variance >= 0);
   TRACE ( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, low->variance));
  }

  /* update min and max */
  max = (low->next->max > low->max) ? low->next->max : low->max;
  min = (low->next->min < low->min) ? low->next->min : low->min;
  if (min == 0) {
    low->min = (low->min == 0) ? low->next->min : low->min;
  } else {
    low->min = min;
  }
  low->max = max;
  low->frequency += low->next->frequency;

  /* Merge the bins' participant lists. */
  low->participants->merge(*low->next->participants);
  
  /* free up one bin which was merged */
  histogram_bin* tmp_bin = low->next;
  low->next = low->next->next;
  delete tmp_bin;

  /* Split the highest frequency bin into two bins */
  tmp_bin = high->next;
  high->next = new histogram_bin();

  high->next->participants.reset(high->participants->clone());

  high->next->next = tmp_bin;
  high->next->end = high->end;

  // set the midpoint of the new bin.
  high->end = high->next->start = high->average;

  // this if check is superfluous considering we return if this condition is
  // false.
  if (high->frequency <= 1) {
    // this should never occur as we don't split bins with 1 or less items
    cerr << "SPLITTING BIN WITH 1 ITEM" << endl;
    exit(1);
  }

  //assert (high->variance >= 0);
  TRACE ( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, high->variance));
  sigma = sqrt (high->variance/(high->frequency - 1));
  //FIXME: change this to utilise sigma
  if (!(sigma <= (high->end - high->start)))
    sigma = high->end - high->start;
  //assert (sigma >= 0);

  high->average -= sigma/2;
  if (high->average < 0)
    high->average = high->start;
  //assert (high->average >= 0);

  high->next->average = high->average + sigma;

  if (high->next->average < 0)
    high->next->average = high->next->start;
  //assert (high->next->average >= 0);

  /*  maintain properties - if sigma value is too high, average calculated
   *  above can fall outside the range. these checks ensure that the average
   *  always falls in the range.
   */
  if (high->average > high->end)
    high->average = high->end;

  if (high->average < high->start)
    high->average = high->start;

  if (high->next->average > high->next->end)
    high->next->average = high->next->end;

  if (high->next->average < high->next->start)
    high->next->average = high->next->start;

  high->next->frequency = high->frequency >> 1;
  high->frequency -= high->next->frequency;

  sigma = sigma / 2;

  high->variance = (high->frequency - 1)*sigma*sigma;
  //assert (high->variance >= 0);
  TRACE ( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, high->variance));

  high->next->variance = (high->next->frequency - 1)*sigma*sigma;
  //assert (high->next->variance >= 0);
  TRACE ( ("[%s:%d] variance=%.2f\n", __FILE__, __LINE__, high->next->variance));
    
  /* assign min and max */
  high->next->max = high->max;
  high->next->min = high->next->start;
  high->max = high->end;
  //high->min remains unchanged


  if (high->start > high->average || high->average > high->end) {
    TRACE( ("high:f = %d, start = %.2f, end = %.2f, avg = %.2f, sigma = %f\n",\
        high->frequency, high->start, high->end, high->average, sigma));
  }
  //assert (high->start <= high->average);
  //assert (high->average <= high->end);

  if (high->next->start > high->next->average) {
    TRACE( ("high->next:f = %d start = %.2f, end = %.2f, avg = %.2f\n",\
          high->next->frequency, high->next->start, high->next->end, \
          high->next->average));
  }

  if (!(high->next->average <= high->next->end)) {
    TRACE( ("high->next:f = %d start = %.2f, end = %.2f, avg = %.2f\n",\
        high->next->frequency, high->next->start, high->next->end,\
        high->next->average));
  }

  //assert (high->next->start <= high->next->average);
  //assert (high->next->average <= high->next->end);
  
  //TODO: ok to create zero-width bins?
  //assert (high->start <= high->end);
}



void Histogram::add(double value, int rank) {  
  if (rank < 0) rank = creator_rank;

  // init bin ranges on first add
  if (num_elems == 0) {
    double bin_size = (value != 0.0) ? (2.0 * value)/num_bins : 1.0;

    histogram_bin *bin = bins;
    for (int i=0; i < num_bins; i++) {
      //assert(bin);
      bin->start = i * bin_size;
      bin->end = (i + 1) * bin_size;
      bin = bin->next;
    }    
  }

  // maintain min/max values.
  update_min(value, rank);
  update_max(value, rank);
  
  // put the valuein the histogram
  histogram_bin *dest = insert(value);

  // if we need to, update participants for the bin this was added to
  if (!dest) dest = find_bin_for_value(value);
  dest->addParticipant(rank);
}



histogram_bin *Histogram::find_bin_for_value(double value) {
  histogram_bin *bin = bins;
  int bin_idx = 0;
  while (bin) {
    bin_idx++;
    if ((value >= bin->start) && (value <= bin->end)) {
      break;
    }

    if (bin_idx == num_bins) {
      bin->end = value;
      break;
    }
    bin = bin->next;
  }

  if (!bin) {
    // no suitable bin was found: search is broken.
    TRACE ( ( "ERROR: no bin found for value %d\n", value));
    exit (1);
  }

  return bin;
}


/*
 * Add a value to a histogram. Depending upon the attributes, the histogram is
 * also flattened periodically.
 */
histogram_bin *Histogram::insert(double value) {
  assert (magic == HISTO_MAGIC);
  TRACE ( ("value=%f\n", value));

  histogram_bin *insert_bin = find_bin_for_value(value);
  insert_bin->add(value, 1);
  num_elems++;

  if (num_elems % smooth_threshold == 0) {
    smooth();
    return NULL;
  } else {
    return insert_bin;
  }
}


/* merge a bin into a histogram */
void Histogram::merge(histogram_bin* bin) {
  if (bin->frequency == 0) return;

  assert (magic == HISTO_MAGIC);
  //assert (bin->variance >= 0);

  double base = 0.0, value = 0.0;//, step = 0.0;  
  if (bin->frequency == 1) {
    base = bin->average;

  } else {
    //FIXME: use avg, sigma
    //base = bin->start;
    //step = (bin->end - bin->start)/bin->frequency;
    base = bin->average;
    //assert (base >= 0);
  }

  // TODO: this seems a little slow... maybe find a better way to add values?
  // TODO: note from todd -- maybe we can use skyline merge for this?
  for (long long i = 0; i < bin->frequency; i++) {
    // interpolate value
    value = base;
    //assert (value >= 0);

    // add the value to h1
    histogram_bin *dest = insert(value);

    if (!dest) dest = find_bin_for_value(value);

    // merge participant lists by oring them together
    dest->participants->merge(*bin->participants);
  }
}


/* merge a bin into a histogram */
void Histogram::skyline_merge(histogram_bin* bin) {
  if (bin->frequency == 0) return;

  assert (magic == HISTO_MAGIC);
  //assert (bin->variance >= 0);

  double base = 0.0, value = 0.0;//, step = 0.0;  
  if (bin->frequency == 1) {
    base = bin->average;

  } else {
    //FIXME: use avg, sigma
    //base = bin->start;
    //step = (bin->end - bin->start)/bin->frequency;
    base = bin->average;
    //assert (base >= 0);
  }
  num_elems += bin->frequency;
  histogram_bin *insert_bin = find_bin_for_value(base);
  //for(int i=0; i < bin->frequency; i++)
    insert_bin->add(base, bin->frequency);
  if (num_elems % smooth_threshold == 0) 
    smooth();
}

 /*
 * Merge two histograms.
 */
void Histogram::merge (const Histogram *other)
{
  histogram_bin* bin;

  if (!other) return;
  assert (magic == HISTO_MAGIC);
  assert (other->magic == HISTO_MAGIC);

  // maintain min/max values for entire histogram.
  update_min(other->min_value, other->min_rank);
  update_max(other->max_value, other->max_rank);

  bin = other->bins;
  while (bin) {
    merge(bin);
    bin = bin->next;
  }
}


/*
 * Merge two histograms.
 */
void Histogram::skyline_merge (const Histogram *other)
{
  histogram_bin* bin;

  if (!other) return;
  assert (magic == HISTO_MAGIC);
  assert (other->magic == HISTO_MAGIC);
  /*char * buf= (char *) malloc(20000);
  output(buf);
  printf("Hist 1: %s\n", buf);
  other->output(buf);
  printf("Hist 2: %s\n", buf);
  free(buf);
  */
  // maintain min/max values for entire histogram.
  update_min(other->min_value, other->min_rank);
  update_max(other->max_value, other->max_rank);

  bin = other->bins;
  while (bin) {
    skyline_merge(bin);
    bin = bin->next;
  }
  smooth();
}

/*
 * Return one value from a histogram. The value is determined by randomly
 * selecting a value in the range 0 - num_elems.
 */
double Histogram::value (int k, int j)
{
  int bin_id = 0, n, i;
  double value = 0, base = 0, step = 0;

  static int seed_generated = 0;
  static int prev_seed = 0;
  
  assert (magic == HISTO_MAGIC);

  histogram_bin* tmp = bins;

  if (!seed_generated || (prev_seed != k))
  {
    prev_seed = k;
    srand (k);
    seed_generated = 1;
  }
  /* get a random number in the range of (0, num_elems). It does not make sense
   * to use rank because there might not be as many elements in the histo
   * use num_elems - 2 (i.e. avoid min, max)
   */
  if (j == -1)
    n = i = (int)rand_max (num_elems - 2); 
  else
    n = i = j;
  while (tmp)
  {
    if (tmp->frequency && n <= tmp->frequency)
      break;
    n -= tmp->frequency;

    bin_id++;
    if (bin_id == num_bins)
    {
      break;
    }
    tmp = tmp->next;
  }

  assert (tmp);
 // assert (tmp->frequency > 0);
  if(tmp->frequency <= 0)
    return min_value;

  if (1) //tmp->frequency == 1)
  //if (tmp->frequency == 1)
  {
    base = tmp->average;
  }
  else
  {
    //FIXME: use avg, sigma
    base = tmp->min;
    step = (tmp->max - tmp->min)/tmp->frequency;
    assert (base >= 0);
  }

  value = base + step * n;
  //FIXME
  //min = max = 25, avg = 25.15 <-- this is a workaround this bug
  //assert ((value == tmp->average) || (value >= 0 && !(value > tmp->max)));
  if (!((value == tmp->average) || (value >= 0 && !(value > tmp->max))))
  {
    fprintf (stderr, "v %.2f start %.2f, min %.2f, avg %.2f, max %.2f, variance %.2f, f = %d, stddev %.2f, n %d, k %d\n",
                     value , tmp->start, tmp->min, tmp->average, tmp->max, tmp->end, tmp->frequency, tmp->variance, n, k);
    //exit (0);
  }
  if(value < 0) {
    value = min_value;
  }
  return value;
}

double Histogram::value_from_vector(RandomVector *rv) {

  int bin_id = 0, n, i;
  double value = 0, base = 0, step = 0;

  assert (magic == HISTO_MAGIC);

  histogram_bin* tmp = bins;

  /* get a random number in the range of (0, num_elems). It does not make sense
   * to use rank because there might not be as many elements in the histo
   * use num_elems - 2 (i.e. avoid min, max)
   */
    n = rv->get_rand_max (num_elems - 2); 
  while (tmp)
  {
    if (tmp->frequency && n <= tmp->frequency)
      break;
    n -= tmp->frequency;

    bin_id++;
    if (bin_id == num_bins)
    {
      break;
    }
    tmp = tmp->next;
  }

  assert (tmp);
 // assert (tmp->frequency > 0);
  if(tmp->frequency <= 0)
    return min_value;

  if (1) //tmp->frequency == 1)
  //if (tmp->frequency == 1)
  {
    base = tmp->average;
  }
  else
  {
    //FIXME: use avg, sigma
    base = tmp->min;
    step = (tmp->max - tmp->min)/tmp->frequency;
    assert (base >= 0);
  }

  value = base + step * n;
  //FIXME
  //min = max = 25, avg = 25.15 <-- this is a workaround this bug
  //assert ((value == tmp->average) || (value >= 0 && !(value > tmp->max)));
  if (!((value == tmp->average) || (value >= 0 && !(value > tmp->max))))
  {
    fprintf (stderr, "v %.2f start %.2f, min %.2f, avg %.2f, max %.2f, variance %.2f, f = %d, stddev %.2f, n %d\n",
                     value , tmp->start, tmp->min, tmp->average, tmp->max, tmp->end, tmp->frequency, tmp->variance, n);
    //exit (0);
  }
  if(value < 0) {
    value = min_value;
  }
  return value;

}

double Histogram::value_for_rank(int rank) {
  assert (magic == HISTO_MAGIC);

  // make sure we seeded the RN generator.
  // If not, just use time of day.
  static int seeded = FALSE;
  if (!seeded) {
    struct timeval time;
    gettimeofday(&time, NULL);
    srand(time.tv_sec * 1000000 + time.tv_usec);
    seeded = TRUE;
  }
  
  // Construct a list of bins for which the rank was a participant
  histogram_bin *rank_bins[num_bins];
  int num_rank_bins = 0;
  double num_rank_elems = 0;
  for (histogram_bin *bin = bins; bin; bin=bin->next) {
    if (bin->participants->member(rank)) {
      rank_bins[num_rank_bins++] = bin;

      // add the frequency divided by the number of participants
      // this will normalize for multi-participant bins
      num_rank_elems += bin->singleRankFrequency();
    }
  }

  // It's an error if the rank asked for wasn't in the participant lists.
  // To avoid this, check with isParticipant() before calling.
  assert(num_rank_bins != 0);

  // Generate a random number to pick among the bins that the
  // target rank participated in.
  double n = ((double)rand() / (RAND_MAX + 1.0)) * num_rank_elems;
  histogram_bin *expected = NULL;
  for (int i=0; i < num_rank_bins; i++) {
    histogram_bin *bin = rank_bins[i];
    if (n <= bin->singleRankFrequency()) {
      expected = bin;
      break;
    }
    n -= bin->singleRankFrequency();
  }

  assert (expected);
  assert (expected->frequency > 0);

  double base = 0.0;
  double step = 0.0;
  if (expected->frequency == 1) {
    base = expected->average;

  } else {
    //FIXME: use avg, sigma
    base = expected->min;
    step = (expected->max - expected->min)/expected->frequency;
  }

  double value = expected->average;
  //double value = base + step * n;

  if (!((value == expected->average) || (value >= 0 && !(value > expected->max)))) {
    fprintf (stderr, 
	     "UNEXPECTED: v %.2f start %.2f, end %.2f, min %.2f, avg %.2f, max %.2f, variance %.2f, f = %d, stddev %.2f, "
	     "base %.2f, step %.2f,n %lf\n",
	     value , expected->start, expected->end, expected->min, expected->average, expected->max, 
	     expected->end, expected->frequency, expected->variance, base, step, n);
    return 0.0;
    //exit (0);
  }

  if (value > 1e25) {
    fprintf (stderr, 
	     "TOO BIG: v %.2f start %.2f, end %.2f, min %.2f, avg %.2f, max %.2f, variance %.2f, f = %d, stddev %.2f, "
	     "base %.2f, step %.2f,n %lf\n",
	     value , expected->start, expected->end, expected->min, expected->average, expected->max, 
	     expected->end, expected->frequency, expected->variance, base, step, n);
    return -5;
  }

  return value;
}


int Histogram::min_max (double * min, double * max, int * min_node, int *
    max_node)
{
  if (!min || !max) return -1;

  // TODO: is it safe to just replace this with accessors for min_value and max_value?
  // Todd: should be safe.
  assert (magic == HISTO_MAGIC);
  
  histogram_bin* bin = bins;
  *min = 0;
  *max = 0;
  while (bin)
  {
    if (bin->frequency > 0)
    {
      if (*min == 0)
        *min = bin->min;
      *max = bin->max;
    }
    bin = bin->next;
  }
  
  *min_node = min_rank;
  *max_node = max_rank;
 
  return 0;
}


double Histogram::average()
{
  double sum = 0;

  if (num_elems == 0)
    return 0;

  histogram_bin* bin = bins;
  while (bin)
  {
    sum += bin->average*bin->frequency;
    bin = bin->next;
  }
  return (sum/num_elems);
}


double Histogram::min()
{
  if (num_elems == 0)
    return 0;

  histogram_bin* bin = bins;
  while (bin)
  {
    if (bin->frequency > 0)
      return bin->min;
    bin = bin->next;
  }
 
  return 0;
}


bool Histogram::isParticipant(int rank) {
  for (histogram_bin *bin = bins; bin; bin = bin->next) {
    //if ((*bin->participants)[rank]) {
    if (bin->participants->member(rank)) {
      return true;
    }
  }
  return false;
}


int Histogram::get_packed_size(MPI_Comm comm) {
  int size = 0;

  size += packed_size(1, MPI_INT, comm);   // magic
  size += packed_size(1, MPI_LONG_LONG, comm);   // num_elems
  size += packed_size(1, MPI_INT, comm);   // num_bins
  size += packed_size(1, MPI_INT, comm);   // smooth_threshold
  size += packed_size(1, MPI_INT, comm);   // merge_limit
  size += packed_size(1, MPI_INT, comm);   // merge_ratio

  size += packed_size(1, MPI_INT, comm);   // rank
  size += packed_size(1, MPI_INT, comm);   // max rank
  size += packed_size(1, MPI_INT, comm);   // min rank
  size += packed_size(1, MPI_DOUBLE, comm);   // max value
  size += packed_size(1, MPI_DOUBLE, comm);   // min value
  size += sig_packed_size(&signature, comm);
  for (histogram_bin *bin = bins; bin; bin = bin->next) {
    size += packed_size(1, MPI_UNSIGNED, comm);    // magic 
    
    size += packed_size(1, MPI_DOUBLE, comm);  //min
    size += packed_size(1, MPI_DOUBLE, comm);  //max
    size += packed_size(1, MPI_DOUBLE, comm);  //start
    size += packed_size(1, MPI_DOUBLE, comm);  //end
    size += packed_size(1, MPI_LONG_LONG, comm);     //frequency
    size += packed_size(1, MPI_DOUBLE, comm);  //average
    size += packed_size(1, MPI_DOUBLE, comm);  //variance
    
    // participant list
    size += bin->participants->packed_size(comm);
  }

  return size;
}


void Histogram::pack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  PMPI_Pack(&magic,            1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&num_elems,        1, MPI_LONG_LONG, buf, bufsize, position, comm);
  PMPI_Pack(&num_bins,         1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&smooth_threshold, 1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&merge_limit,      1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&merge_ratio,      1, MPI_INT, buf, bufsize, position, comm);

  PMPI_Pack(&creator_rank,     1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&max_rank,         1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&min_rank,         1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&max_value,        1, MPI_DOUBLE, buf, bufsize, position, comm);
  PMPI_Pack(&min_value,        1, MPI_DOUBLE, buf, bufsize, position, comm);

  sig_pack(&signature, buf, bufsize, position, comm);

  for (histogram_bin *bin = bins; bin; bin = bin->next) {
    PMPI_Pack(&bin->magic,   1, MPI_UNSIGNED, buf, bufsize, position, comm);

    PMPI_Pack(&bin->min,        1, MPI_DOUBLE, buf, bufsize, position, comm);
    PMPI_Pack(&bin->max,        1, MPI_DOUBLE, buf, bufsize, position, comm);
    PMPI_Pack(&bin->start,      1, MPI_DOUBLE, buf, bufsize, position, comm);
    PMPI_Pack(&bin->end,        1, MPI_DOUBLE, buf, bufsize, position, comm);
    PMPI_Pack(&bin->frequency,  1, MPI_LONG_LONG, buf, bufsize, position, comm);
    PMPI_Pack(&bin->average,    1, MPI_DOUBLE, buf, bufsize, position, comm);
    PMPI_Pack(&bin->variance,   1, MPI_DOUBLE, buf, bufsize, position, comm);
    
    bin->participants->pack(buf, bufsize, position, comm);
  }
}

Histogram* Histogram::unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  Histogram *h = new Histogram();

  PMPI_Unpack(buf, bufsize, position, &h->magic,            1, MPI_INT, comm);
  PMPI_Unpack(buf, bufsize, position, &h->num_elems,        1, MPI_LONG_LONG, comm);
  PMPI_Unpack(buf, bufsize, position, &h->num_bins,         1, MPI_INT, comm);
  PMPI_Unpack(buf, bufsize, position, &h->smooth_threshold, 1, MPI_INT, comm);
  PMPI_Unpack(buf, bufsize, position, &h->merge_limit,      1, MPI_INT, comm);
  PMPI_Unpack(buf, bufsize, position, &h->merge_ratio,      1, MPI_INT, comm);

  PMPI_Unpack(buf, bufsize, position, &h->creator_rank, 1, MPI_INT, comm);
  PMPI_Unpack(buf, bufsize, position, &h->max_rank,     1, MPI_INT, comm);
  PMPI_Unpack(buf, bufsize, position, &h->min_rank,     1, MPI_INT, comm);
  PMPI_Unpack(buf, bufsize, position, &h->max_value,    1, MPI_DOUBLE, comm);
  PMPI_Unpack(buf, bufsize, position, &h->min_value,    1, MPI_DOUBLE, comm);

  sig_unpack(&h->signature, buf, bufsize, position, comm);

  histogram_bin *tail = NULL;
  for (int i=0; i < h->num_bins; i++) {
    histogram_bin *bin = new histogram_bin();

    PMPI_Unpack(buf, bufsize, position, &bin->magic, 1, MPI_UNSIGNED, comm);

    PMPI_Unpack(buf, bufsize, position, &bin->min,        1, MPI_DOUBLE, comm);
    PMPI_Unpack(buf, bufsize, position, &bin->max,        1, MPI_DOUBLE, comm);
    PMPI_Unpack(buf, bufsize, position, &bin->start,      1, MPI_DOUBLE, comm);
    PMPI_Unpack(buf, bufsize, position, &bin->end,        1, MPI_DOUBLE, comm);
    PMPI_Unpack(buf, bufsize, position, &bin->frequency,  1, MPI_LONG_LONG,    comm);
    PMPI_Unpack(buf, bufsize, position, &bin->average,    1, MPI_DOUBLE, comm);
    PMPI_Unpack(buf, bufsize, position, &bin->variance,   1, MPI_DOUBLE, comm);

    bin->participants.reset(histo_rank_set_t::unpack(buf, bufsize, position, comm));

    if (tail == NULL) {
      h->bins = bin;
    } else {
      tail->next = bin;
    }
    tail = bin;
  }

  return h;
}

int Histogram::getCommRank() {
  return get_my_rank ();
}


// ---- C wrapper functions follow ----- //

void histogram_free(Histogram *histo) {
  delete histo;
}

int histogram_output(char *buf, Histogram *histo) {
  return histo->output(buf);
}

Histogram* histogram_input(char *buf) {
  return Histogram::input(buf);
}

int histogram_packed_size(Histogram *histo, MPI_Comm comm) {
  return histo->get_packed_size(comm);
}

void histogram_pack(Histogram *histo, void *buf, int bufsize, int *position, MPI_Comm comm) {
  histo->pack(buf, bufsize, position, comm);
}

Histogram* histogram_unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  return Histogram::unpack(buf, bufsize, position, comm);
}

// used during transfer to merge rsd queues.
// TODO: move this out of here when we compile everything w/the C++ compiler.
void histogram_merge(Histogram **m, long int mtime, Histogram **s, 
		     long int stime, int srank, int B) {
  Histogram *master = *m;
  Histogram *slave = *s;
  //printf("merge:master %p,mtime %ld, slave %p, stime %ld, slave:%d\n",master,mtime,slave,stime,srank);
  if (!master && !slave) {
    master = new Histogram(B);
    master->add(mtime);
    master->add(stime, srank);
    *m = master;

  } else if (master && slave) {
    master->merge(slave);

  } else if (master && !slave) {
    master->add(stime, srank);
    
  } else /* (!master && slave) */ {
    slave->set_rank(Histogram::getCommRank());
    *s = NULL;
    *m = slave;
  }
  //printf("merge end\n");
}


RankSet *Histogram::getParticipants() {
  auto_ptr<RankSet> result(bins->participants->clone());
  
  for (histogram_bin *bin = bins->next; bin; bin = bin->next) {
    result->merge(*bin->participants);
  }
  return result.release();
}


//TODO: move this out of here when we compile wtih the C++ compiler
void histogram_init(Histogram **histo, int B) {
  if (*histo) return;
  Histogram *h = new Histogram(B);
  *histo = h;
}

void histogram_add (Histogram *histo, int v)
{
  int rank = get_my_rank ();
  histo->add (v, rank);
}

int histogram_min_max (Histogram *histo, double * min, double * max, int *
    min_rank, int * max_rank) {
  return histo->min_max(min, max, min_rank, max_rank);
}

void histogram_average (Histogram *histo, double *avg) {
  *avg = histo->average();
}

int histogram_has_other (Histogram *histo, int rank)
{
  return histo->has_other(rank);
}

double histogram_value (Histogram *h, int k, int j)
{
  return h->value (k, j);
}

double histogram_value_from_vector (Histogram *h, RandomVector *rv)
{
  return h->value_from_vector (rv);
}

ostream& operator<<(ostream& out, const Histogram& histo) {
  char buf[32768]; // TODO: fix this class to use streamed output.
  histo.output(buf);
  out << buf;
  return out;
}

/* vim: set tw=80 ts=2 sw=2 expandtab: */
