/*
 * Filename: TimeList.C
 * Author:   ratn1
 * Created:  Wed Sep 19 11:15:22 2007
 * $Id: TimeList.C,v 1.3 2008/03/20 06:50:42 pratn Exp $
 */

#include <algorithm>
using namespace std;

#include "TimeList.h"
#include "mpi_util.h"

TimeList::TimeList () { }

TimeList::~TimeList() {
  for_each(tList.begin(), tList.end(), list_deleter());
}


int TimeList::output (char * buf, int index) const
{
  int i = 0;
  char * pos = buf;
  for (trc_iter iter = tList.begin(); iter != tList.end(); iter++, i++) {
    pos += sprintf (pos, "TIME%d-%d: ", index, i);
    pos += (*iter)->output (pos);
    pos += sprintf(pos, "\n");
  }
  return (pos - buf);
}


int TimeList::get_packed_size (MPI_Comm comm) const
{
  int size = packed_size(1, MPI_INT, comm);   // number of lists
  for (trc_iter iter = tList.begin(); iter != tList.end(); iter++)
  {
    size += (*iter)->get_packed_size(comm);
  }

  return size;
}

void TimeList::pack (void *buf, int bufsize, int *position, MPI_Comm comm) const
{
  int list_size = tList.size ();
  PMPI_Pack(&list_size, 1, MPI_INT, buf, bufsize, position, comm);
  for (trc_iter iter = tList.begin(); iter != tList.end(); iter++)
  {
    (*iter)->pack(buf, bufsize, position, comm);
  }

  return ;
}


TimeList * TimeList::unpack (void * buf, int bufsize, int *position, MPI_Comm comm)
{
  TimeList * var = new TimeList();
  int list_size = 0;

  PMPI_Unpack(buf, bufsize, position, &list_size, 1, MPI_INT, comm);
  for (int i = 0; i < list_size; i++)
  {
    TimeRecord * t = TimeRecord::unpack(buf, bufsize, position, comm);
    var->insert (t); 
  }

  return var;
}


TimeRecord * TimeList::begin ()
{
  return *(tList.begin());
}


void TimeList::insert (TimeRecord *t)
{
  tList.push_back (t);
}


TimeRecord * TimeList::find_match (const stack_sig *sig) const
{
  for (trc_iter miter = tList.begin(); miter != tList.end(); miter++)
  {
    if ((*miter)->sig_match(sig))
      return *miter;
  }

  return NULL;
}


void TimeList::merge ( 
  long int dest_time, const stack_sig *dest_sig, int dest_rank,
  TimeList * psrc,
  long int src_time, const stack_sig* src_sig, int src_rank,  int B
 )
{
  TRList * dest_list = &tList;
  TRList * src_list = NULL;
  if (psrc) 
  {
    src_list = &psrc->tList;
  }

  if (dest_list->empty())
  {
    TimeRecord *h = new TimeRecord (dest_time, B, dest_sig, dest_rank, dest_rank, dest_rank);
    dest_list->push_back (h);
  }
  if (!src_list)
  {
    TimeRecord *h = find_match(src_sig);
    if (!h)
    {
      // if this is intra-node merge, a new path was found
      // if inter-node , one histogram is part of a PRSD while the other isn't
      h = new TimeRecord (src_time, B, src_sig, src_rank, src_rank, src_rank);
      dest_list->push_back (h);
    }
    else
    {
      h->add ((double)src_time, src_rank);
    }
  }
  else
  {
    tr_iter src_iter, dest_iter, tmp_iter;

    for (src_iter = src_list->begin(); src_iter != src_list->end(); )
    {
      int merged = 0;
      for (dest_iter = dest_list->begin(); dest_iter != dest_list->end(); dest_iter++)
      {
        if ((*dest_iter)->sig_match (*src_iter))
        {
          (*dest_iter)->merge (*src_iter);
          merged = 1;
          src_iter++;
          break;
        }
      }
      if (!merged)
      {
        // src_iter points into the other list after the splice
        // save src_iter++
        tmp_iter = src_iter;
        tmp_iter++;
        dest_list->splice (dest_iter, *src_list, src_iter);
        src_iter = tmp_iter;
      }
    }
  }
}

/* -- C interface -- */

void timing_list_free (TimeList ** rec_list)
{
  delete *rec_list;
  *rec_list = NULL;
}

int timing_list_output (char * buf, TimeList * list, int index)
{
  return list->output (buf, index);
}

int timing_list_packed_size (TimeList * list, MPI_Comm comm)
{
  return list->get_packed_size (comm);
}

void timing_list_pack (TimeList * list, void *buf, int bufsize, 
		       int *position, MPI_Comm comm)
{
  list->pack (buf, bufsize, position, comm);
}

void timing_list_unpack (TimeList ** list, void * buf, int bufsize, int *position,
    MPI_Comm comm)
{
  *list = TimeList::unpack (buf, bufsize, position, comm);
}

TimeRecord * timing_find_match (TimeList * list, const stack_sig *sig)
{
  return list->find_match(sig);
}

TimeRecord * timing_list_begin (TimeList* list) 
{
  return list->begin();
}

int timing_list_size (TimeList * list)
{
  return list->size();
}

void timing_list_insert (TimeList ** list, TimeRecord * rec)
{
  assert (rec);

  TimeList *hlist = *list;
  if (!hlist) {
    hlist = new TimeList();
  }
  hlist->insert(rec);

  *list = hlist;
}

void timing_list_merge (TimeList ** pdest, 
    long int dest_time, const stack_sig *dest_sig, int dest_rank,
    TimeList ** psrc,
    long int src_time,  const stack_sig *src_sig,  int src_rank, int B
) {
  TimeList *destList = *pdest;
  TimeList *srcList = *psrc;
  if (!destList) {
    destList = new TimeList();
  }

  destList->merge ( dest_time, dest_sig, dest_rank, 
		    srcList, src_time, src_sig, src_rank, B);

  *pdest = destList;
}

/* vim: set tw=80 ts=2 sw=2 expandtab: */
