#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <assert.h>
#include <mpi.h>

#include "opstruct.h"
#include "rsd_queue.h"
#include "setjmp.h"
#include "trace.h"
#include "rank_recursive.h"
#include "stack_sig.h"
#include "ptr_handler.h"
#include "trace.h"
#include "mpi_util.h"

#ifdef FEATURE_TIMING_HISTO
#include "Histogram.h"
#include "TimeList.h"
#elif defined (FEATURE_TIMING_BITS)
#include "Intseq.h"
#include "TimeList.h"
#endif

#include "VParam.h"


#define FORMAT_OUT(x) (log2pow2((field_mask_t)x))
#define TWO_COMPLEMENT(x) (FORMAT_OUT(x)^0xff)+1

#define err_dump stderr

void init_rsd_node(rsd_node *node)
{
  node->node = get_my_rank();
  node->rlist = NULL;
  sig_init(&node->signature);
  node->rsd_info = SINGLE_RSD;
  init_prsd_stack(&node->prsd_info);
  node->match = MATCH;
  node->next = NULL;
  node->prev = NULL;

  /*Added by Karthik - Set event to NULL*/  
  node->event = NULL;
  node->hist_param = NULL;
  node->hist_op = 0;

  node->graph.index = -1;
  node->graph.in_edges = NULL;
  node->graph.out_edges = NULL;
  node->graph.rank_list = NULL;
}

void free_rsd_node(rsd_node *node) {

  //printf("Inside free node: Event value is %p\n", node->event);
  if(!node) return;

  //TODO: Possible memory leak: Graph structures are not freed

  /* free participants */
  if(node->rlist) {
    free(node->rlist);
  }
  node->rlist = NULL;
  
  /* free string signature */
  sig_destroy(&node->signature);

  /*Added by Karthik - Free event*/
  if(node->event)
    free_event(&node->event);
  if(node->hist_param){
    histogram_free(node->hist_param);
    node->hist_param = NULL;
  }
  /* free prsd info */
  if(node->prsd_info.head)
    free_prsd_stack(&node->prsd_info);
  node->prsd_info.head = NULL;
  node->prsd_info.tail = NULL;

  /* free array parameters */
  if(node->data.array1) {
    free(node->data.array1);
  }
  if(node->data.array2) {
    free(node->data.array2);
  }
  if(node->data.array3) {
    free(node->data.array3);
  }
  if(node->data.array4) {
    free(node->data.array4);
  }

  if (node->data.rlparam) {
    rlparam_free ((RLParamList*)&node->data.rlparam);
  }

#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
    /* free histogram data */
    for (int i = 0; i < TIMING_MAX; i++)
    {
      if (node->data.time_info[i].time_list)
        timing_list_free (&(node->data.time_info[i].time_list));
    }
#endif
/*Added by Karthik - Start*/
#ifdef FEATURE_FILE_IO
  if(node->data.filepath!=NULL) { //&& strlen(node->data.filepath)>0) {
    free(node->data.filepath);
  }
  if(node->data.filename!=NULL) { // && strlen(node->data.filename)>0) {
    free(node->data.filename);
  }
  if(node->data.datarep != NULL) {
    free(node->data.datarep);
  }
#endif
/*Added by Karthik - End*/
  /* avoid dangling pointers */
  if(node->prev)
    node->prev->next = NULL;
  
  free(node);
}

void free_rsd_nodes(rsd_queue *queue) {
  rsd_node *node;
  while ((node = rsd_pop(queue))) {
    free_rsd_node(node);
  }
}


// FIXME: need to refactor this so that buffer overflows can be avoided.
static char buf[PBUF];

void output_rsd_node_event(FILE *fp, rsd_node *node) {
  field_mask_t arr;    /* For determining which arrays were used. */
  char *pos = buf;     /* Position in buffer for printing. */

  buf [PBUF - 1] = '!';

  /* rank list */
  pos += sprintf(pos, "ranks: ");
  if(node->rlist)
    pos += soutput_range_list(node->rlist, pos);

  /* rsd info */
  pos += sprintf(pos, "\nTERM: %d %d %d " , node->data.op, node->data.seq_num, node->rsd_info);
  pos += sig_soutput(pos, &node->signature);
  pos += sprintf(pos, "\n");

  if(node->event) {
    pos += soutput_event_timing(pos, node->event);
  }

  /* prsd info */
  if(node->prsd_info.head)
    pos += soutput_prsd_stack(pos, PBUF, &node->prsd_info);

  /*Print Event data*/
  if(node->event)
    pos += soutput_event_params(pos, node->event);
  pos += sprintf(pos,"\n");
  fprintf(fp, "%s", buf);

  assert (buf [PBUF - 1] == '!');

}

void output_rsd_node(FILE * fp, rsd_node *node)
{
  field_mask_t arr;    /* For determining which arrays were used. */
  char *pos = buf;     /* Position in buffer for printing. */

  buf [PBUF - 1] = '!';

  /* rank list */
  pos += sprintf(pos, "ranks: ");
  if(node->rlist)
    pos += soutput_range_list(node->rlist, pos);

  /* rsd info */
  pos += sprintf(pos, "\nTERM: %d %d %d " , node->data.op, node->data.seq_num, node->rsd_info);
  pos += sig_soutput(pos, &node->signature);
  pos += sprintf(pos, "\n");

#if defined(FEATURE_TIMING_HISTO) || defined(FEATURE_TIMING_BITS)
  for (int i = 0; i < TIMING_MAX; i++)
  {
    if (node->data.time_info[i].time_list)
      pos += timing_list_output (pos, node->data.time_info[i].time_list, i);
    else
      pos += sprintf (pos, "TIME%d-0: %ld\n", i, node->data.time_info[i].time);
  }
#endif

  /* prsd info */
  if(node->prsd_info.head)
    pos += soutput_prsd_stack(pos, PBUF, &node->prsd_info);

  /* param info */
  if(node->data.fields & GLOBAL_WORLD) 
    pos += sprintf(pos, "%hhx %"PRImpiPTR"\n", FORMAT_OUT(GLOBAL_WORLD),
                   ((mpi_hval_t)node->data.global_world));
  if(node->data.fields & NULL_REQ) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(NULL_REQ), node->data.null_req);
  if(node->data.fields & COMM) 
    pos += sprintf(pos, "%hhx %"PRImpiPTR"\n", FORMAT_OUT(COMM), 
                   ((mpi_hval_t)node->data.comm));
  if(node->data.fields & COUNT) {
    if (node->data.pvectors & COUNT) {
      pos += rlparam_output (node->data.rlparam, COUNT, pos);
    }
    else if (node->data.count_vector) {
      pos += sprintf(pos, "%hhx ", FORMAT_OUT(COUNT));
      pos += vector_output  (pos, node->data.count_vector);
    } else {
      pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(COUNT), node->data.count);
    }
  }
  if(node->data.fields & DATATYPE) 
    pos += sprintf(pos, "%hhx %"PRImpiPTR"\n", FORMAT_OUT(DATATYPE),
                   ((mpi_hval_t)node->data.datatype));
  if(node->data.fields & RECVCOUNT) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(RECVCOUNT), node->data.recvcount);
  if(node->data.fields & RECVTYPE) 
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(RECVTYPE), (mpi_hval_t)node->data.recvtype);
  if(node->data.fields & MPI_OP) 
    pos += sprintf(pos, "%hhx %"PRImpiPTR"\n", FORMAT_OUT(MPI_OP), 
                   (intptr_t)node->data.mpi_op);
  if(node->data.fields & ROOT) {
    if (node->data.pvectors & ROOT) {
      pos += rlparam_output (node->data.rlparam, ROOT, pos);
    }
    else if (node->data.root_vector) {
      pos += sprintf(pos, "%hhx ", FORMAT_OUT(ROOT));
      pos += vector_output  (pos, node->data.root_vector);
    } else {
      pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(ROOT), node->data.root);
    }
  }
  if(node->data.fields & TAG) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(TAG), node->data.tag);
  if(node->data.fields & DEST) {
    if (node->data.pvectors & DEST)
      pos += rlparam_output (node->data.rlparam, DEST, pos);
    else if (node->data.dest_vector) {
      pos += sprintf(pos, "%hhx ", FORMAT_OUT(DEST));
      pos += vector_output  (pos, node->data.dest_vector);
    } else
      pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(DEST), node->data.dest);
  }
  if(node->data.fields & REQUEST) {
    if (node->data.request_vector) {
      pos += sprintf(pos, "%hhx ", FORMAT_OUT(REQUEST));
      pos += vector_output  (pos, node->data.request_vector);
    } else
      pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(REQUEST), node->data.request);
  }
  if(node->data.fields & GROUP) 
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(GROUP), (mpi_hval_t)node->data.group);
  if(node->data.fields & COLOR) 
    pos += sprintf(pos, "%hhx %c %d\n", FORMAT_OUT(COLOR),
        show_rank_offset (node->data, COLOR), node->data.color);
  if(node->data.fields & KEY) 
    pos += sprintf(pos, "%hhx %c %d\n", FORMAT_OUT(KEY),
        show_rank_offset (node->data, KEY),   node->data.key);
  if(node->data.fields & GROUP1) 
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(GROUP1), (mpi_hval_t)node->data.group1);
  if(node->data.fields & GROUP2) 
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(GROUP2), (mpi_hval_t)node->data.group2);
  if(node->data.fields & LOCAL_LEADER) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(LOCAL_LEADER), node->data.local_leader);
  if(node->data.fields & PEER_COMM) 
    pos += sprintf(pos, "%hhx %"PRImpiPTR"\n", FORMAT_OUT(PEER_COMM), ((mpi_hval_t)node->data.peer_comm));
  if(node->data.fields & REMOTE_LEADER) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(REMOTE_LEADER), node->data.remote_leader);
  if(node->data.fields & HIGH) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(HIGH), node->data.high);
  if(node->data.fields & SOURCE) {
    if (node->data.pvectors & SOURCE)
      pos += rlparam_output (node->data.rlparam, SOURCE, pos);
    else if (node->data.source_vector) {
      pos += sprintf(pos, "%hhx ", FORMAT_OUT(SOURCE));
      pos += vector_output  (pos, node->data.source_vector);
    } else
      pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(SOURCE), node->data.source);
  }
  if(node->data.fields & RECVTAG) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(RECVTAG), node->data.recvtag);
  if(node->data.fields & NEWTYPE) 
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(NEWTYPE), (mpi_hval_t)node->data.newtype);
  if(node->data.fields & OLDTYPE) 
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(OLDTYPE), (mpi_hval_t)node->data.oldtype);
  if(node->data.fields & BLOCKLEN) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(BLOCKLEN), node->data.blocklen);
  if(node->data.fields & STRIDE) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(STRIDE), node->data.stride);
  if(node->data.fields & REORDER) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(REORDER), node->data.reorder);
  if(node->data.fields & NEWRANK) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(NEWRANK), node->data.newrank);
  if(node->data.fields & OUTCOUNT) 
    pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(OUTCOUNT), node->data.outcount);

  if (node->data.pvectors & RECVCOUNTS)
    pos += rlparam_output (node->data.rlparam, RECVCOUNTS, pos);
  else if ((arr = array_type1(node->data.arrays))) {
    pos += soutput_array(pos, arr, node->data.array_size1, node->data.array1);
  }

  if (node->data.pvectors & DISPLS)
    pos += rlparam_output (node->data.rlparam, DISPLS, pos);
  else if ((arr = array_type2(node->data.arrays))) {
    pos += soutput_array(pos, arr, node->data.array_size2, node->data.array2);
  }

  if (node->data.pvectors & COUNTS)
    pos += rlparam_output (node->data.rlparam, COUNTS, pos);
  else if ((arr = array_type3(node->data.arrays))) {
    pos += soutput_array(pos, arr, node->data.array_size3, node->data.array3);
  }

  if (node->data.pvectors & RDISPLS) {
    pos += rlparam_output (node->data.rlparam, RDISPLS, pos);
  }
  else if ((arr = array_type4(node->data.arrays))) {
    pos += soutput_array(pos, arr, node->data.array_size4, node->data.array4);
  }
 
  //Added by Karthik   
  #ifdef FEATURE_FILE_IO
  /*Recording any fields from fields_extn should be done by toggling bits.
   *This ensures that REPLAY populates correctly back into fields_extn.
   */
  if(node->data.fields_extn != NO_FIELDS)
    pos += sprintf(pos, "EXTN:\n");
  if(node->data.fields_extn & FILENAME) {
    if(node->data.filepath && node->data.filename && strcmp(node->data.filepath,node->data.filename)!=0)   
      pos += sprintf(pos,"%hhx %s%s\n", FORMAT_OUT(FILENAME), node->data.filepath, node->data.filename);
    else
      pos += sprintf(pos,"%hhx %s\n", FORMAT_OUT(FILENAME), node->data.filename);
  }
  if(node->data.fields_extn & AMODE)
    pos += sprintf(pos,"%hhx %d\n", FORMAT_OUT(AMODE),node->data.amode);
  if(node->data.fields_extn & FH)
    pos += sprintf(pos,"%hhx %d\n", FORMAT_OUT(FH),node->data.fh);
  if(node->data.fields_extn & OFFSET)
    pos += sprintf(pos,"%hhx (%d,%d,%d)\n", FORMAT_OUT(OFFSET),node->data.offsetinfo.startpos,node->data.offsetinfo.stride,node->data.offsetinfo.totalcount);
  if(node->data.fields_extn & COMM_SIZE)
    pos += sprintf(pos,"%hhx %d\n", FORMAT_OUT(COMM_SIZE),node->data.comm_size);
  if(node->data.fields_extn & RANK)
    pos += sprintf(pos,"%hhx %d\n", FORMAT_OUT(RANK),node->data.rank);
  if(node->data.fields_extn & NDIMS)
    pos += sprintf(pos,"%hhx %d\n", FORMAT_OUT(NDIMS),node->data.count);
  if(node->data.fields_extn & ORDER)
    pos += sprintf(pos,"%hhx %d\n", FORMAT_OUT(ORDER),node->data.order);
  if(node->data.fields_extn & DISP) {
    if (node->data.pvectors & DISP)
      pos += rlparam_output (node->data.rlparam, DISP, pos);
    else if (node->data.disp_vector) {
      pos += sprintf(pos, "%hhx ", FORMAT_OUT(DISP));
      pos += vector_output  (pos, node->data.disp_vector);
    } else
      pos += sprintf(pos, "%hhx %d\n", FORMAT_OUT(DISP), node->data.disp);
  }
  if(node->data.fields_extn & DATAREP)
    pos += sprintf(pos,"%hhx %s\n", FORMAT_OUT(DATAREP), node->data.datarep);
  if(node->data.fields_extn & ETYPE)
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(ETYPE), (mpi_hval_t)node->data.etype);
  if(node->data.fields_extn & FILETYPE)
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(FILETYPE), (mpi_hval_t)node->data.filetype);
  if(node->data.fields_extn & WHENCE)
    pos += sprintf(pos, "%hhx %" PRImpiPTR "\n", FORMAT_OUT(WHENCE), node->data.whence);

  // Below code checks whether the op uses each of the array fields
  // if so, it prints the appropriate field to output.
  if ((arr = array_type1(node->data.arrays_extn))) {
    pos += soutput_array(pos, arr, node->data.array_size1, node->data.array1);
  }

  if ((arr = array_type2(node->data.arrays_extn))) {
    pos += soutput_array(pos, arr, node->data.array_size2, node->data.array2);
  }

  if ((arr = array_type3(node->data.arrays_extn))) {
    pos += soutput_array(pos, arr, node->data.array_size3, node->data.array3);
  }

  if ((arr = array_type4(node->data.arrays_extn))) {
    pos += soutput_array(pos, arr, node->data.array_size4, node->data.array4);
  }
  #endif
  //Added by Karthik
  pos += sprintf(pos, "\n");
  fprintf(fp, "%s", buf);

  assert (buf [PBUF - 1] == '!');
}


void init_rsd_queue(rsd_queue *queue)
{
  queue->head = NULL;
  queue->tail = NULL;
  queue->last_op_added = -1;
}

void rsd_push(rsd_queue *queue, replay_op *operation, Event *event)
{
  rsd_node *new_tail = (rsd_node *) calloc(sizeof(rsd_node), 1);

  if(new_tail == NULL) {
    perror("malloc");
    exit(0);
  }

  init_rsd_node(new_tail);

  /*Added by Karthik - Copy Event into queue node - Start*/
  if(event)
    event_copy(&new_tail->event, &event);

  /*Karthik - Copy end*/

  create_recursive_list(get_my_rank(), &new_tail->rlist);

  sig_unwind(&new_tail->signature);

  memcpy(&new_tail->data, operation, sizeof(replay_op));

  new_tail->next = NULL;
  new_tail->prev = queue->tail;

  if(queue->head == NULL) {
    queue->head = new_tail;
    queue->tail = new_tail;
  }
  else {
    queue->tail->next = new_tail;
    queue->tail = new_tail;
  }

  /* add dependence edge */
  if (new_tail -> prev) {
    new_tail->graph.out_edges = list_insert (new_tail->graph.out_edges, new_tail->prev);
    new_tail->prev->graph.in_edges = list_insert (new_tail->graph.in_edges, new_tail);
    create_recursive_list(new_tail->node, &new_tail->graph.rank_list);
    assert (new_tail->node == new_tail->prev->node);
  }

  queue->last_op_added = operation->op;
}

void rsd_push_head(rsd_queue *queue, rsd_node *new_head)
{
  new_head->next = queue->head;
  new_head->prev = NULL;

  if(queue->head == NULL) {
    queue->head = new_head;
    queue->tail = new_head;
  }
  else {
    queue->head->prev = new_head;
    queue->head = new_head;
  }

  queue->last_op_added = new_head->data.op;
}

void rsd_push_tail(rsd_queue *queue, rsd_node *new_tail)
{
  new_tail->next = NULL;
  new_tail->prev = queue->tail;

  if(queue->head == NULL) {
    queue->head = queue->tail = new_tail;
  }
  else {
    queue->tail->next = new_tail;
    queue->tail = new_tail;
  }

  queue->last_op_added = new_tail->data.op;
}

void rsd_insert_post(rsd_queue *queue, rsd_node *target, rsd_node *seq_start, rsd_node *seq_end)
{
  rsd_node *iter, *target_end;

  if(!queue || !target || !seq_start || !seq_end) {
    fprintf(stderr, "error: rsd_insert called with invalid parameter\n");
    fflush(stderr);
    exit(0);
  }

  /* ensure the sequence is continous */
  iter = seq_start;
  while(iter && iter != seq_end)
    iter = iter->next;
  if(!iter) {
    fprintf(stderr, "error: rsd_insert called with invalid sequence\n");
    fflush(stderr);
    exit(0);
  } 
  
  /* insert sequence after target */
  target_end = target->next;

  target->next = seq_start;
  seq_start->prev = target;

  /* special case = end of the queue */
  if(!target_end) {
    queue->tail = seq_end;
    seq_end->next = NULL;
  }
  else {
    //target_end->next = seq_end;
    //seq_end->prev = target_end;
    target_end->prev = seq_end;
    seq_end->next = target_end;
  }
}

void rsd_insert_pre(rsd_queue *queue, rsd_node *target, rsd_node *seq_start, rsd_node *seq_end)
{
  rsd_node *iter, *target_start;

  assert(queue && target && seq_start && seq_end);
  if(seq_end->data.seq_num == 87409) {
    printf("test\n");
  }
  /* ensure the sequence is continous */
  iter = seq_start;
  while(iter && iter != seq_end)
    iter = iter->next;
  if(!iter) {
    fprintf(stderr, "error: rsd_insert called with invalid sequence\n");
    fflush(stderr);
    exit(0);
  }

  /* insert sequence before target */
  target_start = target->prev;

  /* special case = start of the queue */
  if(!target_start) {
    queue->head = seq_start;
    seq_start->prev = NULL;
  }
  else {
    target_start->next = seq_start;
    seq_start->prev = target_start;
  }

  target->prev = seq_end;
  seq_end->next = target;
}

void unsplice(rsd_queue *queue, rsd_node *start, rsd_node *end) {
  assert(queue && start && end);

  /* Connect predecessors to successors */
  if (start == queue->head) {
    queue->head = end->next;
  } else {
    start->prev->next = end->next;
  }

  /* Connect successors to predecessors */
  if (end == queue->tail) {
    queue->tail = start->prev;
  } else {
    end->next->prev = start->prev;
  }

  /* Tie off ends of unspliced sequence. */
  start->prev = NULL;
  end->next = NULL;
}

rsd_node *rsd_pop(rsd_queue *queue)
{
  rsd_node *ptr = queue->head;

  if(!queue->head)
    return NULL;

  /* single element to pop */
  if(!queue->head->next) {
    queue->head = NULL;
    queue->tail = NULL;
  }
  else {
    queue->head = queue->head->next;
    queue->head->prev = NULL;
  }

  return ptr;
}


rsd_node *rsd_pop_tail(rsd_queue *queue)
{
  rsd_node *old_tail = queue->tail;

  if(!queue->tail)
    return NULL;

  /* single element to pop */
  if (queue->head == queue->tail) {
    queue->head = NULL;
    queue->tail = NULL;

  } else {
    queue->tail = queue->tail->prev;
    queue->tail->next = NULL;
  }

  return old_tail;
}


int rsd_get_length(rsd_queue *queue)
{
  int count = 0;
  rsd_node *ptr = queue->head;

  while(ptr != NULL) {
    count++;
    ptr = ptr->next;
  }

  return count;
}

int rsd_checkqueue(rsd_queue *queue, char *msg)
{
  int next_count, prev_count;
  rsd_node *ptr;

  if(!queue->head && !queue->tail)
    return 1;

  if(!queue->head && queue->tail) {
    fprintf(stderr, "%s error: rsd_checkqueue detected a NULL head on task %d\n", msg, get_my_rank());
    exit(0);
  }

  if(queue->head && !queue->tail) {
    fprintf(stderr, "%s error: rsd_checkqueue detected a NULL tail on task %d\n", msg, get_my_rank());
    exit(0);
  }

  ptr = queue->head;
  next_count = 1;
  while(ptr->next) {
    ptr = ptr->next;
    next_count++;

    if(next_count > LINKLIST_DEBUG_MAX) {
      fprintf(stderr, "%s error: LINKLIST_DEBUG_MAX exceeded on task %d\n", msg, get_my_rank());
      exit(0);
    }
  }

  if(ptr != queue->tail) {
    fprintf(stderr, "%s error: rsd_checkqueue detected that a node which is not the tail has a NULL next pointer on task %d\n", msg, get_my_rank());
    exit(0);
  }

  ptr = queue->tail;
  prev_count = 1;
  while(ptr->prev) {
    ptr = ptr->prev;
    prev_count++;

    if(prev_count > next_count)
      break;
  }

  if(ptr != queue->head) {
    fprintf(stderr, "%s error: rsd_checkqueue detected that a node which is not the head has a NULL prev pointer on task %d\n", msg, get_my_rank());
    exit(0);
  }

  if(next_count != prev_count) {
    fprintf(stderr, "%s error: rsd_checkqueue detected that the number of next counters traversed does not equal the number of previous counters traversed on task %d\n", msg, get_my_rank());
    exit(0);
  }

  return 1;
}

void output_rsd_queue(FILE *dest, rsd_queue *queue)
{
  rsd_node *ptr = queue->head;

  while(ptr != NULL) {
    output_rsd_node_event(dest, ptr);
    //output_rsd_node(dest, ptr);
    ptr = ptr->next;
  }
}


int prsd_match(prsd_stack *ptr1, prsd_stack *ptr2, int depth)
{
  int count = 0;
  prsd_node *n1 = ptr1->tail;
  prsd_node *n2 = ptr2->tail;

  for(count = 0; count < depth; count++) {
    if(!n1 || !n2)
      return 0;

 #ifdef FEATURE_COMPRESS_HISTO
    //Karthik - Collecting histogram data to add imprecision
    //if prsd info do not match
  if( n1->loop_mems != n2->loop_mems ||
       !SIG_MATCH(n1->tail_sig, n2->tail_sig)) {
      return 0;
  } else {
    if(n1->loop_iters != n2->loop_iters) {
      if(!n2->prsd_stat) {
        histogram_init (&n2->prsd_stat, 5);//HIST_BUCKETS);
        histogram_add (n2->prsd_stat, n2->loop_iters);
      }
      histogram_add(n2->prsd_stat, n1->loop_iters);
    }
  }
 #else
    if(n1->loop_iters != n2->loop_iters ||
       n1->loop_mems != n2->loop_mems ||
       !SIG_MATCH(n1->tail_sig, n2->tail_sig)) {
      return 0;
    }
    //Karthik - End changes
 #endif

    n1 = n1->prev;
    n2 = n2->prev;
  }
  
  return 1;
}

int dep_graph_packed_size (dep_graph_info * graph, MPI_Comm comm)
{
  int size = 0;

  size += list_packed_size (graph-> in_edges, comm);
  size += list_packed_size (graph->out_edges, comm);
  size += recursive_list_packed_size (graph->rank_list, comm);
  return size;
}

void dep_graph_pack (dep_graph_info * graph, void * buf, int bufsize, int * position, MPI_Comm comm)
{
  list_pack (graph-> in_edges, buf, bufsize, position, comm);
  list_pack (graph->out_edges, buf, bufsize, position, comm);
  recursive_list_pack (graph->rank_list, buf, bufsize, position, comm);
}

void dep_graph_unpack (dep_graph_info * graph, void * buf, int bufsize, int * position, MPI_Comm comm)
{
  if (graph->in_edges)
    graph-> in_edges = list_unpack            (buf, bufsize, position, comm);
  if (graph->out_edges)
    graph->out_edges = list_unpack            (buf, bufsize, position, comm);
  if (graph->rank_list)
    graph->rank_list = recursive_list_unpack  (buf, bufsize, position, comm);
}

void printrsdqueue(rsd_queue *queue)
{
  rsd_node *iter = queue->head;
  printf("rsd_queue: ");
  if(iter!=NULL)
  {
    while(iter && iter!=queue->tail)
    {
      printf("%d-%d-->",iter->data.op,iter->data.seq_num);
      if(iter->event) event_print(iter->event);
      iter = iter->next;
    }
    if(iter) {
      printf("%d-%d-->",iter->data.op, iter->data.seq_num);
      if(iter->event) event_print(iter->event);
    }
  }
  
}
