#include "rsd_reader.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "mpi_util.h"
#include "utils.h"
#include "opstruct.h"
#include "trace.h"
#include "ptr_handler.h"
#include "rank_recursive.h"

#include "VParam.h"
#include "cvector.h"
#include "RandomVector.h"

#ifdef FEATURE_TIMING_HISTO
#include "Histogram.h"
#include "TimeList.h"
#elif defined (FEATURE_TIMING_BITS)
#include "Intseq.h"
#include "TimeList.h"
#endif

#define SIZE          16384
#define BIGSIZE       32768


enum read_array_type{
  READ_ARR_ELT_T,
  READ_MPI_AINT,
  READ_INT64
};

#ifdef FEWER_MALLOCS
void update_op_buf_size (rsd_node*);
#else
#define update_op_buf_size(a)
#endif

rsd_queue *read_rsd_file(const char *filename) {
  return read_rsd_file_rank(filename, -1);
}


unsigned char by_ranks = 1;
unsigned char parse_by_ranks ()
{
  return (by_ranks);
}

  int lineno = 0;
//FIXME: find a better way to do this
static char prsd_util_buf[BIGSIZE];

static int single_mem_prsd_count;

static int prsd_count;

extern RandomVector *timing_rvector;
extern RandomVector *param_rvector;

rsd_queue *read_rsd_file_rank(const char *filename, int rank)
{
  rsd_queue *retval = 0;
  FILE *trace;
  char *res;
  int get_content = 0;
  int recs = 0;
  list_t tail_list = 0;
  int is_field_extn = 0;

  single_mem_prsd_count = 1;
  prsd_count = 1;

  TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));

  by_ranks = (rank < 0) ? 0 : 1;

  trace = fopen(filename, "r");
  if(!trace) {
    perror("fopen");
    exit(0);
  }

  retval = (rsd_queue *) calloc(sizeof(rsd_queue), 1);
  init_rsd_queue(retval);

  while(!feof(trace)) {
    res = fgets(prsd_util_buf, SIZE, trace);
    if(!res && !feof(trace)) {
      perror("fgets");
      exit(0);
    }
    /*if(rank ==0) {
      printrsdqueue(retval);
    }*/
    if(res) {
      if(strlen(res) == SIZE) {
        fprintf(stderr, "fgets: not enough prsd_util_buffer space allocated for line %d\n", lineno);
        exit(0);
      }
    }

    /* note header lines contain 'ranks' whereas content lines do not */
    if(strstr(prsd_util_buf, "ranks")) {
      is_field_extn = 0;
      rsd_node *new_data = (rsd_node *) calloc(sizeof(rsd_node), 1);
      if(!new_data) {
        perror("malloc");
        exit(0);
      }

      init_rsd_node(new_data);
      init_op(&new_data->data);
      parse_rank(prsd_util_buf, new_data);

      if(member_rank(my_rank, new_data->rlist))
        new_data->match = 1; //Reusing variable to identify if a node participates in event or not

      if( !parse_by_ranks() || member_rank(rank, new_data->rlist)) {
        if (retval->tail)
          update_op_buf_size (retval->tail);
        recs++;       
        /*if(rank == 0){
          printf("Output:\n");
	  output_rsd_queue(stdout, retval);
          printf("Output End\n");fflush(stdout);
        }*/
        rsd_push_tail(retval, new_data);
        get_content = 1;
      }
      else {
        get_content = 0;
        if (new_data)
          free_rsd_node(new_data);
      }
    }

    /* only parse these if the header is my rank */
    else if(get_content && strlen(prsd_util_buf) > 1) {
      if(strstr(prsd_util_buf, "TERM")) {
        parse_rsd_header(prsd_util_buf, retval->tail, &tail_list, recs);
      }

      /* timing info */
      else if(strstr(prsd_util_buf, "TIME0") || strstr(prsd_util_buf, "TIME1"))
      {
#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
        char * ch = 0;

        int i = prsd_util_buf[4] - '0';
        ch = prsd_util_buf + strlen("TIME0-0: ");
        if (strchr(ch, '(')) // indicates that there's a histogram
        {
#ifdef FEATURE_TIMING_HISTO
          void * rec = histogram_input(ch);
#else
          void * rec = intseq_input (ch);
#endif

          timing_list_insert (&retval->tail->data.time_info[i].time_list, rec);
        }
        else
        {
          sscanf (ch, "%ld", &retval->tail->data.time_info[i].time);

          retval->tail->data.time_info[i].time_list = 0;
        }
#endif
      }

      /* PRSD info for leaders */
      else if(strstr(prsd_util_buf, "PRSD"))
      {
        TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));
        parse_prsd_header(prsd_util_buf, retval->tail, &tail_list);
      }
#if 0
      /* PARAM info for terminal */
      else if(strstr(prsd_util_buf, "PARAMS"))
      {
        TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));
        parse_params_header(prsd_util_buf, retval);
      }
#endif
      else if(strstr(prsd_util_buf,"all"))
      {
        /*VT_AK: End of file*/	
        TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));
        break;
      }
      else if(strstr(prsd_util_buf,"EXTN"))
      {
        is_field_extn = 1;/*Set field_extn flag to true so that extended fields are parsed henceforth*/
      }
      /* normal params */
      else
      {
        if(!is_field_extn)
        {
          TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));
          parse_prsd_content(prsd_util_buf, retval);
        }
        else
        {
          parse_prsd_content_extn(prsd_util_buf, retval);
        }
      }
      TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));
    }
    TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));
    lineno++;
  }
  TRACE (("[debug]%s:%d\n", __FILE__, __LINE__));

  fclose(trace);

#ifdef FEWER_MALLOCS
  Recvrecvbuf = malloc (Recvrecvbuf_size);
  Sendsendbuf = malloc (Sendsendbuf_size);
#endif
  
  assert (!tail_list);
  //printrsdqueue(retval);
  return retval;
}

void parse_rank(char *buf, rsd_node *target)
{
#ifdef DSTAR
  char *start_pos = strchr(buf, ':') + 1;
#else
  char *start_pos = buf;
#endif

  init_rsd_node(target);

  /* scn in rlist */
  target->rlist = parse_recursive_list(strchr(start_pos, ':') + 1);
}

void parse_rsd_header(char *buf, rsd_node *target, list_t *tail_list, int rec)
{
#ifdef DSTAR
  char *pos = strchr(buf, ':') + 1;
#else
  char *pos = buf;
#endif

  int op, seq, info, i;
  if (3 != sscanf(pos, "TERM: %d %d %d ", &op, &seq, &info)) {
    fprintf(stderr, "Error: malformed RSD record!\n");
    exit(1);
  }

  // skip to 3rd space + 1 where signature is.
  for (i=0; i < 4; i++) pos = strchr(pos, ' ') + 1;
  sig_input(&target->signature, pos);

  target->data.op = op;
  target->data.seq_num = rec;
  target->rsd_info = info;
 #ifndef REPLAY_HISTO
  if (parse_by_ranks ()) { //Karthik - Commented out to test reading file irrespective of rank
 #endif
      patch_prsd_tail (tail_list, target);
 #ifndef REPLAY_HISTO
  }
 #endif
}

void parse_prsd_header(char *buf, rsd_node *target, list_t *tail_list)
{
  int iters, mems, i;
  char *pos = NULL;

#ifdef DSTAR
  pos = strchr(buf, ':') + 1;
#else
  pos = buf;
#endif

  if (2 != sscanf(pos, "PRSD: %d %d", &iters, &mems)) {
    fprintf(stderr, "Error: malformed PRSD record!\n");
    exit(1);
  }

  // skip to signature
  for (i=0; i < 3; i++) pos = strchr(pos, ' ') + 1;

  /* we see them in order - so we don't want to do a stack push */
  prsd_push_end(&target->prsd_info, mems);

  sig_input(&target->prsd_info.tail->tail_sig, pos);
  target->prsd_info.tail->loop_iters = iters;
 #ifdef REPLAY_HISTO
  if(strchr(pos, 'N')) {
    char *ch = strchr(pos, '(');
    Histogram *h = histogram_input(ch);
    double min, max, avg;
    int min_node, max_node;
    histogram_min_max(h, &min, &max, &min_node, &max_node);
    histogram_average (h, &avg);
    int hist_val = (int)histogram_value_from_vector(h,param_rvector);//(int)histogram_value(h, 0, -1);
    if(my_rank == 0 )
      printf("PRSD: Iters: max %f, min %f, avg %f, int avg %d, hist %d\n", max, min, avg, (int)avg, hist_val);
    target->prsd_info.tail->loop_iters = hist_val;//histogram_value(h, 0, -1);
  }
 #endif
  target->prsd_info.tail->parent_ptr = target;
  prsd_count *= (target->prsd_info.tail->loop_iters + 1); 
 #ifndef REPLAY_HISTO
  if (parse_by_ranks ()) {
 #endif
    /* prsd has only one op */
    if (SIG_MATCH(target->signature, target->prsd_info.tail->tail_sig)) {
      target->prsd_info.tail->tail_ptr = target;
    } else {
      *tail_list = list_insert (*tail_list, target->prsd_info.tail);
    }
 #ifndef REPLAY_HISTO
  }
 #endif
}

void patch_prsd_tail (list_t * tail_list, rsd_node *target)
{
  prsd_node *tail = 0;
  list_t tmp_list = 0, new_list = 0;
  int seq_num;

  tmp_list = *tail_list;
  new_list = *tail_list;
  while (tmp_list)
  {
    tail = (prsd_node *) tmp_list->v;
    seq_num = tail->parent_ptr->data.seq_num + tail->loop_mems - 1;
    if (seq_num == target->data.seq_num && SIG_MATCH(target->signature, tail->tail_sig)) {
      if(prsd_count > 1)
        prsd_count = prsd_count/(tail->loop_iters + 1);
      tail->tail_ptr = target;
      tmp_list = tmp_list->next;
      *tail_list = new_list = list_delete (*tail_list, tail, NULL);
    }
    else
    {
      tmp_list = tmp_list->next;
    }
  }
}

#if 0
void parse_params_header(char *buf, rsd_queue *target)
{
  int off, len, i;
  char *start_pos;
  params_queue *pq_ptr;

  pq_ptr = (params_queue *) malloc(sizeof(params_queue));
  if(!pq_ptr) {
    perror("malloc");
    exit(0);
  }
  init_params_queue(pq_ptr);

  sscanf(buf, "PARAMS: %d", &len);
  start_pos = strchr(buf, ' ') + 1;
  for(i = 0; i < len + 1; i++) {
    /* first # is length */
    if(i > 0) {
      sscanf(start_pos, "%d", &off);
      start_pos = strchr(start_pos, ' ') + 1;
      params_push(pq_ptr, off);
    }
    else
      start_pos = strchr(start_pos, ' ') + 1;
  }
}
#endif

void read_array_entry_main (char * start_pos, enum read_array_type a, void **array, int *arrsize)
{
  int len, i=0;
  array_elt_t *  elt_arr = NULL;
  MPI_Aint    * aint_arr = NULL;
  int64_t     *  int_arr = NULL;

  sscanf(start_pos, "%d", &len);
  *arrsize = len;
  start_pos = strchr (start_pos, ' ') + 1;
  if (a == READ_ARR_ELT_T) {
    elt_arr = OP_ARRAY_ALLOC (len);
  } else if (a == READ_MPI_AINT) {
    aint_arr = OP_AINT_ARRAY_ALLOC (len);
  } else {
    int_arr = (int64_t*) malloc (sizeof (int64_t) * len);
  }

  if(!elt_arr && !aint_arr && !int_arr) {
    perror("malloc");
    exit(0);
  }
  for(i = 0; i < len; i++) {
    if (a == READ_ARR_ELT_T)
      sscanf(start_pos, "%"PRImpiARR, &(elt_arr[i]));
    else if (a == READ_MPI_AINT)
      sscanf(start_pos, "%"PRIdPTR, &(aint_arr[i]));
    else
      sscanf(start_pos, "%"PRId64, &(int_arr[i]));
    start_pos = strchr(start_pos, ' ') + 1;
  }
  if (a == READ_ARR_ELT_T)
    *array = elt_arr;
  else if (a == READ_MPI_AINT)
    *array = aint_arr;
  else
    *array = int_arr;
  return;
}

void rlparam_get_array (char * buf, char * rank_pos, enum read_array_type a, void ** array, int * arrsize, int rank)
{
  int found;

  int * rlist = parse_recursive_list (rank_pos+1);
  if (member_rank (rank, rlist)) 
    found = TRUE;
  else
    found = FALSE;
  free (rlist);

  if (!found) return ;

  char * start_pos = strchr(buf, ' ') + 1;
  read_array_entry_main (start_pos, a, array, arrsize);

  return;
}

void read_array_entry (char *buf, enum read_array_type a, void **array, int * arrsize)
{
  char * start_pos = NULL, * rank_pos = NULL;

  start_pos = strchr (buf, ' ') + 1;

  if (*start_pos == '(' && (rank_pos = strchr (start_pos, '['))) {
    /* this is compressed array */
    rlparam_get_array (start_pos, rank_pos, a, array, arrsize, get_my_rank ());
    return ;
  } else {
    /* This is non-compressed array */
    read_array_entry_main (start_pos, a, array, arrsize);
  }
}

void read_vector_param (char * pos, int * param, cvector ** arr)
{
  int64_t * array;
  int size;
  array = rlparam_get_int64 (pos, get_my_rank(), &size);
  if (!array) return;
  if (size == 1) {
    *param = array [0];
  } else {
    assert (arr);
    // the vectors are changed into arrays before the reduction in MPI_Finalize,
    // so it is necessary to maintain the array size parameter. Since the array
    // is printed out with the size as the first number, we skip it here. Ugly.
    //*arr = array_to_vector (array+1, size-1);
    *arr = array_to_vector (array, size-1);
  }
  free (array);
}

void rlvisitor_fill_rank(void *data, int rank) {
  rank_array_t *dlist = (rank_array_t *)data;
  dlist->rank_arr[dlist->curr_ptr] = rank;
  dlist->curr_ptr++;
}

int check_dest(int *rank_list, int *dest_list, int seq_num, rank_array_t *dlist) {

  int src;
  char debugstr[100];
  int dest_arr_sz = member_count(dest_list);
  dlist->rank_arr = (int *)malloc(dest_arr_sz * sizeof(int));
  dlist->curr_ptr = 0;
  traverse_rlist (dest_list, (void *)dlist, rlvisitor_fill_rank);
  for(int i = 0; i < dlist->curr_ptr; i++) {
    src = (my_rank - dlist->rank_arr[i] + my_size ) % my_size;
    
    if(member_rank(src, rank_list) ) {
      //sprintf(debugstr,"chk_dest:seq:%d,src:%d,dl:%d",seq_num,src,dlist.rank_arr[i] );
      //write_to_debug_file(my_rank, debugstr);
      //free(dlist.rank_arr);
      return 1;
    }
  }  
  //free(dlist.rank_arr);
  return 0;
}

void write_to_debug_file(int rank, char *str) {
  FILE *fp;
  char filename[20];
  sprintf(filename, "debug-%d", rank);
  fp = fopen(filename, "a");
  fprintf(fp, "%s\n", str);
  fclose(fp);
}

int get_hist_destination(int h_val, rank_array_t *dest_list, int *rlist ) {
  for(int i=1; i < dest_list->curr_ptr; i++) {
    if( (h_val >= dest_list->rank_arr[i-1] && h_val <= dest_list->rank_arr[i])) {
      if(h_val <= (dest_list->rank_arr[i-1] + dest_list->rank_arr[i])/2)
        return dest_list->rank_arr[i-1];
      else
        return dest_list->rank_arr[i];
    }
  }
}

void fill_dest(rsd_node *node, rank_array_t *dlist) {
  int h_val;
  for(int i=0; i< prsd_count; i++) {
    h_val = histogram_value(node->hist_param, 0, -1);
    h_val = get_hist_destination(h_val, dlist, node->graph.rank_list);
    if(!node->data.dest_vector)
      vector_init(&node->data.dest_vector, h_val);
    else
      vector_push_back(node->data.dest_vector, h_val);
  }
}

void parse_prsd_content(char *buf, rsd_queue *target)
{
  char *start_pos = NULL;
  size_t size;
  int64_t * arr;

#ifdef DSTAR
  start_pos = strchr(buf, ':') + 1;
#else
  start_pos = buf;
#endif
  signed char shift;
  sscanf(start_pos,"%hhx", &shift);
  if (shift < 0) {
    fprintf(stderr, "Malformed trace file; bad field shift: %hhx", shift); 
    exit(1);
  }
  field_mask_t op = 1ll << shift;

  start_pos = strchr(start_pos, ' ');
  if (!start_pos) {
    fprintf(stderr, "Malformed trace file: missing comm value.");
    exit(1);
  }

  /*VT_AK: It is observed that if the destination is not specified in rsd file, it should be interpreted 
    as an offset of -1. If we don't do it here, then the values of src and dest are undefined and can lead
    to arbit results.*/
   //target->tail->data.source = -1;
   //target->tail->data.dest = -1;
  
   /* First check whether the op represents an array or a field parameter.  
    * Mark it as such in the tail. */
   if (op & ARRAY_MASK) {
     target->tail->data.arrays |= op;
   } else {
     target->tail->data.fields |= op;
   }

   /* Now do special handling for each of the ops. */
   switch(op) {
   case COMM:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.comm = 0;//max;//histogram_value(h, 0, -1);
     }
     else {

     sscanf(start_pos, "%d", &target->tail->data.comm);
     }
     break;

   case COUNT:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       //target->tail->hist_param = histogram_input(ch);
       //target->tail->hist_op = op;
       target->tail->data.count = (int)histogram_value_from_vector(h,param_rvector);//histogram_value(h, 0, -1);
       //printf("Rank %d: Root: %d\n", get_my_rank(), target->tail->data.root);
     }
     else {

       if (strchr (start_pos, '[')) {
         //read_vector_param (start_pos, &target->tail->data.count, NULL);
         read_vector_param (start_pos, &target->tail->data.count, &target->tail->data.count_vector);
       } else {
         str2array (strchr(start_pos, ' '), strchr (start_pos, '\n'), &size, &arr);
         if (size == 1)
           target->tail->data.count = arr[0];
         else
           array_to_vector (arr, size);
       }
     }
     break;

   case DATATYPE:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.datatype = 0;//max;//histogram_value(h, 0, -1);
     }
     else {

       sscanf(start_pos, "%d", &target->tail->data.datatype);
     }
     break;

   case RECVCOUNT:
     sscanf(start_pos, "%d", &target->tail->data.recvcount);
     break;

   case RECVTYPE:
     sscanf(start_pos, "%d", &target->tail->data.recvtype);
     break;

   case MPI_OP:
     sscanf(start_pos, "%d", &target->tail->data.mpi_op);
     break;

   case ROOT:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL

       char *ch = strchr(start_pos, '(');
       target->tail->hist_param = histogram_input(ch);
       target->tail->hist_op = op;
       target->tail->data.root = -1; //histogram_value(h, 0, -1);
     }
     else {
       if (strchr (start_pos, '[')) {
         read_vector_param (start_pos, &target->tail->data.root, &target->tail->data.root_vector);
       } else {
         str2array (strchr(start_pos, ' '), strchr (start_pos, '\n'), &size, &arr);
         if (size == 1)
           target->tail->data.root = arr[0];
         else
           target->tail->data.root_vector = array_to_vector (arr, size);
         free (arr);
       }
     }
     break;

   case DEST:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       rank_array_t dlist;
       dlist.rank_arr = NULL;
       dlist.curr_ptr = 0;
       if(strchr(start_pos, '[')) {
         char *rstart = strchr(start_pos, '[');
         char *rend = strchr(start_pos, ']');
         int *rlist;
         //string_to_int_array(rstart+1, rend-1, NULL, &rlist);
         str2intarray(rstart+1, rend-1, NULL, &rlist);
         target->tail->graph.rank_list = rlist;
         //if(!target->tail->match && check_dest(target->tail->rlist, rlist,target->tail->data.seq_num, &dlist)) {
         if( check_dest(target->tail->rlist, rlist,target->tail->data.seq_num, &dlist)) {
           target->tail->match = 1;
         }
       }
       char *ch = strchr(start_pos, '(');
       //Histogram *h = histogram_input(ch);
       target->tail->hist_param = histogram_input(ch);
       target->tail->hist_op = op;
       target->tail->data.dest = -1; //histogram_value(h, 0, -1);
       //printf("Rank %d: Root: %d\n", get_my_rank(), target->tail->data.root);

       //fill_dest(target->tail, &dlist);
       if(dlist.rank_arr)
         free(dlist.rank_arr);

       if(target->tail->prsd_info.tail && target->tail->prsd_info.tail->tail_ptr == target->tail) {
         if(prsd_count > 1)
           prsd_count = prsd_count/(target->tail->prsd_info.tail->loop_iters + 1);
       }
     }
     else {
       if (strchr (start_pos, '[')) {
         read_vector_param (start_pos, &target->tail->data.dest, &target->tail->data.dest_vector);
       } else {
         str2array (strchr(start_pos, ' '), strchr (start_pos, '\n'), &size, &arr);
         if (size == 1)
           target->tail->data.dest = arr[0];
         else
           target->tail->data.dest_vector = array_to_vector (arr, size);
         free (arr);
       }
     }
     break;

   case TAG:
     sscanf(start_pos, "%d", &target->tail->data.tag);
     break;

   case REQUEST:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.request = max;//histogram_value(h, 0, -1);
     }
     else {
       if (strchr (start_pos, '[')) {
         read_vector_param (start_pos, &target->tail->data.request, &target->tail->data.request_vector);
       } else {
         str2array (strchr (start_pos, ' '), strchr (start_pos, '\n'), &size, &arr);
         if (size == 1)
           target->tail->data.request = arr[0];
         else
           target->tail->data.request_vector = array_to_vector (arr, size);
         free (arr);
       }
     }
     break;

   case REORDER:
     sscanf(start_pos, "%d", &target->tail->data.reorder);
     break;

   case NEWRANK:
     break;

   case GROUP:
     sscanf(start_pos, "%d", &target->tail->data.group);
     break;

   case COLOR:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.color = max;//histogram_value(h, 0, -1);
     }
     else {
       sscanf(start_pos, "%d", &target->tail->data.color);
     }
     break;     
		
   case KEY:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.key = max;//histogram_value(h, 0, -1);
     }
     else {
       sscanf(start_pos, "%d", &target->tail->data.key);
     }
     break;
     
   case GROUP1:
   case GROUP2:
   case LOCAL_LEADER:
     sscanf(start_pos, "%d", &target->tail->data.local_leader);
     break;
   case PEER_COMM:
     sscanf(start_pos, "%d", &target->tail->data.peer_comm);
     break;
   case REMOTE_LEADER:
     sscanf(start_pos, "%d", &target->tail->data.remote_leader);
     break;
   case HIGH:
     break;

   case SOURCE:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       //Histogram *h = histogram_input(ch);
       target->tail->hist_param = histogram_input(ch);
       target->tail->hist_op = op;
       target->tail->data.source = -1; //histogram_value(h, 0, -1);
       //printf("Rank %d: Root: %d\n", get_my_rank(), target->tail->data.root);
     }
     else {
       if (strchr (start_pos, '[')) {
         read_vector_param (start_pos, &target->tail->data.source, &target->tail->data.source_vector);
       } else {
         str2array (strchr (start_pos, ' '), strchr (start_pos, '\n'), &size, &arr);
         if (size == 1)
           target->tail->data.source = arr[0];
         else
           target->tail->data.source_vector = array_to_vector (arr, size);
         free (arr);
       }
     }
     break;

   case SENDTAG:
     break;

   case RECVTAG:
     sscanf(start_pos, "%d", &target->tail->data.recvtag);
     break;

   case NEWTYPE:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.newtype = max;//histogram_value(h, 0, -1);
     }
     else {
       sscanf(start_pos, "%d", &target->tail->data.newtype);
     }
     break;

   case OLDTYPE:
     sscanf(start_pos, "%d", &target->tail->data.oldtype);
     break;

   case BLOCKLEN:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.blocklen = max;//histogram_value(h, 0, -1);
     }
     else {
       sscanf(start_pos, "%d", &target->tail->data.blocklen);
     }
     break;

   case STRIDE:
     if (strchr(start_pos, 'N')) { //Histo starts with NULL
       char *ch = strchr(start_pos, '(');
       Histogram *h = histogram_input(ch);
       double min, max;
       int min_node, max_node;
       histogram_min_max(h, &min, &max, &min_node, &max_node);
       target->tail->data.stride = max;//histogram_value(h, 0, -1);
     }
     else {
       sscanf(start_pos, "%d", &target->tail->data.stride);
     }
     break;

   case GLOBAL_WORLD:
     sscanf(start_pos, "%d", &target->tail->data.global_world);
     break;

   case NULL_REQ:
     sscanf(start_pos, "%d", &target->tail->data.null_req);
     break;

   case OUTCOUNT:
     sscanf(start_pos, "%d", &target->tail->data.outcount);
     break;

   case RECVCOUNTS:
   case ARRAY_OF_DIMS:
   case DIMSP:
   case ARRAY_OF_DEGREES:
   case RANGES:
   case ARRAY_OF_RANKS:
   case ARRAY_OF_REQUESTS:
   case ARRAY_OF_BLOCKLENS:
     read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array1, &target->tail->data.array_size1);
     break;
    
   case DISPLS:
   case ARRAY_OF_PERIODS:
   case ARRAY_OF_EDGES:
     read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array2, &target->tail->data.array_size2);
     break;

   case COUNTS:
   case ARRAY_OF_TYPES:
     read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array3, &target->tail->data.array_size3);
     break;

   case RDISPLS:
     read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array4, &target->tail->data.array_size4);
     break;

   case ARRAY_OF_DISPLACEMENTS:
     read_array_entry (start_pos, READ_MPI_AINT, (void**)&target->tail->data.array_of_displacements,
         &target->tail->data.array_size_displs);
     break;

   default:
     // don't print start_pos here, as it's not necessarily inited!
     fprintf(stderr, "error: unhandled parameter with value %hhx:\n%s\n", shift, buf);
   }
}

void parse_prsd_content_extn(char *buf, rsd_queue *target)
{
  char *start_pos = NULL;
  size_t size;
  int64_t * arr;

#ifdef DSTAR
  start_pos = strchr(buf, ':') + 1;
#else
  start_pos = buf;
#endif
  signed char shift;
  sscanf(start_pos,"%hhx", &shift);
  if (shift < 0) {
    fprintf(stderr, "Malformed trace file; bad field shift: %hhx", shift);
    exit(1);
  }
  field_mask_t op = 1ll << shift;
  
  start_pos = strchr(start_pos, ' ');
  if (!start_pos) {
    fprintf(stderr, "Malformed trace file: missing comm value.");
    exit(1);
  }

  /*VT_AK: It is observed that if the destination is not specified in rsd file, it should be interpreted 
    as an offset of -1. If we don't do it here, then the values of src and dest are undefined and can lead
    to arbit results.*/
   target->tail->data.source = -1;
   target->tail->data.dest = -1;
  
   /* First check whether the op represents an array or a field parameter.  
    * Mark it as such in the tail. */
   if (op & ARRAY_MASK) {
     target->tail->data.arrays_extn |= op;
   } else {
     target->tail->data.fields_extn |= op; 
   }
   /* Now do special handling for each of the ops. */
   switch(op) {
     case FILENAME:
       /*Add code to merge filename if there is prefix and postfix*/
       target->tail->data.filename = (char *)malloc((strlen(start_pos)+1)*sizeof(char));
       sscanf(start_pos, "%s", target->tail->data.filename);
       break;

     case AMODE:
       sscanf(start_pos, "%d", &target->tail->data.amode);
       break;

     case FH:
       sscanf(start_pos, "%d", &target->tail->data.fh);
       break;

     case OFFSET:
       sscanf(start_pos, " \(%d,%d,%d\)", &(target->tail->data.offsetinfo.startpos),&(target->tail->data.offsetinfo.stride),
             &(target->tail->data.offsetinfo.totalcount));
       //As of now added code to initialize offset to totalcount*rank, i.e of pattern (k.rank,stride, totalcount),
       //Add code to handle the case of (rank, stride, totalcount)
       target->tail->data.offset = target->tail->data.offsetinfo.totalcount * target->tail->node;
       break;

     case DISP:
       //sscanf(start_pos, "%d", &target->tail->data.disp);
     if (strchr (start_pos, '[')) {
       read_vector_param (start_pos, &target->tail->data.disp, &target->tail->data.disp_vector);
     } else {
       str2array (strchr (start_pos, ' '), strchr (start_pos, '\n'), &size, &arr);
       if (size == 1)
         target->tail->data.disp = arr[0];
       else
         target->tail->data.disp_vector = array_to_vector (arr, size);
       free (arr);
     }
       break;

     case DATAREP:
       target->tail->data.datarep = (char *)malloc((strlen(start_pos)+1)*sizeof(char));
       sscanf(start_pos, "%s", target->tail->data.datarep);
       break;

     case ETYPE:
       sscanf(start_pos, "%d", &target->tail->data.etype);
       break;

     case FILETYPE:
       sscanf(start_pos, "%d", &target->tail->data.filetype);
       break;

     case WHENCE:
       sscanf(start_pos, "%d", &target->tail->data.whence);
       break;

     case COMM_SIZE:
       sscanf(start_pos, "%d", &target->tail->data.comm_size);
       break;

     case RANK:
       sscanf(start_pos, "%d", &target->tail->data.rank);
       break;

     case NDIMS:
       sscanf(start_pos, "%d", &target->tail->data.count);
       break;

     case ORDER:
       sscanf(start_pos, "%d", &target->tail->data.order);
       break;

     case ARRAY_OF_GSIZES:
       read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array1, &target->tail->data.array_size1);
       break;

     case ARRAY_OF_DISTRIBS:
       read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array2, &target->tail->data.array_size2);
       break;

     case ARRAY_OF_DARGS:
       read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array3, &target->tail->data.array_size3);
       break;

     case ARRAY_OF_PSIZES:
       read_array_entry (start_pos, READ_ARR_ELT_T, (void**)&target->tail->data.array4, &target->tail->data.array_size4);
       break;

     default:
       fprintf(stderr, "error: unhandled parameter with value %hhx:\n%s\n", shift, buf);
     
   }
}

