/**
 * Platform-aware stack-sig file by Todd Gamblin, 5/23/2007
 * This attempts to determine what architecture we're on and which stack
 * signature to use.  
 * 
 * Currently thinks that everything that is not x86_64 should use the 
 * regular 32-bit stack_sig file.
 */
#include "stack_sig.h"

#include <mpi.h>
#include "mpi_util.h"

#include <iostream>
#include <cstdlib>
#include <cassert>
#include <cmath>
#include <cstring>
#include <cstdio>
#include <vector>
#include <set>
using namespace std;

#define SIG_DEPTH 1000
extern "C" intptr_t * do_stack_unwind ();
extern "C" void       do_stack_free   (intptr_t*);

int sig_len(const intptr_t *sig) {
  if (!sig) return 0;

  int len = 0;
  while (sig[len] != 0) len++;
  return len;
}

/// Comparator structure for signatures.  Use this to put signatures 
/// in the internal set we keep.  Note that we really only need to use 
/// this expensive comparator internally for the initial signature 
/// lookup.  External data structures can use pointer comparison to 
/// compare, vis sig_lt.
struct ltsig {
  bool operator()(const intptr_t* sig1, const intptr_t* sig2) const {
    if (!sig1 || !sig2) {
      return sig1 < sig2;
    }

    int i;
    for (i=0; sig1[i] && sig2[i]; i++) {
      if (sig1[i] < sig2[i]) {
	return true;
	
      } else if (sig1[i] > sig2[i]) {
	return false;
      }
    }
    return (!sig1[i] && sig2[i]); // only way it's less
  }
};


set<const intptr_t *, ltsig> signatures;

// returns unique pointers to strings
static const intptr_t *sig_lookup(const intptr_t *sig) {
  set<const intptr_t *>::iterator s = signatures.find(sig);

  if (s == signatures.end()) {
    int len = sig_len(sig);
    intptr_t *new_sig = new intptr_t[len + 1];
    for (int i=0; i < len+1; i++) {
      new_sig[i] = sig[i];
    }
    s = signatures.insert(new_sig).first;
  }
  return *s;
}

void sig_init(stack_sig *signature) {
  signature->sig = NULL;
}

void sig_destroy(stack_sig *signature) {
  signature->sig = NULL;
}

void sig_assign(stack_sig *dest, const stack_sig *src) {
  dest->sig = src->sig;
}

bool already_copied (intptr_t sig, intptr_t *arr, int j) {
  for (int i = 0; i < j; i++)
    if (sig == arr[i]) return true;
  return false;
}

void sig_unwind(stack_sig *signature) {
  intptr_t * uw_sig = do_stack_unwind ();
#if 1
  intptr_t tmp_sig [SIG_DEPTH];
  int i = 0, j = 0;
  i = 0; j = 0;
  tmp_sig [j++] = uw_sig[i++];
  while (uw_sig[i])
  {
    if (!already_copied (uw_sig[i], tmp_sig, j)) {
      tmp_sig [j++] = uw_sig[i];
    }
    i++; 
  }
  tmp_sig [j] = 0;
#else
  intptr_t *tmp_sig = uw_sig;
#endif
  signature->sig = sig_lookup(tmp_sig);
  do_stack_free (uw_sig);
}


int sig_packed_size(const stack_sig *signature, MPI_Comm comm) {
  int size = 0;
  size += packed_size(1, MPI_INT, comm); // length
  if (signature->sig) {
    size += packed_size(sig_len(signature->sig) + 1, MPI_INTPTR_TYPE, comm);
  }
  return size;
}


void sig_pack(const stack_sig *signature, void *buf, int bufsize, int *position, MPI_Comm comm) {
  int len = 0;
  if (signature->sig) {
    len = sig_len(signature->sig) + 1;
  }
  PMPI_Pack(&len, 1, MPI_INT, buf, bufsize, position, comm);

  if (len) {
    PMPI_Pack(const_cast<intptr_t*>(signature->sig), len, MPI_INTPTR_TYPE,
	      buf, bufsize, position, comm);
  }
}


void sig_unpack(stack_sig *signature, void *buf, int bufsize, int *position, MPI_Comm comm) {
  int len;
  PMPI_Unpack(buf, bufsize, position, &len, 1, MPI_INT, comm);
  if (len) {
    intptr_t tmp_sig[len+1];
    PMPI_Unpack(buf, bufsize, position, &tmp_sig, len, MPI_INTPTR_TYPE, comm);
    tmp_sig[len] = 0;
    signature->sig = sig_lookup(tmp_sig);
  } else {
    signature->sig = NULL;
  }
}


int sig_soutput(char *buf, const stack_sig *signature) {
  char *pos = buf;

  if (!signature->sig) {
    pos += sprintf(pos, "NULL");
  } else {
    int len = sig_len(signature->sig);
    if (len > 0) {
      pos += sprintf(pos, "%" PRIxPTR, signature->sig[0]);
    }
    
    for (int i=1; i < len; i++) {
      pos += sprintf(pos, "_%" PRIxPTR, signature->sig[i]);
    }
  }
  return (pos - buf);
}


void sig_input(stack_sig *signature, char *buf) {
  sig_init(signature);

  vector<intptr_t> addrs;
  char *start = buf;
  while (true) {
    addrs.push_back(0);
    sscanf(start, "%"PRIxPTR, &addrs.back());
    if (!(start = strchr(start, '_'))) break;
    start++;
  }
  addrs.push_back(0);

  if (addrs.size()) {
    signature->sig = sig_lookup(&addrs[0]);
  }
}
