#include <stdlib.h>
#include <assert.h>
#include <stdint.h>
#include <regex.h>
#include "transfer.h"
#include "umpi_internal.h"
#include "internal.h"
#include "opstruct.h"
#include "op_info.h"
#include "utils.h"
#include "cvector.h"

#ifdef FEATURE_TIMING_HISTO
#include "Histogram.h"
#include "TimeList.h"
#endif /* FEATURE_TIMING_HISTO */
#ifdef FEATURE_TIMING_BITS
#include "Intseq.h"
#include "TimeList.h"
#endif

#include "VParam.h"
#include "Event.h"
#define DEBUG(...)
//#define DEBUG fprintf

/* Weird weird compile error on AIX */
#ifdef rlist
#undef rlist
#endif

#ifndef print_perror
#define print_perror() do {\
    char perrorbuf [32];  \
    sprintf (perrorbuf, "%s: %d", __FILE__, __LINE__);  \
    perror(perrorbuf);  \
    exit(0);  } while (0)
#endif


/* Macros for packing/unpacking rsd queues. ==================================== */
/* Used by rsd_queue_packed_size. */
#define ADD_ARR_SIZE(size, rsd_ptr, comm, num) \
do { \
  if (rsd_ptr->data.array##num) { \
    size += packed_size(1, MPI_INT, comm); \
    size += packed_size(rsd_ptr->data.array_size##num * sizeof(array_elt_t), MPI_BYTE, comm); \
  } \
} while (0)

/* Used by send_rsd_queue. */
#define PACK_ARR(rsd_ptr, buf, bufsize, pos, comm, num) \
do { \
  if (rsd_ptr->data.array##num) { \
    int packSize = rsd_ptr->data.array_size##num * sizeof(array_elt_t); \
    PMPI_Pack(&rsd_ptr->data.array_size##num, 1, MPI_INT, buf, bufsize, &pos, comm);  \
    PMPI_Pack(rsd_ptr->data.array##num, packSize, MPI_BYTE, buf, bufsize, &pos, comm); \
  } \
} while (0)

/* Used by recv_rsd_queue. */
#define UNPACK_ARR(buf, bufsize, pos, rsd_ptr, comm, num) \
do { \
  if (rsd_ptr->data.array##num) { \
    PMPI_Unpack(buf, bufsize, &pos, &rsd_ptr->data.array_size##num, 1, MPI_INT, comm); \
    int packSize = rsd_ptr->data.array_size##num * sizeof(array_elt_t); \
    rsd_ptr->data.array##num = (array_elt_t*)malloc(packSize); \
    PMPI_Unpack(buf, bufsize, &pos, rsd_ptr->data.array##num, packSize, MPI_BYTE, comm); \
  } \
} while (0)

/* For packing histograms. */
# define need_limits(a,b) 1
//# define need_limits(op, i) ((((i) == HISTOGRAM_POST) &&  
//  ((op)==umpi_MPI_Recv || is_wait_after_async (op))) || is_collective(op))
/* =============================================================================== */


extern int my_rank;

/* this looks embarassingly long, but what we're trying to do here
 * is to see whether we have
 * (a) two  nodes with either  the same offset or the same absolute color/key
 * (b) two groups with either  the same offset or the same absolute color/key
 * (c) a group and a node with the same offset or the same absolute color/key
 */
int can_merge_color_or_key (rsd_node * master, rsd_node * slave, int64_t param,
                            int * match)
{
  int master_has_offset = has_rank_offset (master->data, param);
  int  slave_has_offset = has_rank_offset ( slave->data, param);

  int64_t master_param = 0;
  int64_t  slave_param = 0;

  if (param == COLOR) {
      master_param = master->data.color;
      slave_param =  slave->data.color;
  } else if (param == KEY) {
      master_param = master->data.key;
      slave_param =  slave->data.key;
  } else {
    assert (0);
  }

  int master_has_other = (1 != member_count (master->rlist));
  int  slave_has_other = (1 != member_count ( slave->rlist));

  if (!master_has_offset && !slave_has_offset) {
    /* if both are constants ... */

    /* ... and are not the same offset from rank, we don't match them
     * in future, we may try to match based on other criteria such as 
     * mod(rank) etc */
    if ( (master_param - master->node) != (slave_param - slave->node)) {
      return FALSE;
    } else {
      /* merge done by setting flag at the end of the function */
    }
  } else if ( master_has_offset && !slave_has_offset) {

    /* master has a bunch of ranks which match based on offset,
     * whereas slave has a bunch  which  match exactly */ 
    if (slave_has_other) {
      return FALSE;
    }

    if (master_param != ( slave_param -  slave->node)) {
      return FALSE;
    }
    else {
      return  TRUE;
    }

    /* master->data.color/key is already an offset */
  } else if (!master_has_offset &&  slave_has_offset) {

    if (master_has_other) {
      return FALSE;
    }

    if ( slave_param != (master_param - master->node)) {
      return FALSE;
    }

    /* merge done by setting flag at the end of the function */

  } else if (master_has_offset && slave_has_offset) {

  /* If both are offsets and differ, we don't match them
   * in future we may create a rlparam list for these */
    if (master_param != slave_param ) {
      return FALSE;
    } else {
      return  TRUE;
    }
  }

  (*match) |= param;
  return TRUE;
}

/* this is called only when can_merge_color_or_key sets the flag */
void merge_color_or_key (rsd_node *master, int64_t param)
{
  /* change the value in master to an offset and mark it as a offset too */
  if (param == COLOR) {
    master->data.color -= master->node;
  } else if (param == KEY) {
    master->data.key -= master->node;
  } else {
    assert (0);
  }
  set_rank_offset (master->data, param);

  /* clear the flag here */
  master->match &= ~param;
}

#ifdef FEATURE_FILE_IO
void extractNodeFromFileName(char* filename, char* result, int resultsize)
{
  int diff = 0;
  char* lasthyphen = strrchr(filename,'-');
  if(lasthyphen == NULL)
    return;//It is out of notation that is supported.

  char* lastperiod = strrchr(filename,'.');
  if(lastperiod == NULL) //There is no extension in filename.
    lastperiod = strchr(filename,'\0');

  diff = lastperiod-lasthyphen;

  if(diff>resultsize) {
    return; //Node number digits cannot be so huge.
  }
  bcopy(lasthyphen+1,result,(diff-1)*sizeof(char));
  result[diff-1]='\0';
}

void generalizeFileName(char * filename)
{
  char * lasthyphen = strrchr(filename,'-');
  char * lastperiod = strrchr(filename,'.');
  if(lastperiod == NULL)
    lastperiod = strchr(filename,'\0');
  *(lasthyphen+1) = '#';
  bcopy(lastperiod,(lasthyphen+2),(strlen(lastperiod)+1)*sizeof(char));
}
#endif

int can_merge_params (rsd_node * master, rsd_node * slave, int * match)
{
  /* merge alltoallv */
  if (master->data.op == umpi_MPI_Alltoallv &&
      slave->data.op ==  umpi_MPI_Alltoallv && (*match) == MFAIL_ARR) {
    return TRUE;
  }

  #ifdef FEATURE_FILE_IO
  int match_mask = ( ~ (MFAIL_SRC|MFAIL_DST|MFAIL_CNT|MFAIL_ROOT|MFAIL_CLR|MFAIL_KEY|MFAIL_REQ|MFAIL_FLN|MFAIL_OFF|MFAIL_DISP) );
  #else
  int match_mask = ( ~ (MFAIL_SRC|MFAIL_DST|MFAIL_CNT|MFAIL_ROOT|MFAIL_CLR|MFAIL_KEY|MFAIL_REQ) );
  #endif
  if ((*match) & match_mask) {
    return FALSE;
  }
  #ifdef FEATURE_FILE_IO
  if((*match) & MFAIL_FLN) {

    if(strcmp(master->data.filepath,slave->data.filepath)!=0) {
      fprintf(stderr,"Warning:Merging of RSDs with different filepath is not supported\n");
      return FALSE;
    }

    regex_t regexp;
    char* reg = "[a-z0-9A-Z]*\-[0-9#]+\.?[a-z0-9A-Z]*";
    regcomp(&regexp, reg, REG_EXTENDED);

    if(regexec(&regexp,master->data.filename,0,NULL, 0)!=0 || regexec(&regexp,slave->data.filename,0,NULL, 0)!=0) {
      fprintf(stderr,"Warning: For complete compression, filename should be of pattern <characters>-nodenumber[.extn]\n");
      return FALSE;
    }
    
    char nodenum[10];
    extractNodeFromFileName(master->data.filename, &nodenum, 10);
    if((nodenum[0]!='#') && (atoi(&nodenum) != master->node)) {
      fprintf(stderr,"Warning: For complete compression, filename should have corresponding task number in the pattern <characters>-nodenumber[.extn]\n");
      return FALSE;
    }

    memset(&nodenum,'\0',10);
    extractNodeFromFileName(slave->data.filename, &nodenum, 10);
    if((nodenum[0]!='#') && (atoi(&nodenum) != slave->node)) {
      fprintf(stderr,"Warning: For complete compression, filename should have corresponding task number in the pattern <characters>-nodenumber[.extn]\n");
      return FALSE;
    }    
  }
  #endif //FEATURE_FILE_IO
  if ((*match) & MFAIL_CLR)  {
    /* clear the bit; the bit may be set again if anything needs to be done
     * for the merge */
    (*match) &= ~MFAIL_CLR;
    int merged = can_merge_color_or_key (master, slave, COLOR, match);
    if (merged == FALSE) return FALSE;
  }
  if ((*match) & MFAIL_KEY) { /* mismatch in KEY */
    (*match) &= ~MFAIL_KEY;
    int merged = can_merge_color_or_key (master, slave,   KEY, match);
    if (merged == FALSE) return FALSE;
  }

  return TRUE;
}

rsd_node * get_loop_end(rsd_node *node, prsd_node *prsd_iter) {
  rsd_node * ret_node = node;
 
  if(prsd_iter) {
    for(int i = 1; i < prsd_iter->loop_mems; i++)
      ret_node = ret_node->next;
  }
  return ret_node;
}

int match_sequence(rsd_node *master, rsd_node *slave)
{
  prsd_node *sprsd_iter, *mprsd_iter;
  stack_sig *tail, *sig;
  rsd_node *loop_end, *old_master;
  /*if (master->prsd_info.head && !SIG_MATCH(master->signature, master->prsd_info.head->tail_sig)) {
    tail = &master->prsd_info.head->tail_sig;
  } else {
    tail = NULL;
  }*/
  loop_end = get_loop_end(master, master->prsd_info.head);

  do {
    /* sig mismatch */
    if(!SIG_MATCH(master->signature, slave->signature))
      return FALSE;

    /* uneven depths */
    if(get_depth(slave->prsd_info.head) != get_depth(master->prsd_info.head))
      return FALSE;

    /* loop info checks */
    sprsd_iter = slave->prsd_info.head;
    mprsd_iter = master->prsd_info.head;
    while(sprsd_iter && mprsd_iter) {
     #ifdef FEATURE_COMPRESS_HISTO
      if (sprsd_iter->loop_iters != mprsd_iter->loop_iters) {
        histogram_merge(&mprsd_iter->prsd_stat, mprsd_iter->loop_iters, &sprsd_iter->prsd_stat, 
	  sprsd_iter->loop_iters, slave->node, 5);
      }
     #else
      if (sprsd_iter->loop_iters != mprsd_iter->loop_iters)
        return FALSE;
     #endif
      if (sprsd_iter->loop_mems != mprsd_iter->loop_mems)
        return FALSE;
      if (!SIG_MATCH(sprsd_iter->tail_sig, mprsd_iter->tail_sig))
        return FALSE;
      sprsd_iter = sprsd_iter->next;
      mprsd_iter = mprsd_iter->next;
    }
    int match = params_match (&master->data, &slave->data);
    int event_match = compare_event_params(master->event, slave->event);

    /*if(match == event_match) 
      printf("CHECK: transfer.c - both param and event return same value:param_match %d, event_match %d\n", match,event_match);
    else
      printf("CHECK: transfer.c - mismatch in param and event return value: param_match %d, event_match %d\n",match, event_match);
    */
    if ( !(can_merge_params (master, slave, &event_match)) )
        return FALSE;

    master->match = event_match;

    /* Set this *before* moving the master ptr ahead */
    //sig = &master->signature;
    old_master = master;
    master = master->next;
    slave = slave->next;
      
  } while (old_master != loop_end); //(tail && !SIG_MATCH(*sig, *tail));
    
  return TRUE;
}

static void merge_timing (rsd_node * master_iter, rsd_node * slave_iter, int slave_rank)
{
#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
  stack_sig msig, ssig;
  sig_init(&msig);
  sig_init(&ssig);
  
  for (int i = 0; i < TIMING_MAX; i++)
  {
    // the signature aren't important here. We've already established a match.
    timing_list_merge (&master_iter->data.time_info[i].time_list,
        master_iter->data.time_info[i].time,
        &msig, master_iter->node,
        &slave_iter->data.time_info[i].time_list,
        slave_iter->data.time_info[i].time,
        &ssig, slave_iter->node,
        NUM_BUCKETS
        );
    //Event Timing Merge
    event_time_merge(i, master_iter->event, &msig, master_iter->node, 
                     slave_iter->event, &ssig, slave_iter->node, NUM_BUCKETS);
  }
#endif
}

void merge_params (rsd_node * master, rsd_node * slave) {

  int64_t param = 0;
  unsigned char mfail_src = (master->match & MFAIL_SRC);
  unsigned char mfail_dst = (master->match & MFAIL_DST);
  unsigned char mfail_cnt = (master->match & MFAIL_CNT);
  unsigned char mfail_clr = (master->match & MFAIL_CLR);
  unsigned char mfail_key = (master->match & MFAIL_KEY);
  unsigned char mfail_arr = (master->match & MFAIL_ARR);
  unsigned char mfail_req = (master->match & MFAIL_REQ);
  int mfail_root = (master->match & MFAIL_ROOT);

  #ifdef FEATURE_FILE_IO
  int mfail_fln = (master->match & MFAIL_FLN);
  int mfail_off = (master->match & MFAIL_OFF);
  int mfail_disp = (master->match & MFAIL_DISP);
  if ( !(mfail_cnt || mfail_src || mfail_dst || mfail_clr || mfail_key || mfail_arr || mfail_root || mfail_fln!=0 || mfail_off!=0 || mfail_disp!=0) )
                                    return;

  if(mfail_fln) {
    master->match = 0;
    generalizeFileName(master->data.filename);
  }
  if(mfail_off) {
    master->match = 0;//Add code to generalize offset values
  }
  #else
  if ( !(mfail_cnt || mfail_src || mfail_dst || mfail_clr || mfail_key || mfail_arr || mfail_root) )
                                    return;
  #endif
  if (mfail_arr) {
    master->match &= ~(MFAIL_ARR);

    if (master->data.op == umpi_MPI_Alltoallv) {
      rlparam_merge_arr (RECVCOUNTS, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
          master->data.array_size1, master->data.array1, master->rlist, slave->data.array_size1, slave->data.array1, slave->rlist);
       rlparam_merge_arr (   DISPLS, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
          master->data.array_size2, master->data.array2, master->rlist, slave->data.array_size2, slave->data.array2, slave->rlist);
      rlparam_merge_arr (    COUNTS, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
          master->data.array_size3, master->data.array3, master->rlist, slave->data.array_size3, slave->data.array3, slave->rlist);
      rlparam_merge_arr (   RDISPLS, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
          master->data.array_size4, master->data.array4, master->rlist, slave->data.array_size4, slave->data.array4, slave->rlist);
      master->data.pvectors |= (RECVCOUNTS|DISPLS|COUNTS|RDISPLS);
    }
  }

  if (mfail_clr) {
    mfail_clr = 0;
    merge_color_or_key (master, COLOR);
  }
  if (mfail_key) {
    mfail_key = 0;
    merge_color_or_key (master,   KEY);
  }

  int * master_arr, master_size, * slave_arr, slave_size;

  #ifdef FEATURE_FILE_IO
  if (mfail_disp)
  {
    if (master->data.disp_vector) {
      master_arr = vector_to_array (master->data.disp_vector, &master_size);

    } else {
      master_arr = malloc (sizeof (int));
      master_arr[0] = master->data.disp;
      master_size = 1;
    }
    if (slave->data.disp_vector) {
      slave_arr = vector_to_array (slave->data.disp_vector, &slave_size);
    } else {
      slave_arr = malloc (sizeof (int));
      slave_arr[0] = slave->data.disp;
      slave_size = 1;
    }
    master->match &= ~(MFAIL_DISP);
    DEBUG (stderr, "[%d] mfail_disp\n", get_my_rank());
    rlparam_merge_arr (DISP, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
        master_size, master_arr, master->rlist, slave_size, slave_arr, slave->rlist);
    vector_free (&master->data.source_vector);
    vector_free (&slave->data.source_vector);
    free (master_arr);
    free (slave_arr);
    master->data.pvectors |= DISP;
  }
  #endif

  if (mfail_src)
  {
    if (master->data.source_vector) {
      master_arr = vector_to_array (master->data.source_vector, &master_size);

    } else {
      master_arr = malloc (sizeof (int));
      master_arr[0] = master->data.source;
      master_size = 1;
    }
    if (slave->data.source_vector) {
      slave_arr = vector_to_array (slave->data.source_vector, &slave_size);
    } else {
      slave_arr = malloc (sizeof (int));
      slave_arr[0] = slave->data.source;
      slave_size = 1;
    }
    master->match &= ~(MFAIL_SRC);
    DEBUG (stderr, "[%d] mfail_src\n", get_my_rank());
    rlparam_merge_arr (SOURCE, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
        master_size, master_arr, master->rlist, slave_size, slave_arr, slave->rlist);
    vector_free (&master->data.source_vector);
    vector_free (&slave->data.source_vector);
    free (master_arr);
    free (slave_arr);
    master->data.pvectors |= SOURCE;

    //Merge Event Params
    event_merge_param_value_list(master->event, slave->event, TYPE_SOURCE, 
                               master->rlist, slave->rlist);

  }
  if (mfail_dst)
  {
    if (master->data.dest_vector)
      master_arr = vector_to_array (master->data.dest_vector, &master_size);
    else {
      master_arr = malloc (sizeof (int));
      master_arr[0] = master->data.dest;
      master_size = 1;
    }
    if (slave->data.dest_vector)
      slave_arr = vector_to_array (slave->data.dest_vector, &slave_size);
    else {
      slave_arr = malloc (sizeof (int));
      slave_arr[0] = slave->data.dest;
      slave_size = 1;
    }

    master->match &= ~(MFAIL_DST);
    DEBUG (stderr, "mfail_dst\n");
    rlparam_merge_arr (DEST, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
        master_size, master_arr, master->rlist, slave_size, slave_arr, slave->rlist);
    vector_free (&master->data.dest_vector);
    vector_free (&slave->data.dest_vector);
    free (master_arr);
    free (slave_arr);
    master->data.pvectors |= DEST;
    //Merge Event Params
    event_merge_param_value_list(master->event, slave->event, TYPE_DEST, 
                               master->rlist, slave->rlist);
  }
  if (mfail_req)
  {
    if (master->data.request_vector)
      master_arr = vector_to_array (master->data.request_vector, &master_size);
    else {
      master_arr = malloc (sizeof (int));
      master_arr[0] = master->data.request;
      master_size = 1;
    }
    if (slave->data.request_vector)
      slave_arr = vector_to_array (slave->data.request_vector, &slave_size);
    else {
      slave_arr = malloc (sizeof (int));
      slave_arr[0] = slave->data.request;
      slave_size = 1;
    }

    master->match &= ~(MFAIL_REQ);
    DEBUG (stderr, "mfail_req\n");
    rlparam_merge_arr (REQUEST, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
        master_size, master_arr, master->rlist, slave_size, slave_arr, slave->rlist);

    vector_free (&master->data.request_vector);
    vector_free (&slave->data.request_vector);
    free (master_arr);
    free (slave_arr);
    master->data.pvectors |= REQUEST;
    //Merge Event Params
    event_merge_param_value_list(master->event, slave->event, TYPE_REQUEST, 
                               master->rlist, slave->rlist);
  }

  if (mfail_cnt)
  {
    if (master->data.count_vector) {
      master_arr = vector_to_array (master->data.count_vector, &master_size);

    } else {
      master_arr = malloc (sizeof (int));
      master_arr[0] = master->data.count;
      master_size = 1;
    }
    if (slave->data.count_vector) {
      slave_arr = vector_to_array (slave->data.count_vector, &slave_size);
    } else {
      slave_arr = malloc (sizeof (int));
      slave_arr[0] = slave->data.count;
      slave_size = 1;
    }
    master->match &= ~(MFAIL_CNT);
    DEBUG (stderr, "[%d] mfail_cnt\n", get_my_rank());
    rlparam_merge_arr (COUNT, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
        master_size, master_arr, master->rlist, slave_size, slave_arr, slave->rlist);
    vector_free (&master->data.count_vector);
    vector_free (&slave->data.count_vector);
    free (master_arr);
    free (slave_arr);
    master->data.pvectors |= COUNT;
    //Merge Event Params
    event_merge_param_value_list(master->event, slave->event, TYPE_COUNT, 
                               master->rlist, slave->rlist);
  }

  if (mfail_root)
  {
    if (master->data.root_vector) {
      master_arr = vector_to_array (master->data.root_vector, &master_size);

    } else {
      master_arr = malloc (sizeof (int));
      master_arr[0] = master->data.root;
      master_size = 1;
    }
    if (slave->data.root_vector) {
      slave_arr = vector_to_array (slave->data.root_vector, &slave_size);
    } else {
      slave_arr = malloc (sizeof (int));
      slave_arr[0] = slave->data.root;
      slave_size = 1;
    }
    master->match &= ~(MFAIL_ROOT);
    DEBUG (stderr, "[%d] mfail_cnt\n", get_my_rank());
    rlparam_merge_arr (ROOT, (RLParamList*)&master->data.rlparam, (RLParamList*)& slave->data.rlparam,
        master_size, master_arr, master->rlist, slave_size, slave_arr, slave->rlist);
    vector_free (&master->data.root_vector);
    vector_free (&slave->data.root_vector);
    free (master_arr);
    free (slave_arr);
    master->data.pvectors |= ROOT;
    //Merge Event Params
    event_merge_param_value_list(master->event, slave->event, TYPE_ROOT, 
                               master->rlist, slave->rlist);
  }

  while (master->match) {

    int64_t val[2];
    if (mfail_cnt) {    /* mismatch in COUNT */
      val[0] = master->data.count;
      val[1] =  slave->data.count;
      param = COUNT;
      mfail_cnt = 0;
      master->match &= ~(MFAIL_CNT);
    }

    rlparam_merge (param,
        (RLParamList*)&master->data.rlparam,
        (RLParamList*)& slave->data.rlparam,
        val[0], master->rlist, 
        val[1],  slave->rlist);
    if (master->data.rlparam)
      master->data.pvectors |= param;

  }
}

int rsd_queue_packed_size(rsd_queue *queue, MPI_Comm comm) {
  const int terminal_size = packed_size(sizeof(rsd_node), MPI_BYTE, comm);
  const int loopinfo_size = packed_size(sizeof(prsd_node), MPI_BYTE, comm);

  int size = packed_size(1, MPI_INT, comm);  // start with one int for size.

  rsd_node *rsd_ptr = queue->head;
  while(rsd_ptr) {
    size += terminal_size;

    /*Event Size calculation*/
    size += event_packed_size(rsd_ptr->event, comm);
    /* Stack signature */
    size += sig_packed_size(&rsd_ptr->signature, comm);

    /* rank list */
    if(rsd_ptr->rlist) {
      size += packed_size(1, MPI_INT, comm);
      size += packed_size(size_recursive_list(rsd_ptr->rlist), MPI_INT, comm);
    }

    /* parameter arrays */
    ADD_ARR_SIZE(size, rsd_ptr, comm, 1);
    ADD_ARR_SIZE(size, rsd_ptr, comm, 2);
    ADD_ARR_SIZE(size, rsd_ptr, comm, 3);
    ADD_ARR_SIZE(size, rsd_ptr, comm, 4);

    /* prsd stack size */
    if (rsd_ptr->prsd_info.head) {
      size += packed_size(1, MPI_INT, comm); // for size of list

      prsd_node *prsd_ptr = rsd_ptr->prsd_info.head;
      while(prsd_ptr) {
        size += loopinfo_size;
	size += sig_packed_size(&prsd_ptr->tail_sig, comm);  // tail signature 
	//Added by Karthik - start
	size += packed_size(1, MPI_INT, comm);
	if(prsd_ptr->prsd_stat)
	  size += histogram_packed_size(prsd_ptr->prsd_stat, comm);
	//Added by Karthik - end
        prsd_ptr = prsd_ptr->next;
      }
    }

    size += vector_packed_size  (rsd_ptr->data.source_vector, comm);
    size += vector_packed_size  (rsd_ptr->data.dest_vector, comm);
    size += vector_packed_size  (rsd_ptr->data.request_vector, comm);
    size += rlparam_packed_size (rsd_ptr->data.rlparam, comm);
    size += vector_packed_size  (rsd_ptr->data.count_vector, comm);
    size += vector_packed_size  (rsd_ptr->data.root_vector, comm);
#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
    /* histogram bins size */
    for (int i = 0; i < TIMING_MAX; i++) {
      if (rsd_ptr->data.time_info[i].time_list) {
        size += timing_list_packed_size
          (rsd_ptr->data.time_info[i].time_list, comm);
      }
    }
#endif

//Added by Karthik -Start
#ifdef FEATURE_FILE_IO
    size += packed_size(3, MPI_INT, comm);//3 int size to track strlen of filepath,filename,datarep
    if(rsd_ptr->data.filepath)
      size += packed_size(strlen(rsd_ptr->data.filepath)+1, MPI_CHAR, comm);
    if(rsd_ptr->data.filename)
      size += packed_size(strlen(rsd_ptr->data.filename)+1, MPI_CHAR, comm);
    if(rsd_ptr->data.datarep)
      size += packed_size(strlen(rsd_ptr->data.datarep)+1, MPI_CHAR, comm);

    size += vector_packed_size  (rsd_ptr->data.disp_vector, comm);
#endif
//Added by Karthik - End
    /* graph size */
    size += dep_graph_packed_size (&rsd_ptr->graph, comm); // out edge

    rsd_ptr = rsd_ptr->next;
  }

  return size;
}

void dep_graph_ptr_to_index (rsd_node * rsd_ptr)
{
  int index = 0;
  
  for (rsd_node *ptr = rsd_ptr; ptr; ptr = ptr->next)
    ptr->graph.index = index++;

  for (rsd_node *ptr = rsd_ptr; ptr; ptr = ptr->next)
  {
    for (list_t l = ptr->graph.in_edges; l; l = l->next) {
      //DEBUG (stderr, "[send] ptr = %p, index = %d\n", l->v,((rsd_node*)(l->v))->graph.index);
      l->v = (void*)((rsd_node*)(l->v))->graph.index;
      assert (l->v >= 0);
    }
    for (list_t l = ptr->graph.out_edges; l; l = l->next) {
      //DEBUG (stderr, "[send] ptr = %p, index = %d\n", l->v,((rsd_node*)(l->v))->graph.index);
      l->v = (void*)((rsd_node*)(l->v))->graph.index;
      assert (l->v >= 0);
    }
  }
  fflush (stderr);
}

void dep_graph_index_to_ptr (rsd_node * rsd_ptr)
{
  int num = 0;

  for (rsd_node *ptr = rsd_ptr; ptr; ptr = ptr->next)
    num++;

  void ** table = malloc (num * sizeof (void*));
  int index = 0;
  for (rsd_node *ptr = rsd_ptr; ptr; ptr = ptr->next) {
    ptr->location = Slave_Queue;
    ptr->graph.color = White;
    table [index++] = ptr;
  }

  for (rsd_node *ptr = rsd_ptr; ptr; ptr = ptr->next)
  {
    for (list_t l = ptr->graph.in_edges; l; l = l->next) {
      //DEBUG (stderr, "l->v = %d, table[l->v] = %p\n", l->v, table[(intptr_t)(l->v)]);
      l->v = table [(intptr_t)(l->v)];
      assert (l->v >= 0);
    }
    for (list_t l = ptr->graph.out_edges; l; l = l->next) {
      //DEBUG (stderr, "l->v = %d, table[l->v] = %p\n", l->v, table[(intptr_t)(l->v)]);
      l->v = table [(intptr_t)(l->v)];
      assert (l->v >= 0);
    }
  }
  free (table);
  //fflush (stderr);
}

void send_rsd_queue(rsd_queue *queue, MPI_Comm comm, int dest) {
  //printf("send_rsd_queue: %d::",get_my_rank());
  //printrsdqueue(queue);
  int bufsize = rsd_queue_packed_size(queue, comm);  // get an upper bound on packed size
  void *buf = malloc(bufsize);                       // buffer for MPI packing.
  int position = 0;                                  // position in the buffer (for MPI)
  rsd_node *rsd_ptr = queue->head;

  dep_graph_ptr_to_index (queue->head);

  // first send the upper bound on buffer size so the other side knows how much to allocate
  // doing this here should also let the other side do the allocating while we do the packing,
  // even if it might be a little larger than sending the actual size once we're done.
  PMPI_Send(&bufsize, 1, MPI_INT, dest, 0, comm);

  // pack number of rsd nodes in the send.
  int size = rsd_get_length(queue);
  PMPI_Pack(&size, 1, MPI_INT, buf, bufsize, &position, comm);

  // pack the rest of the queue
  while(rsd_ptr) {
    /* rsd node */
    PMPI_Pack(rsd_ptr, sizeof(rsd_node), MPI_BYTE, buf, bufsize, &position, comm);

    /*Pack Event structure*/
    //int backpos = position;
    event_pack(rsd_ptr->event, buf, bufsize, &position, comm);
    /*int esize = event_packed_size(rsd_ptr->event, comm);
    char * bbuf = malloc(esize+1);
    memset(bbuf, 0, esize+1);
    event_pack(rsd_ptr->event, bbuf, esize, &backpos, comm);
    printf("PPPPPPacked event is %s\n",bbuf);*/
    /* Stack signature */
    sig_pack(&rsd_ptr->signature, buf, bufsize, &position, comm);

    /* rank list */
    if (rsd_ptr->rlist) {
      int len = size_recursive_list(rsd_ptr->rlist);
      PMPI_Pack(&len, 1, MPI_INT, buf, bufsize, &position, comm);
      PMPI_Pack(rsd_ptr->rlist, len, MPI_INT, buf, bufsize, &position, comm);
    }

    PACK_ARR(rsd_ptr, buf, bufsize, position, comm, 1);
    PACK_ARR(rsd_ptr, buf, bufsize, position, comm, 2);
    PACK_ARR(rsd_ptr, buf, bufsize, position, comm, 3);
    PACK_ARR(rsd_ptr, buf, bufsize, position, comm, 4);

    /* prsd stack */
    if (rsd_ptr->prsd_info.head) {
      prsd_node *prsd_ptr = rsd_ptr->prsd_info.head;
      int prsd_size = get_depth(prsd_ptr);
      PMPI_Pack(&prsd_size, 1, MPI_INT, buf, bufsize, &position, comm);

      while (prsd_ptr) {
        PMPI_Pack(prsd_ptr, sizeof(prsd_node), MPI_BYTE, buf, bufsize, &position, comm);
	sig_pack(&prsd_ptr->tail_sig, buf, bufsize, &position, comm);	// tail signature
	//Added by Karthik - Start
	int histo_present = 0;
	if(prsd_ptr->prsd_stat){
	  histo_present = 1;
	}
	PMPI_Pack(&histo_present, 1, MPI_INT, buf, bufsize, &position, comm);
	if(prsd_ptr->prsd_stat) 
	  histogram_pack(prsd_ptr->prsd_stat, buf, bufsize, &position, comm);
	//Added by Karthik - End
        prsd_ptr = prsd_ptr->next;
      }
    }

    rlparam_pack (rsd_ptr->data.rlparam, buf, bufsize, &position, comm);
    vector_pack  (rsd_ptr->data.source_vector, buf, bufsize, &position, comm);
    vector_pack  (rsd_ptr->data.dest_vector, buf, bufsize, &position, comm);
    vector_pack  (rsd_ptr->data.request_vector, buf, bufsize, &position, comm);
    vector_pack  (rsd_ptr->data.count_vector, buf, bufsize, &position, comm);
    vector_pack  (rsd_ptr->data.root_vector, buf, bufsize, &position, comm);
    #ifdef FEATURE_FILE_IO
    vector_pack  (rsd_ptr->data.disp_vector, buf, bufsize, &position, comm);
    #endif
#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
    for (int i = 0; i < TIMING_MAX; i++) {
      if (rsd_ptr->data.time_info[i].time_list) {
        timing_list_pack (rsd_ptr->data.time_info[i].time_list, buf, bufsize, &position, comm);
      }
    }
#endif

//Added by Karthik - Start
#ifdef FEATURE_FILE_IO
    int namelen = 0;
    if(rsd_ptr->data.filepath!=NULL && strlen(rsd_ptr->data.filepath)>0) {
      namelen = strlen(rsd_ptr->data.filepath)+1;
      PMPI_Pack(&namelen,1,MPI_INT,buf,bufsize,&position,comm);
      PMPI_Pack(rsd_ptr->data.filepath,namelen,MPI_CHAR,buf,bufsize,&position,comm);
    }
    else {
      PMPI_Pack(&namelen,1,MPI_INT,buf,bufsize,&position,comm);
    }
    namelen=0;
    if(rsd_ptr->data.filename!=NULL && strlen(rsd_ptr->data.filename)>0) {
      namelen = strlen(rsd_ptr->data.filename)+1;
      PMPI_Pack(&namelen,1,MPI_INT,buf,bufsize,&position,comm);
      PMPI_Pack(rsd_ptr->data.filename,namelen,MPI_CHAR,buf,bufsize,&position,comm);
    } 
    else {
      PMPI_Pack(&namelen,1,MPI_INT,buf,bufsize,&position,comm);
    }
    namelen=0;
    if(rsd_ptr->data.datarep!=NULL && strlen(rsd_ptr->data.datarep)>0) {
      namelen = strlen(rsd_ptr->data.datarep)+1;
      PMPI_Pack(&namelen,1,MPI_INT,buf,bufsize,&position,comm);
      PMPI_Pack(rsd_ptr->data.datarep,namelen,MPI_CHAR,buf,bufsize,&position,comm);
    }
    else {
      PMPI_Pack(&namelen,1,MPI_INT,buf,bufsize,&position,comm);
    }   
#endif
//Added by Karthik - End

    dep_graph_pack (&rsd_ptr->graph, buf, bufsize, &position, comm);

    rsd_ptr = rsd_ptr->next;
  }

  // finally, after all that packing, ship out the buffer.
  PMPI_Send(buf, position, MPI_PACKED, dest, 0, comm);

  free(buf);
}


void recv_rsd_queue(rsd_queue* queue, MPI_Comm comm, int src) {
  int bufsize = 0;                // size of received buffer
  void *buf;                      // buffer for MPI packing.
  int position = 0;               // position in the buffer (for MPI_Unpack)

  // receive size and allocate something large enough to hold the data when we get it.
  PMPI_Recv(&bufsize, 1, MPI_INT, src, 0, comm, MPI_STATUS_IGNORE);
  buf = malloc(bufsize);

  // now start building an rsd queue from what we receive
  init_rsd_queue(queue);

  PMPI_Recv(buf, bufsize, MPI_PACKED, src, 0, comm, MPI_STATUS_IGNORE);

  int total_rsds;
  PMPI_Unpack(buf, bufsize, &position, &total_rsds, 1, MPI_INT, comm);

  for (int rsd = 0; rsd < total_rsds; rsd++) {
    // create a new node and unpack it.
    rsd_node *rsd_ptr = (rsd_node*)malloc(sizeof(rsd_node));
    PMPI_Unpack(buf, bufsize, &position, rsd_ptr, sizeof(rsd_node), MPI_BYTE, comm);

    assert (rsd_ptr->data.op > umpi_MPI_op_min && rsd_ptr->data.op < umpi_MPI_op_max);

    /*Unpack Event structure*/
    event_unpack(&rsd_ptr->event, buf, bufsize, &position, comm);

    /* Stack signature */
    sig_unpack(&rsd_ptr->signature, buf, bufsize, &position, comm);

    /* rank list */
    if (rsd_ptr->rlist) {
      int num_ranks;
      PMPI_Unpack(buf, bufsize, &position, &num_ranks, 1, MPI_INT, comm);

      rsd_ptr->rlist = (int*)malloc(num_ranks * sizeof(int));
      PMPI_Unpack(buf, bufsize, &position, rsd_ptr->rlist, num_ranks, MPI_INT, comm);
    }

    UNPACK_ARR(buf, bufsize, position, rsd_ptr, comm, 1);
    UNPACK_ARR(buf, bufsize, position, rsd_ptr, comm, 2);
    UNPACK_ARR(buf, bufsize, position, rsd_ptr, comm, 3);
    UNPACK_ARR(buf, bufsize, position, rsd_ptr, comm, 4);

    /* prsd stack */
    if (rsd_ptr->prsd_info.head) {
      init_prsd_stack(&rsd_ptr->prsd_info);

      int histo_present;
      int prsd_size;
      PMPI_Unpack(buf, bufsize, &position, &prsd_size, 1, MPI_INT, comm);
      for (int p = 0; p < prsd_size; p++) {
        prsd_node *prsd_ptr = (prsd_node*)malloc(sizeof(prsd_node));
        PMPI_Unpack(buf, bufsize, &position, prsd_ptr, sizeof(prsd_node), MPI_BYTE, comm);
        // unpack the tail signature
        sig_unpack(&prsd_ptr->tail_sig, buf, bufsize, &position, comm);
	//Added by Karthik - Start
	PMPI_Unpack(buf, bufsize, &position, &histo_present, 1, MPI_INT, comm);
	if(histo_present)
	  prsd_ptr->prsd_stat = histogram_unpack(buf, bufsize, &position, comm);
	//Added by Karthik - End
        prsd_push_node_end(&rsd_ptr->prsd_info, prsd_ptr);
      }
    }

    rsd_ptr->data.rlparam = rlparam_unpack (buf, bufsize, &position, comm);
    rsd_ptr->data.source_vector = vector_unpack  ( buf, bufsize, &position, comm);
    rsd_ptr->data.dest_vector   = vector_unpack  ( buf, bufsize, &position, comm);
    rsd_ptr->data.request_vector= vector_unpack  ( buf, bufsize, &position, comm);
    rsd_ptr->data.count_vector = vector_unpack  (buf, bufsize, &position, comm);
    rsd_ptr->data.root_vector = vector_unpack  (buf, bufsize, &position, comm);
    #ifdef FEATURE_FILE_IO
    rsd_ptr->data.disp_vector= vector_unpack  ( buf, bufsize, &position, comm);
    #endif
#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
    for (int i = 0; i < TIMING_MAX; i++) {
      if (rsd_ptr->data.time_info[i].time_list) {
        timing_list_unpack (&rsd_ptr->data.time_info[i].time_list, buf,
            bufsize, &position, comm);
      }
    }
#endif /* FEATURE_TIMING_HISTO */

//Added by Karthik - Start
#ifdef FEATURE_FILE_IO
    int namelen = 0;
    PMPI_Unpack(buf, bufsize, &position, &namelen, 1, MPI_INT, comm);
    if(namelen>0) {
      rsd_ptr->data.filepath = (char*)malloc(namelen*sizeof(char));
      PMPI_Unpack(buf, bufsize, &position, rsd_ptr->data.filepath, namelen, MPI_CHAR, comm);
    }
    namelen = 0;
    PMPI_Unpack(buf, bufsize, &position, &namelen, 1, MPI_INT, comm);
    if(namelen>0) {
      rsd_ptr->data.filename = (char*)malloc(namelen*sizeof(char));
      PMPI_Unpack(buf, bufsize, &position, rsd_ptr->data.filename, namelen, MPI_CHAR, comm);
    }
    namelen = 0;
    PMPI_Unpack(buf, bufsize, &position, &namelen, 1, MPI_INT, comm);
    if(namelen>0) {
      rsd_ptr->data.datarep = (char*)malloc(namelen*sizeof(char));
      PMPI_Unpack(buf, bufsize, &position, rsd_ptr->data.datarep, namelen, MPI_CHAR, comm);
    }
#endif
//Added by Karthik - End
    dep_graph_unpack (&rsd_ptr->graph, buf, bufsize, &position, comm);

    // put the new rsd at the end of the queue.
    rsd_push_tail(queue, rsd_ptr);
  }
  dep_graph_index_to_ptr (queue->head);
 
  free(buf);
  //printf("recv_rsd_queue: %d:\n",get_my_rank());
  //printrsdqueue(queue);
}

void merge_graphs (rsd_node * master_iter, rsd_node * slave_iter) {
  list_t si = NULL, s0 = NULL, m0 = NULL;

  /* point all incoming edges to the master. move all reverse incoming edges to
   * the masters list of edges */
  foreach (si, slave_iter->graph.in_edges) {
    DEBUG (stderr, "in edge: %d :(%d,seq %d %p) \n",
            ((rsd_node*)(si->v))->node, ((rsd_node*)(si->v))->data.op, ((rsd_node*)(si->v))->data.seq_num, ((rsd_node*)(si->v))
        );
    foreach (s0, ((rsd_node*)(si->v))->graph.out_edges){
      if (s0->v == (void*)slave_iter) {
        DEBUG (stderr, "moved pointer for %d :(%d,seq %d %p) "
            "from %d : (%d, seq %d %p "
            "to %d : (%d, seq %d %p ""\n",
            ((rsd_node*)(si->v))->node, ((rsd_node*)(si->v))->data.op, ((rsd_node*)(si->v))->data.seq_num, ((rsd_node*)(si->v)),
            slave_iter->node, slave_iter->data.op, slave_iter->data.seq_num, slave_iter,
            master_iter->node, master_iter->data.op, master_iter->data.seq_num, master_iter
            );
        s0->v = (void*)master_iter;
      }
    }
    int already_exists = 0;
    foreach (m0, master_iter->graph.in_edges) {
      if (si->v == m0->v) {
        already_exists = 1;
      }
    }
    if (!already_exists)
      master_iter->graph.in_edges = list_insert (master_iter->graph.in_edges, si->v);
  }

  foreach (si, slave_iter->graph.out_edges)
  {
    foreach (s0, ((rsd_node*)(si->v))->graph.in_edges){
      if (s0->v == (void*)slave_iter) {
        s0->v = (void*)master_iter;
      }
    }

    int already_exists = 0;
    foreach (m0, master_iter->graph.out_edges) {
      if (si->v == m0->v) {
        already_exists = 1;
      }
    }
    if (!already_exists)
      master_iter->graph.out_edges = list_insert (master_iter->graph.out_edges, si->v);

  }
}

#define merge_nodes_internal(master_iter, slave_iter) do {              \
  master_iter->location = Master_Queue;                                 \
  merge_params (master_iter, slave_iter);                               \
  merge_recursive_lists(&master_iter->rlist, slave_iter->rlist);        \
  merge_timing (master_iter, slave_iter, slave_iter->node);             \
  merge_graphs (master_iter, slave_iter);                               \
} while (0)

rsd_node * merge_nodes (rsd_node * master_iter, rsd_node * slave_iter)
{
  rsd_node * last = NULL;
  rsd_node * loop_end = NULL;

  merge_nodes_internal (master_iter, slave_iter);

  /* unwind master & slave */
  if(master_iter->prsd_info.head) {
    //stack_sig tail = master_iter->prsd_info.head->tail_sig;
    loop_end = get_loop_end(master_iter, master_iter->prsd_info.head);
    //while(!SIG_MATCH(master_iter->signature, tail)) {
    while(master_iter != loop_end) {
      master_iter = master_iter->next;
      slave_iter = slave_iter->next;

      merge_nodes_internal (master_iter, slave_iter);
    }
  }

  last = slave_iter;
  return last;
}

/* get next element in the master queue
   note that we should skip loop members */
/*#define go_next(iter) do {\
        if(iter->prsd_info.head) {  \
          stack_sig tail = iter->prsd_info.head->tail_sig;  \
          while(!SIG_MATCH(iter->signature, tail))  \
            iter = iter->next;  \
        } \
        iter = iter->next;  \
} while (0);
*/
#define dep_next(node) (node->graph.out_edges != NULL ) ? node->graph.out_edges->v: NULL

list_t dep_list = NULL;

rsd_node * go_next(rsd_node *node) {
  int loop_mems = 0;
  rsd_node *iter = node;
  if(iter->prsd_info.head) {
    loop_mems = iter->prsd_info.head->loop_mems;
    for(int i = 1; i < loop_mems; i++) {
      iter = iter->next;
    }
  }
  iter = iter->next;
  return iter;
}

void dfs_visit (rsd_queue *master, rsd_node *master_iter, rsd_queue *slave, rsd_node * node)
{
  list_t ptr;

  node->graph.color = Gray;
  foreach (ptr, node->graph.out_edges)  {
    rsd_node * n = (rsd_node*) ptr->v;
    if (n->graph.color == White && n->location == Slave_Queue)
      dfs_visit (master, master_iter, slave, n);
  }
  node->graph.color = Black;
  DEBUG (stderr, "inserted: %d :(%d,seq %d %p) \n", 
            node->node, node->data.op, node->data.seq_num, node);
  dep_list = list_insert (dep_list, node);
}

void dfs (rsd_queue * master, rsd_node * master_iter, rsd_queue *slave, rsd_node * node)
{
  list_t ptr;
  foreach (ptr, node->graph.out_edges) {
    rsd_node * n = (rsd_node*) ptr->v;
    if (n->graph.color == White && n->location == Slave_Queue)
      dfs_visit (master, master_iter, slave, n);
  }
}

int yank (rsd_queue *master, rsd_node *master_iter, rsd_queue *slave)
{
  int count = 0;
  list_t ptr;
  foreach (ptr, dep_list)
  {
    rsd_node * node = (rsd_node*)ptr->v;
    DEBUG (stderr, "unspliced: %d :(%d,seq %d %p) \n", 
        node->node, node->data.op, node->data.seq_num, node);
    unsplice(slave, node, node);
    rsd_insert_pre(master, master_iter, node, node);
    node->location = Master_Queue;
    count++;
  }
  return count;
}

void clear_list ()
{
  list_free (&dep_list, NULL);
}

int compare_prsd_iters(prsd_node *mprsd, prsd_node *sprsd)
{
  while(mprsd && sprsd) {
    if(mprsd->loop_iters != sprsd->loop_iters)
      return FALSE;
    mprsd = mprsd->next;
    sprsd = sprsd->next;
  }
  return TRUE;
}

int check_match_and_merge(rsd_queue *master, rsd_node *master_iter, rsd_queue *slave, rsd_node *slave_iter) {
  int new_mem_count = 0;
  if( SIG_MATCH(master_iter->signature, slave_iter->signature) ) {
    dfs(master, master_iter, slave, slave_iter);
    new_mem_count = yank(master, master_iter, slave);
    clear_list();

    merge_nodes_internal(master_iter, slave_iter);

    unsplice(slave, slave_iter, slave_iter);
    return new_mem_count;

  } //end - if(SIG_MATCH...

  return -1;
}
#if 0
void find_dep_and_merge(rsd_queue *master, rsd_node *master_iter, rsd_queue *slave, rsd_node *slave_iter)
{
  dfs(master, master_iter, slave, slave_iter);
  yank(master, master_iter, slave);
  clear_list();

  merge_nodes_internal(master_iter, slave_iter);
}
#endif

int check_loop_collective(rsd_node *iter) {
  rsd_node *end;
  end = iter;
  //go_next(end);
  end = go_next(end);

  while(iter->data.seq_num != end->data.seq_num) {
    if(check_collective(iter) == TRUE) {
      return TRUE;
    }
    iter = iter->next;
  }
  return FALSE;
}

/*rsd_node * get_loop_end(rsd_node *node, prsd_node *prsd_iter) {
  rsd_node * ret_node = node;
  if(prsd_iter) {
    //stack_sig tail_sig = prsd_iter->tail_sig;
    while(!SIG_MATCH(prsd_iter->tail_sig, ret_node->signature)){
      ret_node = ret_node->next;
    }
  }
  return ret_node;
}*/


int append_rsd_nodes(rsd_queue *master, rsd_node *dest, rsd_queue *slave, rsd_node *slave_end, int is_pre) {
  int count = 0;
  rsd_node *start, *end, *iter;
  start = slave->head;
  iter = slave->head;
  end = slave_end->prev;
  while(iter != slave_end) {
    count++;
    iter = iter->next;
  }
  unsplice(slave, start, end);
  if(is_pre) {
    rsd_insert_pre (master, dest, start, end);
  } else {
    rsd_insert_post (master, dest, start, end);
  }
  return count;
}

void modify_prsd_tailsig(prsd_node *prsd, stack_sig *old_sig, stack_sig *new_sig) {

  prsd_node *iter = prsd;
  while(iter != NULL) {

    if(SIG_MATCH(iter->tail_sig, *old_sig) ) {
      sig_assign(&iter->tail_sig, new_sig);
    }
    iter = iter->prev;
  }
}

int merge_loops (rsd_queue *master, rsd_node *mnode, rsd_node *m_end, prsd_node *m_prsd, 
      rsd_queue *slave, rsd_node *snode, rsd_node *s_end, prsd_node *s_prsd ) {
  int new_memcnt = 0;
  rsd_node *new_mend, *new_send, *mstart, *sstart;
  rsd_node *mloop_next, *sloop_next;
  prsd_node *new_sprsd;
  mstart = mnode;
  sstart = snode;
  mloop_next = m_end->next;
  sloop_next = s_end->next;
  //Add code to copy slave->head to s_end
  //Update new_memcnt after copy
  
  /*if(slave->head != snode) {
    new_memcnt += append_rsd_nodes(master, mnode, slave, snode, 1);
    if(my_rank == 31)
      printf("merge_loops:after first append\n");
  }*/

  do {
    do {
      if(m_prsd && s_prsd && m_prsd->loop_iters == s_prsd->loop_iters ) {
        if( m_prsd->loop_mems > 1) {
          new_mend = get_loop_end(mnode, m_prsd);
          new_send = get_loop_end(snode, s_prsd);
          new_sprsd = s_prsd->next;
          //Remove sprsd from snode->prsd_info stack
          prsd_pop (&snode->prsd_info);//Chk if this works properly
          int extra_op_count = 0;
          rsd_node *inter_loop_next = new_send->next;
          if(slave->head != snode) {
            new_memcnt += append_rsd_nodes(master, mnode, slave, snode, 1);
          }
          extra_op_count = merge_loops (master, mnode, new_mend, m_prsd->next,
                                        slave, snode, new_send, new_sprsd);
          //if(my_rank == 18)
            //printf("after recur. merge_loops: mi %d, si %d, new_memcnt %d, extra_op %d\n", mnode->data.seq_num, snode->data.seq_num, new_memcnt,extra_op_count);
          if(slave->head != inter_loop_next) {
            //append after new_mend
            int app_cnt = 0;
            app_cnt = append_rsd_nodes(master, new_mend, slave, inter_loop_next, 0);
            //mprsd shld be modified with correct end signature
            //all mprsd->prev's shld be updated with count and signature
            extra_op_count += app_cnt;
            stack_sig *old_sig = &new_mend->signature;
            for(int i=0;i<app_cnt;i++)
              new_mend = new_mend->next;
            
            modify_prsd_tailsig(m_prsd, old_sig, &new_mend->signature);
          }
          //update new_memcnt
          m_prsd->loop_mems += extra_op_count;
          new_memcnt += extra_op_count;
          //if(my_rank == 18)
            //printf("recur. merge_loops b4 break: mi %d, si %d, new_memcnt %d, extra_op %d\n", mnode->data.seq_num, snode->data.seq_num, new_memcnt,extra_op_count);
          mnode = new_mend;
          break;
        } else {
          int cnt = check_match_and_merge(master, mnode, slave, snode);
          if(  cnt >= 0 ) {
            //update new_memcnt
            if(mnode != mstart) {
              new_memcnt += cnt;
            }
            //if(my_rank == 18)
              //printf("merge_loops else 1: mi %d, si %d, new_memcnt %d, cnt %d\n", mnode->data.seq_num, snode->data.seq_num, new_memcnt,cnt);
            break;
          }         
        }
      } else if(m_prsd) {
        if (mnode == mstart) {
          mnode = get_loop_end(mnode, m_prsd);
          break;
        }
      } else if(s_prsd) {
        snode = get_loop_end(snode, s_prsd);
        if (mnode == mstart) {
          mnode = get_loop_end(mnode, m_prsd);
          break;
        }
      } else {
        int cnt = check_match_and_merge(master, mnode, slave, snode);
        if(  cnt >= 0 ) {
          //update new_memcnt
          if(mnode != mstart){
            new_memcnt += cnt;
          }
            //if(my_rank == 18)
              //printf("merge_loops else 2: mi %d, si %d, new_memcnt %d, cnt %d\n", mnode->data.seq_num, snode->data.seq_num, new_memcnt,cnt);
          break;

        }
        if(mnode == mstart)
          break;
      }
      snode = snode->next;
      if(snode)
        s_prsd = snode->prsd_info.head;
      if(mnode->data.op != snode->data.op && check_collective(snode) == TRUE){
        break;
      }//TODO: Untested condition
    } while (snode && snode != sloop_next);
    mnode = mnode->next;
    if(mnode)
      m_prsd = mnode->prsd_info.head;
    snode = slave->head;
    s_prsd = snode->prsd_info.head;
  } while (mnode && mnode != mloop_next && snode != sloop_next);
  return new_memcnt;
}

int merge_prsd_events(rsd_queue *master, rsd_node *master_iter, rsd_queue *slave, rsd_node *slave_iter) {
  rsd_node *mend, *send, *nxt_loop;
  prsd_node *mprsd, *sprsd, *new_sprsd;
  mprsd = master_iter->prsd_info.head;
  sprsd = slave_iter->prsd_info.head;
  int new_mem_cnt = 0;
  if(mprsd && sprsd) {
    if(mprsd->loop_iters == sprsd->loop_iters && check_loop_collective(master_iter)
       && check_loop_collective(slave_iter) ) {
      //if(my_rank == 18)
        //printf("Rank: %d: Before loop end master: %d, slave: %d,slvhd %d\n", my_rank, master_iter->data.seq_num, slave_iter->data.seq_num, slave->head->data.seq_num);
      mend = get_loop_end(master_iter, mprsd);
      send = get_loop_end(slave_iter, sprsd);
      //printf("Rank: %d: After loop end master: %d, slave: %d\n", my_rank, master_iter->data.seq_num, slave_iter->data.seq_num);
      new_sprsd = sprsd->next;
      prsd_pop(&slave_iter->prsd_info);
      nxt_loop = send->next;

      if(slave->head != slave_iter) {
      //if(my_rank == 18)
        //printf("Rank: %d: Before append_rsd node master: %d, slave: %d, slvhd: %d\n", my_rank, master_iter->data.seq_num, slave_iter->data.seq_num, slave->head->data.seq_num);
        append_rsd_nodes (master, master_iter, slave, slave_iter, 1);
      }

      new_mem_cnt = merge_loops(master, master_iter, mend, mprsd->next,
                                slave, slave_iter, send, new_sprsd);
      //if(my_rank == 18)
        //printf("merge_prsd_events: mi %d, si %d, new_mem_cnt %d\n", master_iter->data.seq_num, slave_iter->data.seq_num, new_mem_cnt);

      if(slave->head != nxt_loop) {
        int app_cnt = 0;
        app_cnt = append_rsd_nodes(master, mend, slave, nxt_loop, 0);
        new_mem_cnt += app_cnt;
        stack_sig *old_sig = &mend->signature;
        for(int i=0; i<app_cnt; i++)
          mend = mend->next;
        modify_prsd_tailsig(mprsd, old_sig, &mend->signature);
      }
      mprsd->loop_mems += new_mem_cnt;
      //if(my_rank == 18)
        //printf("merge_prsd_events b4 return: mi %d, si %d, new_mem_cnt %d\n", master_iter->data.seq_num, slave_iter->data.seq_num, new_mem_cnt);
      return TRUE;
    }
  }
  return FALSE;
}

#if 0
int merge_prsd_events(rsd_queue *master, rsd_node *master_iter, rsd_queue *slave, rsd_node *slave_iter, int skip_coll_check)
{
  /*This function merges all possible events inside a prsd.
    If master_iter is not a PRSD, merge is not done.
  */
  prsd_node *mprsd_iter, *sprsd_iter;
  rsd_node *master_start, *slave_start, *master_end, *slave_end;
  stack_sig *mtail, *msig, *stail, *ssig;
  int sig_match = 0, can_proceed = 0;

  if(SIG_MATCH(master_iter->signature, slave_iter->signature) &&
      (master_iter->prsd_info.head || slave_iter->prsd_info.head)) 
      //master_iter->prsd_info.head && slave_iter->prsd_info.head) 
      //&& get_depth(master_iter->prsd_info.head) == get_depth(slave_iter->prsd_info.head)) 
  {
    sig_match = 1;
    can_proceed = 1;
  } else {
    if(skip_coll_check || 
        check_loop_collective (master_iter) && check_loop_collective (slave_iter) ) {
      can_proceed = 1;
    }
  }
  
  if(can_proceed ) {
  
    master_end = master_iter;
    go_next(master_end);
    slave_end = slave_iter;
    go_next(slave_end);
    mprsd_iter = master_iter->prsd_info.head;
    sprsd_iter = slave_iter->prsd_info.head;
     
    /*No point in merging collectives if loop iterations are different.
      Collectives couldn't be used in the case of different iterations
      in different nodes
    */
    if(compare_prsd_iters(mprsd_iter, sprsd_iter) == FALSE)
      return FALSE;

    if(master_end)
      mtail = &master_end->signature;
    else
      mtail = NULL;

    if(slave_end)
      stail = &slave_end->signature;
    else
      stail = NULL;

    /*Copy all slave events before the slave_iter PRSD into master*/
    if(slave->head != slave_iter) {
      rsd_node* start = slave->head;
      rsd_node* end = slave_iter->prev;
      unsplice(slave, start, end);
      rsd_insert_pre(master, master_iter, start, end);
      //TODO: check graph connections
    }
    if(sig_match) {
      find_dep_and_merge (master, master_iter, slave, slave_iter);
      unsplice(slave, slave_iter, slave_iter);
      master_iter = master_iter->next;
    } else {
      slave_iter->rsd_info = 0;
      slave_iter->rsd_info = MEMBER_RSD;
      slave_iter->prsd_info.head = NULL;
      slave_iter->prsd_info.tail = NULL;
      rsd_node *start = slave_iter;
      unsplice(slave, start, start);
      master_iter = master_iter->next;
      rsd_insert_pre(master, master_iter, start, start);
    }

    //while(mtail && master_iter && !SIG_MATCH(master_iter->signature, *mtail) && (master_iter->data.seq_num != master_end->data.seq_num) ) {
    while(mtail && master_iter && (master_iter->data.seq_num != master_end->data.seq_num) ) {
      slave_iter = slave->head;

      while(stail && slave_iter && (slave_iter->data.seq_num != slave_end->data.seq_num) ) {
        if(check_match_and_merge(master, master_iter, slave, slave_iter) == TRUE ) {
          break;
        }
        slave_iter = slave_iter->next;       

      } //end - while(stail....
      //master_iter = master_iter->next;//Karthik - bug fix for moving to next event after PRSD
      go_next(master_iter);
    } //end - while(mtail...
    //master_iter = master_iter->prev;
    if(SIG_MATCH(master_iter->signature, *mtail) && (master_iter->data.seq_num == master_end->data.seq_num) && !SIG_MATCH(slave->head->signature, *stail)) {
      rsd_node* start = slave->head;
      rsd_node* end = slave_end->prev;
      unsplice(slave, start, end);
      rsd_insert_pre(master, master_iter, start, end);
    }
    
    return TRUE;
  }
  return FALSE;
}
#endif

int check_collective(rsd_node * node) {
  switch(node->data.op) {
    case umpi_MPI_Allgather:
    case umpi_MPI_Allgatherv:
    case umpi_MPI_Allreduce:
    case umpi_MPI_Alltoall:
    case umpi_MPI_Alltoallv:
    case umpi_MPI_Barrier:
    case umpi_MPI_Bcast:
    case umpi_MPI_Gather:
    case umpi_MPI_Gatherv:
    case umpi_MPI_Reduce:
    case umpi_MPI_Reduce_scatter:
    case umpi_MPI_Scatter:
    case umpi_MPI_Scatterv:
    case umpi_MPI_File_open:
    case umpi_MPI_File_write_all:
    case umpi_MPI_File_read_all:
    case umpi_MPI_File_close:
      return TRUE;
    default:
      return FALSE;
    
  } 
}

void merge_queues(rsd_queue *master, rsd_queue *slave, int slave_rank)
{
  rsd_node *master_iter, *slave_iter;
  rsd_node * dnext, * dprev;
  rsd_node * start, * end;
  rsd_node * insert;

  master_iter = master->head;

    //dump_queue_as_graph (master, get_my_rank(), slave_rank, "premaster");
    //dump_queue_as_graph (slave, get_my_rank(), slave_rank, "preslave");
  int i = 0;
  while(master_iter) {
    slave_iter = slave->head;
    /*if(my_rank == 25 && master_iter->data.seq_num == 87409 && slave_iter->data.seq_num == 89809) {
      dump_queue_with_label (master, get_my_rank(), i, "qmaster");
      dump_queue_as_graph (master, get_my_rank(), i, "master");
      dump_queue_with_label (slave, get_my_rank(), i, "qslave");
      dump_queue_as_graph (slave, get_my_rank(), i, "slave");
      exit(1);
    }*/
    while(slave_iter) {
        DEBUG (stderr, "trying to match master %d :(%d,seq %d %p) slave %d:(%d,seq %d %p)\n", 
            master_iter->node, master_iter->data.op, master_iter->data.seq_num, master_iter,
            slave_iter->node, slave_iter->data.op, slave_iter->data.seq_num, slave_iter);
      /*printf("trying to match master %d :(%d,seq %d %p) slave %d:(%d,seq %d %p)\n", 
            master_iter->node, master_iter->data.op, master_iter->data.seq_num, master_iter,
            slave_iter->node, slave_iter->data.op, slave_iter->data.seq_num, slave_iter);
*/


      if(match_sequence(master_iter, slave_iter)) {

        DEBUG (stderr, "match!\n");
        /* build a list of dependencies */
        dfs (master, master_iter, slave, slave_iter);
        yank (master, master_iter, slave);
        clear_list ();

        start = slave_iter;
        end = merge_nodes (master_iter, slave_iter);
        unsplice (slave, start, end);
        //TODO: Possible memory leak, start to end nodes are not freed
        break;
      } else {
       #ifdef FEATURE_COMPRESS_HISTO
        if(master_iter->data.op != slave_iter->data.op && check_collective(slave_iter) == TRUE ) {
          break;
        }
        if(merge_prsd_events(master, master_iter, slave, slave_iter) == TRUE){
          /*start = slave_iter;
          go_next(slave_iter);
          end = slave_iter->prev;
          unsplice(slave, start, end);
          dump_queue_with_label (master, get_my_rank(), master_iter->data.seq_num, "qmaster");
          dump_queue_with_label (slave, get_my_rank(), slave_iter->data.seq_num, "qslave");*/
          break;
        } else {
          slave_iter = go_next (slave_iter);
        }
       #else
        slave_iter = go_next (slave_iter);
       #endif
      }
    }
#ifdef DEBUG_MERGE
   if(master_iter->node == 1 && slave_iter && slave_iter->node == 5) {
    dump_queue_with_label (master, get_my_rank(), i, "qmaster");
    dump_queue_with_label (slave, get_my_rank(), i, "qslave");
    dump_queue_as_graph (master, get_my_rank(), i, "master");
    dump_queue_as_graph (slave, get_my_rank(), i++, "slave");
   }
#endif
    master_iter = go_next (master_iter);
  }

  /* unwinding has finished master queue 
     but slave queue still has remaining ops */
  if(!master_iter && slave->head) {
    rsd_node* start = slave->head;
    rsd_node* end = slave->tail;

    unsplice(slave, start, end);
    rsd_insert_post(master, master->tail, start, end);
  }
    //dump_queue_as_graph (master, get_my_rank(), slave_rank, "master");
    //dump_queue_as_graph (slave, get_my_rank(), slave_rank, "slave");
}


/* EOF */
