#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include "prsd_traverse.h"
#include "rsd_reader.h"
#include "op_generator.h"
#include "trace.h"
#include "utils.h"
#include "mpi_util.h"

//#define DEBUG 0
var_dump_t * var;
derived_type_dump_t * type;
FILE* skel_type_fh;
file_handle_dump_t* fh_dump;
/**
 * Visitor routine for executing replay operations.  This is called
 * by the traversal routines to translate ops to actual MPI function 
 * calls.
 */
static int code_generator(rsd_node *node, int rank, int depth, int iter, int
    last, FILE * fh) {
  return generate_prsd_code(node, rank, depth, iter, fh);
}

void generate_max_size_function(FILE* skel_type_fh, derived_type_dump_t* type) {
  int i=0;
  fprintf(skel_type_fh, "\nint get_max_size() {\n");
  fprintf(skel_type_fh, "\tint value = 0;\n\tMPI_Aint size = 0;\n");

  if(type != NULL && type->total_count > 0) {
    for(i=0;i<type->total_count;i++) {
      if(type->type_usage_count[i]>0) {
        fprintf(skel_type_fh, "\tMPI_Type_extent(%s,&size);\n",type->type_name[i]);
        fprintf(skel_type_fh, "\tif(value < (%d * size)) value += %d*size;\n", type->type_usage_count[i],type->type_usage_count[i]);
      }
    }
  }

  fprintf(skel_type_fh, "\treturn value;\n");
  fprintf(skel_type_fh, "}\n");

  fprintf(skel_type_fh, "\nvoid skel_sleep(long long sec, long long nsec) {\n");
  fprintf(skel_type_fh, "struct timespec ts;\n");
  fprintf(skel_type_fh, "ts.tv_sec = sec;\n");
  fprintf(skel_type_fh, "ts.tv_nsec = nsec;\n");
  fprintf(skel_type_fh, "nanosleep(&ts, NULL);\n");
  fprintf(skel_type_fh, "}\n");
}

int main(int argc, char *argv[])
{
  int rank = -1, len = -1, whole, frac;
  int retval = -1;
  int per_node_trace = FALSE;
  struct timeval t_start, t_stop;
  struct timezone tz;
  rsd_queue *prsd_queue;
  char * dir = 0;
  char format = 'x';
  FILE * skel_fh = NULL;
  
  var = init_var_dump();
  type = init_derived_type_dump();
  fh_dump = init_fh_dump();
  /* determine input type */
  if(argc < 2) {
    fprintf(stderr, "usage: %s -b|-a|-p <trace dir>\n", argv[0]);
    return 0;
  }
  while (1)
  {
    retval = getopt (argc, argv, "p:X");

    if (retval == -1)
      break;
    switch (retval)
    {
      case 'p':
        format = 'p';
        dir = optarg;
        break;
      case 'X':
        per_node_trace = TRUE;
        break;
    }
  }
  if (format == -1)
  {
    fprintf(stderr, "error: must specify the input file's format w/ -p\n");
    return 0;
  }

  /* read in op trace for each node */
  char filename[4096];
  sprintf (filename, "%s/%d", dir, (per_node_trace ? rank : 0));
  prsd_queue = read_rsd_file(filename);
  len = rsd_get_length(prsd_queue);

  /* playback ops in queue */
  gettimeofday(&t_start, &tz);

  /*skeleton_type.c code will contain methods to create all
   *derived data types. Also this file will have method to 
   *calculate the max buffer size to be used for all calls
   *like Send, Write, etc.
   */
  skel_type_fh = fopen("skeleton_type.c", "w");
  fprintf(skel_type_fh, "#include \"skeleton_code.h\"\n");
  fprintf(skel_type_fh, "#include <time.h>\n");
  fprintf(skel_type_fh, "void create_derived_type() {\n");

  /*skeleton_code.c will have main function*/
  skel_fh = fopen("skeleton_code.c", "w");
  fprintf(skel_fh, "#include <stdio.h>\n");
  fprintf(skel_fh, "#include <stdlib.h>\n");
  fprintf(skel_fh, "#include <string.h>\n");
  fprintf(skel_fh, "#include <mpi.h>\n");
  fprintf(skel_fh, "#include \"skeleton_code.h\"\n");
  fprintf(skel_fh, "int main(int argc, char *argv[]) {\n\tDEFINE_LOCAL\n");
  fprintf(skel_fh, "\tDEFINE_FH\n");
  fprintf(skel_fh, "\tint buffersize = 0;\n");
  fprintf(skel_fh, "\tvoid* buffer = NULL;\n");
  fprintf(skel_fh, "\tvoid* buffer2 = NULL;\n");
  fprintf(skel_fh, "\tstruct timeval tv, tv1;\n");
  fprintf(skel_fh, "\tgettimeofday(&tv, NULL);\n");
#if DEBUG
  if (rank == DEBUG)  output_rsd_queue (stdout, prsd_queue);
  fprintf (stdout, "output rsd queue\n");
  fflush (stdout);
#else
  traverse_queue_for_generation(prsd_queue->head, rank, code_generator, skel_fh);
#endif

  /* stop timer */
  gettimeofday(&t_stop, &tz);
  whole = t_stop.tv_sec - t_start.tv_sec;
  frac = t_stop.tv_usec - t_start.tv_usec;
  if(frac < 0) {
    frac *= -1;
    whole --;
  }
  if(rank == 0)
    printf("execution time:  %d.%d\n", whole, frac);
  fprintf(skel_fh, "\tif(buffersize > 0) {\n\tfree(buffer);\n\tfree(buffer2);\n\t}\n");
    fprintf(skel_fh, "\tgettimeofday(&tv1, NULL);\n");
  fprintf(skel_fh,"}\n");
  fclose(skel_fh);
  fprintf(skel_type_fh, "}\n");
  generate_max_size_function(skel_type_fh, type);
  fclose(skel_type_fh);

  skel_fh = fopen("skeleton_code.h", "w");
  fprintf(skel_fh, "#include <mpi.h>\n");
  /*Add all local variable in the header file - Start*/
  fprintf(skel_fh, "#define DEFINE_LOCAL ");
  printf("Value of var is %p\n",var);
  if(var != NULL && var->total_count>0) {
    printf("Value of total count is %d\n", var->total_count);
    for(len = 0; len < var->total_count-1; len++) {
      if(var->is_global[len] == 0)
        fprintf(skel_fh, "\t%s \\\n",var->var_detail[len]);
    }
    if(var->is_global[var->total_count-1] == 0)
      fprintf(skel_fh, "\t%s\n", var->var_detail[var->total_count-1]);
  }
  /*Add all local variable in the header file - End*/

  /*Add all local variable in the header file - Start*/
  fprintf(skel_fh, "\n#define DEFINE_FH ");
  if(fh_dump != NULL && fh_dump->total_count>0) {
    for(len = 0; len < fh_dump->total_count-1; len++) {
      fprintf(skel_fh, "\t%s \\\n",fh_dump->fh_detail[len]);
    }
    fprintf(skel_fh, "\t%s\n", fh_dump->fh_detail[fh_dump->total_count-1]);
  }
  /*Add all local variable in the header file - End*/

  /*Add all global variable in the header file - Start*/
  fprintf(skel_fh, "\n/*All global variables used*/\n");
  if(var != NULL && var->total_count>0) {
    for(len = 0; len < var->total_count-1; len++) {
      if(var->is_global[len] == 1)
        fprintf(skel_fh, "\t%s\n",var->var_detail[len]);
    }
    if(var->is_global[var->total_count-1] == 1)
    fprintf(skel_fh, "\t%s\n", var->var_detail[var->total_count-1]);
  }
  /*Add all global variable in the header file - End*/

  /*Add all derived type init details in the header file - Start*/
  fprintf(skel_fh, "\n/*Initialization of all derived types used*/\n");
  if(type!=NULL && type->total_count>0) {
    for(len = 0; len < type->total_count-1; len++) {
      if(type->is_derived_type[len] == 1)
        fprintf(skel_fh, "\t%s\n", type->type_detail[len]);
    }
    if(type->is_derived_type[type->total_count-1] == 1)
      fprintf(skel_fh, "\t%s\n", type->type_detail[type->total_count-1]);
  }
  /*Add all derived type init details in the header file - End*/

  fclose(skel_fh);
  free_var_dump(var);
  free_derived_type_dump(type);
  free_fh_dump(fh_dump);
  return 0;
}
/* EOF */
