#include "op_generator.h"
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>
#include "mpi_util.h"

#include "opstruct.h"
#include "trace.h"
#include "ptr_handler.h"
#include "req_handler.h"
#include "mem_handler.h"
#ifdef FEATURE_TIMING_HISTO
#include "Histogram.h"
#include "TimeList.h"
#include "busy_wait.h"
#endif // FEATURE_TIMING_HISTO

//# define INLINE inline
#define INLINE
#define LOCAL 0
#define GLOBAL 1
#define DEFINED 0
#define DERIVED 1
//# define busy_wait nanosleep

/* Prints an error message that the operation is unimplemented,
 * then exits.
 */
static void unimplemented_error(const char *op) {
  fprintf(stderr, "ERROR: %s not implemented\n", op);
  exit(0); 
}

/* Prints a warning message that the op is unimplemented, but 
 * continues execution as normal.
 */
static void unimplemented_warning(const char *op) {
  fprintf(stderr, "WARNING: %s not implemented.  Continuing without executing.\n", op);
}



#ifdef FEATURE_TIMING_HISTO

long int get_compute_time  ();
int      set_compute_time  ();
int      get_sleep_time    (rsd_node *, int, rsd_node *, int);
#else

#define simulate_compute_time(a, b, c)
#define set_compute_time()

#endif // FEATURE_TIMING_HISTO

int RECORDED_MPI_REQUEST_NULL;
ptr_handler_t replay_ptr;

req_handler_t replay_req;

extern var_dump_t * var;
extern derived_type_dump_t * type;
int new_type_tracker;
extern FILE* skel_type_fh;
extern file_handle_dump_t* fh_dump;
int comm_total;

//#define PRSD_DEBUG
#ifdef PRSD_DEBUG
extern FILE *replay_dbg_dump;
#endif //PRSD_DEBUG

void update_op_buf_size (rsd_node * node);

//#define FEWER_MALLOCS
#ifdef FEWER_MALLOCS

int nmalloc, nrealloc, nnone;

#define Op_xbuf_alloc(Op, xbuf, type, size, buf)  do  {     \
    if (Op##xbuf##_size && Op##xbuf##_size >= (size))       \
    {                                                       \
      buf = (type*) Op##xbuf;                               \
      nnone++;                                              \
    }                                                       \
    else if (Op##xbuf##_size == 0)                          \
    {                                                       \
      Op##xbuf##_size = (size);                             \
      buf = Op##xbuf = (type*) malloc ((size));             \
      nmalloc++;                                            \
    }                                                       \
    else if (Op##xbuf##_size < (size))                      \
    {                                                       \
      Op##xbuf##_size = (size);                             \
      buf = Op##xbuf = (type*) realloc (Op##xbuf, (size));  \
      nrealloc++;                                           \
    }                                                       \
  } while (0)

#define DECLARE_XBUF(Op, xbuf)                  \
  void * Op##xbuf;                              \
  int    Op##xbuf##_size; 

#define DECLARE_SENDBUF(Op)  DECLARE_XBUF(Op, sendbuf)
#define DECLARE_RECVBUF(Op)  DECLARE_XBUF(Op, recvbuf)

#define FREE_XBUF(Op, xbuf)                     \
  free (Op##xbuf);                              \
  Op##xbuf = 0;                                 \
  Op##xbuf##_size = 0;

#define FREE_SENDBUF(Op)  FREE_XBUF(Op, sendbuf)
#define FREE_RECVBUF(Op)  FREE_XBUF(Op, recvbuf)

DECLARE_SENDBUF (Allgather)
  DECLARE_SENDBUF (Allgatherv)
  DECLARE_SENDBUF (Allreduce)
  DECLARE_SENDBUF (Alltoall)
  DECLARE_SENDBUF (Alltoallv)
  DECLARE_SENDBUF (Bcast)
  DECLARE_SENDBUF (Gather)
  DECLARE_SENDBUF (Reduce)
  DECLARE_SENDBUF (Send)
 
  DECLARE_RECVBUF (Allgather)
  DECLARE_RECVBUF (Allgatherv)
  DECLARE_RECVBUF (Allreduce)
  DECLARE_RECVBUF (Alltoall)
  DECLARE_RECVBUF (Alltoallv)
  DECLARE_RECVBUF (Gather)
  DECLARE_RECVBUF (Recv)
  DECLARE_RECVBUF (Reduce)

#define FREE(a)

#define Free_sendrecv_bufs  do  {               \
    FREE_SENDBUF (Allgather)                    \
      FREE_SENDBUF (Allgatherv)                 \
      FREE_SENDBUF (Allreduce)                  \
      FREE_SENDBUF (Alltoall)                   \
      FREE_SENDBUF (Alltoallv)                  \
      FREE_SENDBUF (Bcast)                      \
      FREE_SENDBUF (Gather)                     \
      FREE_SENDBUF (Reduce)                     \
      FREE_SENDBUF (Send)                       \
                                                \
      FREE_RECVBUF (Allgather)                  \
      FREE_RECVBUF (Allgatherv)                 \
      FREE_RECVBUF (Allreduce)                  \
      FREE_RECVBUF (Alltoall)                   \
      FREE_RECVBUF (Alltoallv)                  \
      FREE_RECVBUF (Gather)                     \
      FREE_RECVBUF (Recv)                       \
      FREE_RECVBUF (Reduce)                     \
      } while (0)

#define Print_sendrecv_alloc_stats                                      \
  fprintf (stderr, "nnone = %d, nmalloc = %d, nrealloc = %d\n", nnone,  \
           nmalloc, nrealloc);
#else

#define Op_xbuf_alloc(Op, xbuf, type, size, buf)  \
  buf = (type*) malloc ((size))

#define FREE free

#define Free_sendrecv_bufs
#define Print_sendrecv_alloc_stats

#endif // FEWER_MALLOCS

#define Op_sendbuf_alloc(Op, type, size, buf)   \
  Op_xbuf_alloc(Op, sendbuf, type, size, buf)

#define Op_recvbuf_alloc(Op, type, size, buf)   \
  Op_xbuf_alloc(Op, recvbuf, type, size, buf)


  rsd_node * g_cur_op, * g_prev_op;
void set_cur_op (rsd_node * op)
{
  g_prev_op = g_cur_op;
  g_cur_op = op;
}

#ifdef FEATURE_TIMING_HISTO
int prsd_sleep (FILE* fh, int sleep_duration)
{
  struct timespec tptr, dummy;

  // enable printf only for debugging on small number of nodes.
  tptr.tv_sec = sleep_duration/1000000;
  tptr.tv_nsec = 1000*(sleep_duration - tptr.tv_sec*1000000);
  printf ("Comp.:(%d) Sleep for %d.%09d s\n", my_rank, 
    tptr.tv_sec, tptr.tv_nsec);
  dummy.tv_nsec = dummy.tv_sec = 0;
  fprintf(fh, "\t\tskel_sleep(%lld, %lld);\n", tptr.tv_sec, tptr.tv_nsec );
  if(busy_wait (&tptr, &dummy) < 0)
    perror("busy_wait");

  return 0;
}


int leftover = 0;
void simulate_compute_time(FILE* fh, rsd_node *op, int i) {
  int sleep_duration = -1;
  
  sleep_duration = get_sleep_time (op, TIMING_PRE, g_prev_op, i);
  printf ("Comp.:(%d) op is %d, get_sleep_time: sleep duration %d\n", my_rank, op->data.op, sleep_duration);
//  sleep_duration -= get_compute_time ();
//  printf ("Comp.:(%d) op is %d, after get_compute_time: sleep duration %d\n", my_rank, op->data.op, sleep_duration);
//  sleep_duration += leftover;
//  leftover = 0;
  if (sleep_duration > 0)
  {
    printf ("Comp.:(%d) op is %d, sleep duration %d\n", my_rank, op->data.op, sleep_duration);
    prsd_sleep (fh, sleep_duration);
  }
  else
  {
    leftover = sleep_duration;
    /*
      if (sleep_duration>10000)
      fprintf (stderr, "(%d) compute op:=%d Overshot by %d.%d ms \n", my_rank,
      op->data.op, (float)sleep_duration/1000, sleep_duration%1000);
    */
  }
}

int get_sleep_time (rsd_node * op, int i, rsd_node * prev, int j) {
  int sleep_duration = 0;
  void * h = 0;

  if (op->data.time_info[i].time_list)
  {
    if (op->data.replay_prev_op != prev)
    {
      if (!prev || timing_list_size (op->data.time_info[i].time_list) == 1) {
        h = timing_list_begin (op->data.time_info[i].time_list);
      } else  {
        h = timing_find_match (op->data.time_info[i].time_list, &prev->signature);
      }

      assert(h);

      op->data.replay_prev_op = prev;
      op->data.replay_histo = h;

    }
    else
    {
      h = op->data.replay_histo;
    }
    if (h)
    {
      double min, max;
      int min_node, max_node;
      histogram_min_max (h, &min, &max, &min_node, &max_node);
      if (histogram_has_other (h, my_rank))
      {
        if (my_rank == max_node)
          sleep_duration = max;
        else if (my_rank == min_node)
          sleep_duration = min;
        else
          sleep_duration = (int) histogram_value (h, my_rank, -1);

        if (sleep_duration == -1)
        {
          output_rsd_node (stderr, op);
        }
      }
      else
      {
        sleep_duration = (int) histogram_value (h, my_rank, j);
      }
    }
  }
  else
  {
    sleep_duration = op->data.time_info[i].time;
  }
  if (sleep_duration < 0)
  {
    fprintf (stderr, "[%s:%d] Invalid sleep_duration\n", __FILE__, __LINE__);
    exit (0);
  }

  return sleep_duration;
}

#endif // FEATURE_TIMING_HISTO

void update_op_buf_size (rsd_node * node)
{
#ifdef FEWER_MALLOCS
  int size = 0;

  if (node && node->data.datatype != -1 && node->data.datatype != 0)
  {
    MPI_Datatype datatype = index_to_type (replay_ptr, node->data.datatype);
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS)
    {
      fprintf(stderr, "error: MPI_Type_extent (%d,%p) failed - probably invalid" \
              "type\n", node->data.datatype, datatype);
      fflush (stderr);
      exit(0);
    }
    assert (size);
    size *= node->data.count;
  }
  switch (node->data.op)
  {
  case umpi_MPI_Send:
    Sendsendbuf_size = (Sendsendbuf_size > size) ?
      Sendsendbuf_size : size;
    break;
  case umpi_MPI_Recv:
    Recvrecvbuf_size = (Recvrecvbuf_size > size) ?
      Recvrecvbuf_size : size;
    break;
  default:
    break;
  }
#endif // FEWER_MALLOCS
}

/*
 * populate_datatype - Method to populate derived_type_dump_t with datatype used in
 * particular MPI Call.
 */
void populate_datatype(derived_type_dump_t* type, rsd_node *op, char* typename)
{
    int i = 0;
    MPI_Datatype datatype = 0;
    //char typename[100];
    //memset(typename, 0, 100);
    //Get datatype from index
    datatype = index_to_type (replay_ptr, op->data.datatype);

    //Check if the datatype is part of defined datatype
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }
    
    //If typename is null, then get the name from already populated derived datatype
    if(strlen(typename)==0) {
      strcpy(typename, get_type_name_from_index(type, op->data.datatype));
    }
    else {
      //Add defined type in the type struct for printing purpose
      if(add_defined_type(type, typename, op->data.datatype)) {
        fprintf(stderr,"Write buffer to hold types is possibly full.");
        exit(0);
      }
    }
    //Update usage count for allocating buffer
    update_usage_count(type, op->data.count, op->data.datatype);
}

int generate_prsd_code(rsd_node *op, int rank, int depth, int iter, FILE * skel_fh)
{
  int *sendelem, *recvelem ;
  int *sdisp, *rdisp;
  MPI_Op mpi_op;
  MPI_Group group, group_out;
  MPI_Comm comm, comm_out, peer_comm;
  MPI_Status status, *statuses;
  MPI_Request request, *requests;
  MPI_Datatype newtype = 0, datatype = 0, recvtype = 0, * datatypes = 0;
  void *sendbuf = 0, *recvbuf = 0;
  MPI_Aint size = 0;
  int cnt_total, comm_size, ret = MPI_ERR_OTHER, oc, cur_oc, *indices;
  int src = -1, dest = -1;
  int sendmax = -1, recvmax = -1, si = -1, ri = -1;
  MPI_Info info;
  MPI_File file;
  int disp = -1;
  int i = 0;
  char commname[100];
  char typename[100];
  char* new_typename;
  char* new_fh;
  char statusname[100];
  char statusdet[100];
  char opname[100];
  int replay_type_index;
  set_cur_op (op);

  memset(commname, 0, 100);
  memset(typename, 0, 100);
  memset(statusname, 0, 100);
  memset(statusdet, 0, 100);
  memset(opname, 0, 100);
#ifdef PRSD_DEBUG
  if (!replay_dbg_dump) {
    char buf[20];
    sprintf (buf, "debug/%d", rank);
    replay_dbg_dump = fopen (buf, "w");
  }
  output_rsd_node (replay_dbg_dump, op);
#endif // PRSD_DEBUG

  switch(op->data.op) {
  case umpi_MPI_Abort: 
    unimplemented_error("MPI_Abort");
    break;

  case umpi_MPI_Address: 
    unimplemented_error("MPI_Address");
    break;

  case umpi_MPI_Allgather:
    unimplemented_error("MPI_Allgather");
    break;

  case umpi_MPI_Allgatherv:
    unimplemented_error("MPI_Allgatherv");
    break;

  case umpi_MPI_Allreduce:
    i=0;
    comm = index_to_comm(replay_ptr, op->data.comm);
    while(MPI_DEFINED_COMMS[i].comm != MPI_COMM_NULL) {
      if(MPI_DEFINED_COMMS[i].comm == comm) {
        strcpy(&commname, MPI_DEFINED_COMMS[i].comm_name);
        break;
      }
      i++;
    }

    i = 0;
    datatype = index_to_type (replay_ptr, op->data.datatype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }
    if(strlen(typename)==0) {
      strcpy(&typename, get_type_name_from_index(type, op->data.datatype));
    }
    else {
      if(add_defined_type(type, typename, op->data.datatype)) {
        fprintf(stderr,"Write buffer to hold types is possibly full.");
        exit(0);
      }
    }

    update_usage_count(type, op->data.count, op->data.datatype);

    i=0;
    mpi_op = index_to_op (replay_ptr, op->data.mpi_op);
    while(MPI_DEFINED_OPS[i].op != MPI_OP_NULL) {
      if(MPI_DEFINED_OPS[i].op == mpi_op) {
        strcpy(&opname, MPI_DEFINED_OPS[i].op_name);
        break;
      }
      i++;
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_Allreduce(buffer, buffer2, %d, %s, %s, %s);\n",
              op->data.count, typename, opname, commname);
    set_compute_time();
    break;

  case umpi_MPI_Alltoall:
    unimplemented_error("MPI_Alltoall");
    break;

  case umpi_MPI_Alltoallv: 
    unimplemented_error("MPI_Alltoallv");
    break;

  case umpi_MPI_Attr_delete: 
    unimplemented_error("MPI_Attr_delete");
    break;

  case umpi_MPI_Attr_get: 
    unimplemented_error("MPI_Attr_get");
    break;

  case umpi_MPI_Attr_put: 
    unimplemented_error("MPI_Attr_put");
    break;

  case umpi_MPI_Barrier: 
    i=0;
    comm = index_to_comm(replay_ptr, op->data.comm);
    while(MPI_DEFINED_COMMS[i].comm != MPI_COMM_NULL) {
      if(MPI_DEFINED_COMMS[i].comm == comm) {
        strcpy(&commname, MPI_DEFINED_COMMS[i].comm_name);
        break;
      }
      i++;
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_Barrier(%s);\n",commname);
    set_compute_time();
    break;

  case umpi_MPI_Bcast: 
    i=0;
    comm = index_to_comm(replay_ptr, op->data.comm);
    while(MPI_DEFINED_COMMS[i].comm != MPI_COMM_NULL) {
      if(MPI_DEFINED_COMMS[i].comm == comm) {
        strcpy(&commname, MPI_DEFINED_COMMS[i].comm_name);
        break;
      }
      i++;
    }

    i = 0;
    datatype = index_to_type (replay_ptr, op->data.datatype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }
    if(strlen(typename)==0) {
      strcpy(&typename, get_type_name_from_index(type, op->data.datatype));
    }
    else {
      if(add_defined_type(type, typename, op->data.datatype)) {
        fprintf(stderr,"Write buffer to hold types is possibly full.");
        exit(0);
      }
    }
    update_usage_count(type, op->data.count, op->data.datatype);

    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_Bcast(buffer, %d, %s, %d, %s);\n", op->data.count, typename, op->data.root, commname);
    set_compute_time();
    break;

  case umpi_MPI_Bsend:
    unimplemented_error("MPI_Bsend");
    break;

  case umpi_MPI_Bsend_init:
    unimplemented_error("MPI_Bsend_init");
    break;

  case umpi_MPI_Buffer_attach:
    unimplemented_error("MPI_Buffer_attach");
    break;

  case umpi_MPI_Buffer_detach:
    unimplemented_error("MPI_Buffer_detach");
    break;

  case umpi_MPI_Cancel:
    unimplemented_error("MPI_Cancel");
    break;

  case umpi_MPI_Cart_coords: // informational only
    break;

  case umpi_MPI_Cart_create: 
    break;

  case umpi_MPI_Cart_get: // informational only
    break;

  case umpi_MPI_Cart_map:
    unimplemented_error("MPI_Cart_map");
    break;

  case umpi_MPI_Cart_rank: // informational only
    break;

  case umpi_MPI_Cart_shift:  // informational only
    break;

  case umpi_MPI_Cart_sub: 
    break;

  case umpi_MPI_Cartdim_get:
    unimplemented_error("MPI_Cartdim_get");
    break;

  case umpi_MPI_Comm_compare:
    unimplemented_error("MPI_Comm_compare");
    break;

  case umpi_MPI_Comm_create: 
    break;

  case umpi_MPI_Comm_dup:
    break;

  case umpi_MPI_Comm_free:
    break;

  case umpi_MPI_Comm_get_name:  // informational only
    break;

  case umpi_MPI_Comm_group:
    break;

  case umpi_MPI_Comm_rank:
    if(add_var_dump(var, "int myrank = 0;", LOCAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh,"\tMPI_Comm_rank(MPI_COMM_WORLD, &myrank);\n");
    set_compute_time();
    break;

  case umpi_MPI_Comm_remote_group:
    unimplemented_error("MPI_Comm_remote_group");
    break;

  case umpi_MPI_Comm_remote_size:
    unimplemented_error("MPI_Comm_remote_size");
    break;

  case umpi_MPI_Comm_set_name:
    unimplemented_error("MPI_Comm_set_name");
    break;

  case umpi_MPI_Comm_size:
    if(add_var_dump(var, "int comm_size = 0;", LOCAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh,"\tMPI_Comm_size(MPI_COMM_WORLD, &comm_size);\n");
    set_compute_time();
    break;

  case umpi_MPI_Comm_split:
    break;    

  case umpi_MPI_Comm_test_inter: 
    unimplemented_error("MPI_Comm_test_inter");
    break;

  /*#ifndef FEATURE_FILE_IO
  case umpi_MPI_Dims_create: 
    unimplemented_error("MPI_Dims_create");
    break;
  #endif*/

  case umpi_MPI_Errhandler_create: 
    unimplemented_warning("MPI_Errhandler_create");
    break;

  case umpi_MPI_Errhandler_free: 
    unimplemented_warning("MPI_Errhandler_free");
    break;

  case umpi_MPI_Errhandler_get: 
    unimplemented_warning("MPI_Errhandler_get");
    break;

  case umpi_MPI_Errhandler_set: 
    unimplemented_warning("MPI_Errhandler_set");
    break;

  case umpi_MPI_Error_class: 
    unimplemented_warning("MPI_Errhandler_class");
    break;

  case umpi_MPI_Error_string:  // informational only
    break;

  case umpi_MPI_Finalize:  // build into replay app
    finalize_req(replay_req);
    finalize_mem(REQ_LIST_SIZE);
    finalize_ptr (replay_ptr);
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_Finalize();\n");
    set_compute_time();
    break;

  case umpi_MPI_Finalized:  // informational only
    break;

  case umpi_MPI_Gather: 
    break;

  case umpi_MPI_Gatherv:
    unimplemented_error("MPI_Gatherv");
    break;

  case umpi_MPI_Get_count:
    unimplemented_error("MPI_Get_count");
    break;

  case umpi_MPI_Get_elements:
    unimplemented_error("MPI_Get_elements");
    break;

  case umpi_MPI_Get_processor_name:
    unimplemented_error("MPI_Get_processor_name");
    break;

  case umpi_MPI_Get_version:
    unimplemented_error("MPI_Get_version");
    break;

  case umpi_MPI_Graph_create: 
    break;

  case umpi_MPI_Graph_get:
    unimplemented_error("MPI_Graph_get");
    break;

  case umpi_MPI_Graph_map:
    unimplemented_error("MPI_Graph_map");
    break;

  case umpi_MPI_Graph_neighbors:
    unimplemented_error("MPI_Graph_neighbors");
    break;

  case umpi_MPI_Graph_neighbors_count:
    unimplemented_error("MPI_Graph_neighbors_count");
    break;

  case umpi_MPI_Graphdims_get:
    unimplemented_error("MPI_Graphdims_get");
    break;

  case umpi_MPI_Group_compare:
    unimplemented_error("MPI_Group_compare");
    break;

  case umpi_MPI_Group_difference:
    unimplemented_error("MPI_Group_difference");
    break;

  case umpi_MPI_Group_excl:
    unimplemented_error("MPI_Group_excl");
    break;

  case umpi_MPI_Group_free: 
    break;

  case umpi_MPI_Group_incl:
    break;

  case umpi_MPI_Group_intersection:
    unimplemented_error("MPI_Group_intersection");
    break;

  case umpi_MPI_Group_range_excl:
    unimplemented_error("MPI_Group_range_excl");
    break;

  case umpi_MPI_Group_range_incl:
    unimplemented_error("MPI_Group_range_incl");
    break;

  case umpi_MPI_Group_rank:
    unimplemented_error("MPI_Group_rank");
    break;

  case umpi_MPI_Group_size:
    unimplemented_error("MPI_Group_size");
    break;

  case umpi_MPI_Group_translate_ranks:
    unimplemented_error("MPI_Group_translate_ranks");
    break;

  case umpi_MPI_Group_union:
    unimplemented_error("MPI_Group_union");
    break;

  case umpi_MPI_Handle2int:
    unimplemented_error("MPI_Handle2int");
    break;

  case umpi_MPI_Ibsend:
    unimplemented_error("MPI_Ibsend");
    break;

  case umpi_MPI_Info_c2f:
    unimplemented_error("MPI_Info_c2f");
    break;

  case umpi_MPI_Info_create:
    unimplemented_error("MPI_Info_create");
    break;

  case umpi_MPI_Info_delete:
    unimplemented_error("MPI_Info_delete");
    break;

  case umpi_MPI_Info_dup:
    unimplemented_error("MPI_Info_dup");
    break;

  case umpi_MPI_Info_f2c:
    unimplemented_error("MPI_Info_f2c");
    break;

  case umpi_MPI_Info_free:
    unimplemented_error("MPI_Info_free");
    break;

  case umpi_MPI_Info_get:
    unimplemented_error("MPI_Info_get");
    break;

  case umpi_MPI_Info_get_nkeys:
    unimplemented_error("MPI_Info_get_nkeys");
    break;

  case umpi_MPI_Info_get_nthkey:
    unimplemented_error("MPI_Info_get_nthkey");
    break;

  case umpi_MPI_Info_get_valuelen:
    unimplemented_error("MPI_Info_get_valuelen");
    break;

  case umpi_MPI_Info_set:
    unimplemented_error("MPI_Info_set");
    break;

  case umpi_MPI_Init:
    RECORDED_MPI_REQUEST_NULL = -1;
    init_req(&replay_req, REQ_LIST_SIZE);
    init_mem(REQ_LIST_SIZE);
    replay_ptr = init_ptr (PTR_HANDLER_SZ);
    comm_total = member_count(op->rlist); //Save comm size for use in prsd_traverse.
    fprintf(skel_fh,"\tMPI_Init(&argc,&argv);\n");
    if(add_var_dump(var, "int myrank;", GLOBAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }
    fprintf(skel_fh,"\tMPI_Comm_rank(MPI_COMM_WORLD, &myrank);\n");

    if(add_var_dump(var, "int comm_size;", GLOBAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }
    fprintf(skel_fh,"\tMPI_Comm_size(MPI_COMM_WORLD, &comm_size);\n");
    fprintf(skel_fh, "\tcreate_derived_type();\n");
    fprintf(skel_fh, "\tbuffersize = get_max_size();\n");
    fprintf(skel_fh, "\tif(buffersize > 0) {\n");
    fprintf(skel_fh, "\t  buffer = malloc(buffersize);\n");
    fprintf(skel_fh, "\t  buffer2 = malloc(buffersize);\n");
    fprintf(skel_fh, "\t}\n");
    break;

  case umpi_MPI_Initialized:
    unimplemented_error("MPI_Initialized");
    break;

  case umpi_MPI_Int2handle:
    unimplemented_error("MPI_Int2handle");
    break;

  case umpi_MPI_Intercomm_create: 
    break;

  case umpi_MPI_Intercomm_merge: 
    break;


  case umpi_MPI_Iprobe:
    unimplemented_error("MPI_Iprobe");
    break;

  case umpi_MPI_Recv_init: 
  case umpi_MPI_Irecv: 
    break;

  case umpi_MPI_Irsend:
    unimplemented_error("MPI_Irsend");
    break;

  case umpi_MPI_Send_init:
  case umpi_MPI_Isend:
    break;

  case umpi_MPI_Issend:
    unimplemented_error("MPI_Issend");
    break;

  case umpi_MPI_Keyval_create:
    unimplemented_error("MPI_Keyval_create");
    break;

  case umpi_MPI_Keyval_free:
    unimplemented_error("MPI_Keyval_free");
    break;

  case umpi_MPI_Op_create:
    unimplemented_error("MPI_Op_create");
    break;

  case umpi_MPI_Op_free:
    unimplemented_error("MPI_Op_free");
    break;

  case umpi_MPI_Pack:
    unimplemented_error("MPI_Pack");
    break;

  case umpi_MPI_Pack_size:
    unimplemented_error("MPI_Pack_size");
    break;

  case umpi_MPI_Pcontrol:
    unimplemented_error("MPI_Pcontrol");
    break;

  case umpi_MPI_Probe:
    unimplemented_error("MPI_Probe");
    break;

  case umpi_MPI_Recv: 
    break;


  case umpi_MPI_Reduce: 
    break;
    
  case umpi_MPI_Reduce_scatter: 
    unimplemented_error("MPI_Reduce_scatter");
    break;

  case umpi_MPI_Request_c2f: 
    unimplemented_error("MPI_Request_c2f");
    break;
    
  case umpi_MPI_Request_free: 
    break;

  case umpi_MPI_Rsend:
    unimplemented_error("MPI_Rsend");
    break;

  case umpi_MPI_Rsend_init:
    unimplemented_error("MPI_Rsend_init");
    break;

  case umpi_MPI_Scan:
    unimplemented_error("MPI_Scan");
    break;

  case umpi_MPI_Scatter:
    unimplemented_error("MPI_Scatter");
    break;

  case umpi_MPI_Scatterv:
    unimplemented_error("MPI_Scatterv");
    break;

  case umpi_MPI_Send:
    /*Arguments: 
      IN buf initial address of send buffer (choice)
      IN count number of elements in send buffer (nonnegative integer)
      IN datatype datatype of each send buffer element (handle)
      IN dest rank of destination (integer)
      IN tag message tag (integer)
      IN comm communicator (handle)
    */
    printf("inside send\n");
    populate_datatype(type, op, &typename);
    printf("after populate datatype\n");
    i=0;
    comm = index_to_comm(replay_ptr, op->data.comm);
    while(MPI_DEFINED_COMMS[i].comm != MPI_COMM_NULL) {
      if(MPI_DEFINED_COMMS[i].comm == comm) {
        strcpy(&commname, MPI_DEFINED_COMMS[i].comm_name);
        break;
      }
      i++;
    }
    dest = op->data.dest_vector ? vector_get_next (op->data.dest_vector) : op->data.dest;
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_Send(buffer, %d, %s, (%d+myrank)/comm_size, %d, %s);", 
            op->data.count, typename, dest, op->data.tag, commname);
    set_compute_time();
 
    break;


  case umpi_MPI_Sendrecv:
    unimplemented_error("MPI_Sendrecv");
    break;

  case umpi_MPI_Sendrecv_replace:
    unimplemented_error("MPI_Sendrecv_replace");
    break;

  case umpi_MPI_Ssend:
    unimplemented_error("MPI_Ssend");
    break;

  case umpi_MPI_Ssend_init:
    unimplemented_error("MPI_Ssend_init");
    break;

  case umpi_MPI_Start:
    break;

  case umpi_MPI_Startall: 
    break;

  case umpi_MPI_Status_c2f:
    unimplemented_error("MPI_Status_c2f");
    break;

  case umpi_MPI_Status_f2c:
    unimplemented_error("MPI_Status_f2c");
    break;

  case umpi_MPI_Test:
    unimplemented_error("MPI_Test");
    break;

  case umpi_MPI_Test_cancelled:
    unimplemented_error("MPI_Test_cancelled");
    break;

  case umpi_MPI_Testall:
    unimplemented_error("MPI_Testall");
    break;

  case umpi_MPI_Testany:
    unimplemented_error("MPI_Testany");
    break;

  case umpi_MPI_Testsome:
    unimplemented_error("MPI_Testsome");
    break;

  case umpi_MPI_Topo_test:
    unimplemented_error("MPI_Topo_test");
    break;

  case umpi_MPI_Type_commit:
    if(!(new_typename = get_type_name_from_index(type, op->data.newtype))) {
      fprintf(stderr,"Derived type is null or new type is not added in type.");
      exit(0);
    }
    simulate_compute_time (skel_type_fh, op, iter);
    fprintf(skel_type_fh, "\tMPI_Type_commit(&%s);\n",new_typename);
    set_compute_time();
    break;

  case umpi_MPI_Type_contiguous:
    break;

  case umpi_MPI_Type_count:
    unimplemented_error("MPI_Type_count");
    break;

  case umpi_MPI_Type_create_darray:
    i = 0;
    datatype = index_to_type(replay_ptr, op->data.datatype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }
    printf("darray: new type tracker %d\n",new_type_tracker);
    new_type_tracker--;
    add_type_entry(replay_ptr, new_type_tracker);
    replay_type_index = type_to_index(replay_ptr, new_type_tracker); 
    if(add_derived_type(type, replay_type_index)) {
      fprintf(stderr,"Write buffer to hold types is possibly full.");
      exit(0);
    }
    if(!(new_typename = get_type_name_from_index(type, replay_type_index))) {
      fprintf(stderr,"Derived type is null or new type is not added in type.");
      exit(0);
    }
    if(add_var_dump(var, "int order,array_of_gsizes[10], array_of_distribs[10],array_of_dargs[10];", GLOBAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }
    fprintf(skel_type_fh, "\torder = %d;\n", op->data.order);
    for(i=0;i<op->data.count;i++) {
      fprintf(skel_type_fh, "\tarray_of_gsizes[%d] = %d;\n",i,op->data.array1[i]);
      fprintf(skel_type_fh, "\tarray_of_distribs[%d] = %d;\n",i,op->data.array2[i]);
      fprintf(skel_type_fh, "\tarray_of_dargs[%d] = %d;\n",i,op->data.array3[i]);
    }

    simulate_compute_time (skel_type_fh, op, iter);
    fprintf(skel_type_fh, 
           "\tMPI_Type_create_darray(comm_size, myrank, ndims, array_of_gsizes, array_of_distribs, array_of_dargs,array_of_psizes, order,%s,&%s);\n",
           typename, new_typename);
    set_compute_time();
    break;

  case umpi_MPI_File_open:
    printf("Inside file open\n");
    i=0;
    comm = index_to_comm(replay_ptr, op->data.comm);
    while(MPI_DEFINED_COMMS[i].comm != MPI_COMM_NULL) {
      printf("after while\n");
      if(MPI_DEFINED_COMMS[i].comm == comm) {
        printf("inside if\n");
        strcpy(&commname, MPI_DEFINED_COMMS[i].comm_name);
        printf("after str copy\n");
        break;
      }
      i++;
    }
    if(add_file_handle(fh_dump, op->data.fh)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }
    if(!(new_fh = get_file_handle_from_index(fh_dump, op->data.fh))) {
      fprintf(stderr,"File handle is null or new file handle is not added in data structure.");
      exit(0);
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh,"\tMPI_File_open(%s,\"%s\",%d,MPI_INFO_NULL, &%s);\n",commname, op->data.filename, op->data.amode, new_fh);
    set_compute_time();
    break;

  case umpi_MPI_File_close:
    if(!(new_fh = get_file_handle_from_index(fh_dump, op->data.fh))) {
      fprintf(stderr,"File handle is null or new file handle is not added in data structure.");
      exit(0);
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_File_close(&%s);\n",new_fh);
    set_compute_time();
    break;

  case umpi_MPI_File_set_view:
    i = 0;
    int disp = 0;
    //TODO: Add code to get proper displacement from vector.
    datatype = index_to_type(replay_ptr, op->data.etype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }
    if(add_derived_type(type, op->data.filetype)) {
      fprintf(stderr,"Write buffer to hold types is possibly full.");
      exit(0);
    }
    if(!(new_typename = get_type_name_from_index(type, op->data.filetype))) {
      fprintf(stderr,"Derived type is null or new type is not added in type.");
      exit(0);
    }
    if(!(new_fh = get_file_handle_from_index(fh_dump, op->data.fh))) {
      fprintf(stderr,"File handle is null or new file handle is not added in data structure.");
      exit(0);
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_File_set_view(%s, %d, %s, %s, \"%s\", MPI_INFO_NULL);\n", new_fh, disp, typename, new_typename, op->data.datarep);
    set_compute_time();
    break;

  case umpi_MPI_File_write_at:
    i = 0;
    
    datatype = index_to_type (replay_ptr, op->data.datatype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }

    if(strlen(typename)==0) {
      strcpy(&typename, get_type_name_from_index(type, op->data.datatype));
    }
    else {
      if(add_defined_type(type, typename, op->data.datatype)) {
        fprintf(stderr,"Write buffer to hold types is possibly full.");
        exit(0);
      }
    } 

    update_usage_count(type, op->data.count, op->data.datatype);
    sprintf(&statusdet, "MPI_Status mpistatus%d;", op->data.seq_num);
    sprintf(&statusname, "mpistatus%d", op->data.seq_num);
    if(add_var_dump(var, &statusdet, LOCAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }

    if(!(new_fh = get_file_handle_from_index(fh_dump, op->data.fh))) {
      fprintf(stderr,"File handle is null or new file handle is not added in data structure.");
      exit(0);
    }
    if(op->data.offsetinfo.totalcount > 0) {
    simulate_compute_time (skel_fh, op, iter);
      fprintf(skel_fh, "\tMPI_File_write_at(%s, myrank*%d+(i*%d), buffer, %d, %s, &%s);\n", 
                new_fh,( op->data.offsetinfo.stride * op->data.offsetinfo.totalcount), 
		op->data.offsetinfo.stride, op->data.count, typename, statusname);
    set_compute_time();
    }
    else {
    simulate_compute_time (skel_fh, op, iter);
      fprintf(skel_fh, "\tMPI_File_write_at(%s, %d, buffer, %d, %s, &%s);\n",new_fh, op->data.offset, op->data.count, typename, statusname);
    set_compute_time();
    }
    break;

  case umpi_MPI_File_write_all:
    i = 0;

    datatype = index_to_type (replay_ptr, op->data.datatype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }

    if(strlen(typename)==0) {
      strcpy(&typename, get_type_name_from_index(type, op->data.datatype));
    }
    else {
      if(add_defined_type(type, typename, op->data.datatype)) {
        fprintf(stderr,"Write buffer to hold types is possibly full.");
        exit(0);
      }
    }

    update_usage_count(type, op->data.count, op->data.datatype);
    sprintf(&statusdet, "MPI_Status mpistatus%d;", op->data.seq_num);
    sprintf(&statusname, "mpistatus%d", op->data.seq_num);
    if(add_var_dump(var, &statusdet, LOCAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }

    if(!(new_fh = get_file_handle_from_index(fh_dump, op->data.fh))) {
      fprintf(stderr,"File handle is null or new file handle is not added in data structure.");
      exit(0);
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_File_write_all(%s, buffer, %d, %s, &%s);\n",
                new_fh, op->data.count, typename, statusname);
    set_compute_time();

    break;

  case umpi_MPI_File_read_all:
    i = 0;

    datatype = index_to_type (replay_ptr, op->data.datatype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == datatype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }

    if(strlen(typename)==0) {
      strcpy(&typename, get_type_name_from_index(type, op->data.datatype));
    }
    else {
      if(add_defined_type(type, typename, op->data.datatype)) {
        fprintf(stderr,"Write buffer to hold types is possibly full.");
        exit(0);
      }
    }

    update_usage_count(type, op->data.count, op->data.datatype);
    sprintf(&statusdet, "MPI_Status mpistatus%d;", op->data.seq_num);
    sprintf(&statusname, "mpistatus%d", op->data.seq_num);
    if(add_var_dump(var, &statusdet, LOCAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }

    if(!(new_fh = get_file_handle_from_index(fh_dump, op->data.fh))) {
      fprintf(stderr,"File handle is null or new file handle is not added in data structure.");
      exit(0);
    }
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_File_read_all(%s, buffer, %d, %s, &%s);\n",
                new_fh, op->data.count, typename, statusname);
    set_compute_time();

    break;

  case umpi_MPI_File_seek:
    if(!(new_fh = get_file_handle_from_index(fh_dump, op->data.fh))) {
      fprintf(stderr,"File handle is null or new file handle is not added in data structure.");
      exit(0);
    }
    //TODO: Add code to parse offset.
    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_File_seek(%s,0,%d);\n", new_fh, op->data.whence);
    set_compute_time();
    break;

  case umpi_MPI_Type_create_indexed_block:
    unimplemented_error("MPI_Type_create_indexed_block");
    break;

  case umpi_MPI_Type_create_subarray:
    unimplemented_error("MPI_Type_create_subarray");
    break;

  case umpi_MPI_Type_extent:
    unimplemented_error("MPI_Type_extent");
    break;

  case umpi_MPI_Type_free:
    /*i = 0;
    newtype = index_to_type (replay_ptr, op->data.newtype);
    while(MPI_DEFINED_DATATYPES[i].type != MPI_DATATYPE_NULL) {
      if(MPI_DEFINED_DATATYPES[i].type == newtype) {
        strcpy(&typename, MPI_DEFINED_DATATYPES[i].type_name);
        break;
      }
      i++;
    }*/
    
    //if(strlen(typename)==0) {
      strcpy(&typename, get_type_name_from_index(type, op->data.newtype));
    //}

    simulate_compute_time (skel_fh, op, iter);
    fprintf(skel_fh, "\tMPI_Type_free(&%s);\n", typename);
    set_compute_time();
    break;

  case umpi_MPI_Type_get_contents:
    unimplemented_error("MPI_Type_get_contents");
    break;

  case umpi_MPI_Type_get_envelope:
    unimplemented_error("MPI_Type_get_envelope");
    break;

  case umpi_MPI_Type_hindexed: 
    break;

  case umpi_MPI_Type_hvector:
    break;

  case umpi_MPI_Type_indexed: 
#if 0
    datatype = index_to_type (replay_ptr, op->data.datatype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_indexed (op->data.count,
                            op->data.array1, op->data.array2, datatype,
                            &newtype);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_indexed did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_type_entry(replay_ptr, newtype);
#endif
    break;

  case umpi_MPI_Type_lb:
    unimplemented_error("MPI_Type_lb");
    break;

  case umpi_MPI_Type_size:
    unimplemented_error("MPI_Type_size");
    break;

  case umpi_MPI_Type_struct: 
    break;

  case umpi_MPI_Type_ub:
    unimplemented_error("MPI_Type_ub");
    break;

  case umpi_MPI_Type_vector: 
    break;

  case umpi_MPI_Unpack:
    unimplemented_error("MPI_Unpack");
    break;

  case umpi_MPI_Wait:
    break;

  case umpi_MPI_Waitall:
    break;

  case umpi_MPI_Waitany: 
    break;


  case umpi_MPI_Waitsome:
    break;

  case umpi_MPI_Wtick:  // informational only
    break;

  case umpi_MPI_Wtime:  // informational only
    break;

  case umpi_MPI_Dims_create:
    if(add_var_dump(var, "int ndims,array_of_psizes[10];", GLOBAL)) {
      fprintf(stderr,"Write buffer to hold variables is possibly full.");
      exit(0);
    }
    fprintf(skel_type_fh, "\tndims = %d;\n", op->data.count);
    simulate_compute_time (skel_type_fh, op, iter);
    fprintf(skel_type_fh, "\tMPI_Dims_create(comm_size, ndims, array_of_psizes);\n");
    set_compute_time();
    break;

  default:
    fprintf(stderr, "error: undefined operation number specified (%d)\n", op->data.op);
    //exit(0);
  }

  return 0;
}

#ifdef FEATURE_TIMING_HISTO

long long cTime;
int set_compute_time ()
{
  cTime = (long long int)(get_cur_time_ns ()/1000);
}

long int get_compute_time ()
{
  long long t;

  t = (long long int)(get_cur_time_ns ()/1000);
  return t - cTime;
}
#endif //FEATURE_TIMING_HISTO

/* EOF */
