#include "op_collect.h"
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>
#include "mpi_util.h"
//#include "../../record/libsrc/glob/umpi_mpi.h"
#include "umpi_internal.h"
#include "opstruct.h"
#include "trace.h"
#include "ptr_handler.h"
#include "req_handler.h"
#include "mem_handler.h"
#ifdef FEATURE_TIMING_HISTO
#include "Histogram.h"
#include "TimeList.h"
#include "busy_wait.h"
#endif // FEATURE_TIMING_HISTO

//# define INLINE inline
#define INLINE

//# define busy_wait nanosleep

/* Prints an error message that the operation is unimplemented,
 * then exits.
 */
static void unimplemented_error(const char *op) {
  fprintf(stderr, "ERROR: %s not implemented\n", op);
  exit(0); 
}

/* Prints a warning message that the op is unimplemented, but 
 * continues execution as normal.
 */
static void unimplemented_warning(const char *op) {
  fprintf(stderr, "WARNING: %s not implemented.  Continuing without executing.\n", op);
}



#ifdef FEATURE_TIMING_HISTO

INLINE long int get_compute_time  ();
INLINE int      set_compute_time  ();
INLINE int      get_sleep_time    (rsd_node *, int, rsd_node *, int);
#else

#define simulate_compute_time(a, b)
#define set_compute_time()

#endif // FEATURE_TIMING_HISTO

int RECORDED_MPI_REQUEST_NULL;
ptr_handler_t replay_ptr;

req_handler_t replay_req;

//#define PRSD_DEBUG
#ifdef PRSD_DEBUG
extern FILE *replay_dbg_dump;
#endif //PRSD_DEBUG

void update_op_buf_size (rsd_node * node);

//#define FEWER_MALLOCS
#ifdef FEWER_MALLOCS

int nmalloc, nrealloc, nnone;

#define Op_xbuf_alloc(Op, xbuf, type, size, buf)  do  {     \
    if (Op##xbuf##_size && Op##xbuf##_size >= (size))       \
    {                                                       \
      buf = (type*) Op##xbuf;                               \
      nnone++;                                              \
    }                                                       \
    else if (Op##xbuf##_size == 0)                          \
    {                                                       \
      Op##xbuf##_size = (size);                             \
      buf = Op##xbuf = (type*) malloc ((size));             \
      nmalloc++;                                            \
    }                                                       \
    else if (Op##xbuf##_size < (size))                      \
    {                                                       \
      Op##xbuf##_size = (size);                             \
      buf = Op##xbuf = (type*) realloc (Op##xbuf, (size));  \
      nrealloc++;                                           \
    }                                                       \
  } while (0)

#define DECLARE_XBUF(Op, xbuf)                  \
  void * Op##xbuf;                              \
  int    Op##xbuf##_size; 

#define DECLARE_SENDBUF(Op)  DECLARE_XBUF(Op, sendbuf)
#define DECLARE_RECVBUF(Op)  DECLARE_XBUF(Op, recvbuf)

#define FREE_XBUF(Op, xbuf)                     \
  free (Op##xbuf);                              \
  Op##xbuf = 0;                                 \
  Op##xbuf##_size = 0;

#define FREE_SENDBUF(Op)  FREE_XBUF(Op, sendbuf)
#define FREE_RECVBUF(Op)  FREE_XBUF(Op, recvbuf)

DECLARE_SENDBUF (Allgather)
  DECLARE_SENDBUF (Allgatherv)
  DECLARE_SENDBUF (Allreduce)
  DECLARE_SENDBUF (Alltoall)
  DECLARE_SENDBUF (Alltoallv)
  DECLARE_SENDBUF (Bcast)
  DECLARE_SENDBUF (Gather)
  DECLARE_SENDBUF (Reduce)
  DECLARE_SENDBUF (Send)
 
  DECLARE_RECVBUF (Allgather)
  DECLARE_RECVBUF (Allgatherv)
  DECLARE_RECVBUF (Allreduce)
  DECLARE_RECVBUF (Alltoall)
  DECLARE_RECVBUF (Alltoallv)
  DECLARE_RECVBUF (Gather)
  DECLARE_RECVBUF (Recv)
  DECLARE_RECVBUF (Reduce)

#define FREE(a)

#define Free_sendrecv_bufs  do  {               \
    FREE_SENDBUF (Allgather)                    \
      FREE_SENDBUF (Allgatherv)                 \
      FREE_SENDBUF (Allreduce)                  \
      FREE_SENDBUF (Alltoall)                   \
      FREE_SENDBUF (Alltoallv)                  \
      FREE_SENDBUF (Bcast)                      \
      FREE_SENDBUF (Gather)                     \
      FREE_SENDBUF (Reduce)                     \
      FREE_SENDBUF (Send)                       \
                                                \
      FREE_RECVBUF (Allgather)                  \
      FREE_RECVBUF (Allgatherv)                 \
      FREE_RECVBUF (Allreduce)                  \
      FREE_RECVBUF (Alltoall)                   \
      FREE_RECVBUF (Alltoallv)                  \
      FREE_RECVBUF (Gather)                     \
      FREE_RECVBUF (Recv)                       \
      FREE_RECVBUF (Reduce)                     \
      } while (0)

#define Print_sendrecv_alloc_stats                                      \
  fprintf (stderr, "nnone = %d, nmalloc = %d, nrealloc = %d\n", nnone,  \
           nmalloc, nrealloc);
#else

#define Op_xbuf_alloc(Op, xbuf, type, size, buf)  \
  buf = (type*) malloc ((size))

#define FREE free

#define Free_sendrecv_bufs
#define Print_sendrecv_alloc_stats

#endif // FEWER_MALLOCS

#define Op_sendbuf_alloc(Op, type, size, buf)   \
  Op_xbuf_alloc(Op, sendbuf, type, size, buf)

#define Op_recvbuf_alloc(Op, type, size, buf)   \
  Op_xbuf_alloc(Op, recvbuf, type, size, buf)


  rsd_node * g_cur_op, * g_prev_op;
INLINE void set_cur_op (rsd_node * op)
{
  g_prev_op = g_cur_op;
  g_cur_op = op;
}

#ifdef FEATURE_TIMING_HISTO
INLINE int prsd_sleep (int sleep_duration)
{
  struct timespec tptr, dummy;

  // enable printf only for debugging on small number of nodes.
  tptr.tv_sec = sleep_duration/1000000;
  tptr.tv_nsec = 1000*(sleep_duration - tptr.tv_sec*1000000);
  //printf ("Comp.:(%d) op:=%d Sleep for %d.%09d s\n", my_rank, op->data.op, 
  //    tptr.tv_sec, tptr.tv_nsec);
  printf("Comp.:(%d) Sleep for %d.%09d s\n", my_rank, tptr.tv_sec, tptr.tv_nsec);
  dummy.tv_nsec = dummy.tv_sec = 0;
  if(busy_wait (&tptr, &dummy) < 0)
    perror("busy_wait");

  return 0;
}


int leftover = 0;
INLINE void simulate_compute_time(rsd_node *op, int i) {
  int sleep_duration = -1;
  
  sleep_duration = get_sleep_time (op, TIMING_PRE, g_prev_op, i);
  printf("Comp.:(%d) %d op is %d get_sleep_time: sleep_duration %d\n",my_rank,op->data.seq_num, op->data.op, sleep_duration);
  sleep_duration = get_sleep_time (op, TIMING_POST, g_prev_op, i);//get_compute_time ();
  printf("Comp.:(%d) %d op is %d get_compute_time: sleep_duration %d\n",my_rank, op->data.seq_num, op->data.op, sleep_duration);
  /*sleep_duration += leftover;
  leftover = 0;
  if (sleep_duration > 0)
  {
    printf("Comp.:(%d) %d op is %d sleep_duration %d\n",my_rank, op->data.seq_num, op->data.op, sleep_duration);
    //prsd_sleep (sleep_duration);
  }
  else
  {
    leftover = sleep_duration;
    
      if (sleep_duration>10000)
      fprintf (stderr, "(%d) compute op:=%d Overshot by %d.%d ms \n", my_rank,
      op->data.op, (float)sleep_duration/1000, sleep_duration%1000);
    
  }*/
}

INLINE int get_sleep_time (rsd_node * op, int i, rsd_node * prev, int j) {
  int sleep_duration = 0;
  void * h = 0;

  if (op->data.time_info[i].time_list)
  {
    //if (op->data.replay_prev_op != prev)
    //{
      if (!prev || timing_list_size (op->data.time_info[i].time_list) == 1) {
        h = timing_list_begin (op->data.time_info[i].time_list);
      } else  {
        h = timing_find_match (op->data.time_info[i].time_list, &prev->signature);
      }

      assert(h);
      if(i == TIMING_POST) {
        op->data.replay_prev_op = prev;
        op->data.replay_histo = h;
      }
    //}
    //else
    //{
      //h = op->data.replay_histo;
    //}
    if (h)
    {
      double min, max;
      int min_node, max_node;
      histogram_min_max (h, &min, &max, &min_node, &max_node);
      if (histogram_has_other (h, my_rank))
      {
        if (my_rank == max_node)
          sleep_duration = max;
        else if (my_rank == min_node)
          sleep_duration = min;
        else
          sleep_duration = (int) histogram_value (h, my_rank, -1);

        if (sleep_duration == -1)
        {
          output_rsd_node (stderr, op);
        }
      }
      else
      {
        sleep_duration = (int) histogram_value (h, my_rank, j);
      }
    }
  }
  else
  {
    sleep_duration = op->data.time_info[i].time;
  }
  if (sleep_duration < 0)
  {
    fprintf (stderr, "[%s:%d] Invalid sleep_duration\n", __FILE__, __LINE__);
    exit (0);
  }

  return sleep_duration;
}

#endif // FEATURE_TIMING_HISTO

void update_op_buf_size (rsd_node * node)
{
#ifdef FEWER_MALLOCS
  int size = 0;

  if (node && node->data.datatype != -1 && node->data.datatype != 0)
  {
    MPI_Datatype datatype = index_to_type (replay_ptr, node->data.datatype);
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS)
    {
      fprintf(stderr, "error: MPI_Type_extent (%d,%p) failed - probably invalid" \
              "type\n", node->data.datatype, datatype);
      fflush (stderr);
      exit(0);
    }
    assert (size);
    size *= node->data.count;
  }
  switch (node->data.op)
  {
  case umpi_MPI_Send:
    Sendsendbuf_size = (Sendsendbuf_size > size) ?
      Sendsendbuf_size : size;
    break;
  case umpi_MPI_Recv:
    Recvrecvbuf_size = (Recvrecvbuf_size > size) ?
      Recvrecvbuf_size : size;
    break;
  default:
    break;
  }
#endif // FEWER_MALLOCS
}

int exec_prsd_call(rsd_node *op, int rank, int depth, int iter) {
  umpi_mpi_lookup_t *lookup;
  rsd_collect_t rsd_info;
  assert (op);
  set_cur_op(op);
  lookup = get_struct_by_opcode ((int)op->data.op);
  //simulate_compute_time(op, iter);
  rsd_info.is_io_op=0;
  rsd_info.op = op;
  rsd_info.rank = rank;
  rsd_info.depth = depth;
  rsd_info.iter = iter;
  //rsd_info.compute_time = get_sleep_time (op,TIMING_PRE,g_prev_op,iter);
  //rsd_info.function_time = get_sleep_time(op,TIMING_POST,g_prev_op,iter);
  io_det.total_ops++;
//  if(rank==0)
//    printf("op %d, compute: %d, func time %d\n", op->data.op, rsd_info.compute_time, rsd_info.function_time);
  io_det.rsd_det = &rsd_info;
  lookup->replay_op (&io_det);
  //io_det.total_compute_time += rsd_info.compute_time;
  //io_det.total_function_time += rsd_info.function_time;
  //io_det.total_time += rsd_info.compute_time+rsd_info.function_time;
  if(io_det.rsd_det->is_io_op)
    io_det.io_ops++;
  else
    io_det.comm_ops++;
  return 0;
}

#ifdef FEATURE_TIMING_HISTO

long long cTime;
INLINE int set_compute_time ()
{
  cTime = (long long int)(get_cur_time_ns ()/1000);
}

INLINE long int get_compute_time ()
{
  long long t;

  t = (long long int)(get_cur_time_ns ()/1000);
  return t - cTime;
}
#endif //FEATURE_TIMING_HISTO

/* EOF */
