#include "op_exec.h"
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>
#include "mpi_util.h"

#include "opstruct.h"
#include "trace.h"
#include "ptr_handler.h"
#include "req_handler.h"
#include "mem_handler.h"
#ifdef FEATURE_TIMING_HISTO
#include "Histogram.h"
#include "TimeList.h"
#include "busy_wait.h"
#endif // FEATURE_TIMING_HISTO

//# define INLINE inline
#define INLINE


//# define busy_wait nanosleep

/* Prints an error message that the operation is unimplemented,
 * then exits.
 */
static void unimplemented_error(const char *op) {
  fprintf(stderr, "ERROR: %s not implemented\n", op);
  exit(0); 
}

/* Prints a warning message that the op is unimplemented, but 
 * continues execution as normal.
 */
static void unimplemented_warning(const char *op) {
  fprintf(stderr, "WARNING: %s not implemented.  Continuing without executing.\n", op);
}



#ifdef FEATURE_TIMING_HISTO

INLINE long int get_compute_time  ();
INLINE int      set_compute_time  ();
INLINE int      get_sleep_time    (rsd_node *, int, rsd_node *, int);
#else

#define simulate_compute_time(a, b)
#define set_compute_time()

#endif // FEATURE_TIMING_HISTO

int RECORDED_MPI_REQUEST_NULL;
ptr_handler_t replay_ptr;

req_handler_t replay_req;

static long long int rl_parse_time;
static long long int tot_sleep_time;
static long long int time_no_work;
static long long int total_recv;
extern RandomVector *timing_rvector;
extern RandomVector *param_rvector;
//#define PRSD_DEBUG
#ifdef PRSD_DEBUG
extern FILE *replay_dbg_dump;
#endif //PRSD_DEBUG

void update_op_buf_size (rsd_node * node);

//#define FEWER_MALLOCS
#ifdef FEWER_MALLOCS

int nmalloc, nrealloc, nnone;

#define Op_xbuf_alloc(Op, xbuf, type, size, buf)  do  {     \
    if (Op##xbuf##_size && Op##xbuf##_size >= (size))       \
    {                                                       \
      buf = (type*) Op##xbuf;                               \
      nnone++;                                              \
    }                                                       \
    else if (Op##xbuf##_size == 0)                          \
    {                                                       \
      Op##xbuf##_size = (size);                             \
      buf = Op##xbuf = (type*) malloc ((size));             \
      nmalloc++;                                            \
    }                                                       \
    else if (Op##xbuf##_size < (size))                      \
    {                                                       \
      Op##xbuf##_size = (size);                             \
      buf = Op##xbuf = (type*) realloc (Op##xbuf, (size));  \
      nrealloc++;                                           \
    }                                                       \
  } while (0)

#define DECLARE_XBUF(Op, xbuf)                  \
  void * Op##xbuf;                              \
  int    Op##xbuf##_size; 

#define DECLARE_SENDBUF(Op)  DECLARE_XBUF(Op, sendbuf)
#define DECLARE_RECVBUF(Op)  DECLARE_XBUF(Op, recvbuf)

#define FREE_XBUF(Op, xbuf)                     \
  free (Op##xbuf);                              \
  Op##xbuf = 0;                                 \
  Op##xbuf##_size = 0;

#define FREE_SENDBUF(Op)  FREE_XBUF(Op, sendbuf)
#define FREE_RECVBUF(Op)  FREE_XBUF(Op, recvbuf)

DECLARE_SENDBUF (Allgather)
  DECLARE_SENDBUF (Allgatherv)
  DECLARE_SENDBUF (Allreduce)
  DECLARE_SENDBUF (Alltoall)
  DECLARE_SENDBUF (Alltoallv)
  DECLARE_SENDBUF (Bcast)
  DECLARE_SENDBUF (Gather)
  DECLARE_SENDBUF (Reduce)
  DECLARE_SENDBUF (Send)
 
  DECLARE_RECVBUF (Allgather)
  DECLARE_RECVBUF (Allgatherv)
  DECLARE_RECVBUF (Allreduce)
  DECLARE_RECVBUF (Alltoall)
  DECLARE_RECVBUF (Alltoallv)
  DECLARE_RECVBUF (Gather)
  DECLARE_RECVBUF (Recv)
  DECLARE_RECVBUF (Reduce)

#define FREE(a)

#define Free_sendrecv_bufs  do  {               \
    FREE_SENDBUF (Allgather)                    \
      FREE_SENDBUF (Allgatherv)                 \
      FREE_SENDBUF (Allreduce)                  \
      FREE_SENDBUF (Alltoall)                   \
      FREE_SENDBUF (Alltoallv)                  \
      FREE_SENDBUF (Bcast)                      \
      FREE_SENDBUF (Gather)                     \
      FREE_SENDBUF (Reduce)                     \
      FREE_SENDBUF (Send)                       \
                                                \
      FREE_RECVBUF (Allgather)                  \
      FREE_RECVBUF (Allgatherv)                 \
      FREE_RECVBUF (Allreduce)                  \
      FREE_RECVBUF (Alltoall)                   \
      FREE_RECVBUF (Alltoallv)                  \
      FREE_RECVBUF (Gather)                     \
      FREE_RECVBUF (Recv)                       \
      FREE_RECVBUF (Reduce)                     \
      } while (0)

#define Print_sendrecv_alloc_stats                                      \
  fprintf (stderr, "nnone = %d, nmalloc = %d, nrealloc = %d\n", nnone,  \
           nmalloc, nrealloc);
#else

#define Op_xbuf_alloc(Op, xbuf, type, size, buf)  \
  buf = (type*) malloc ((size))

#define FREE free

#define Free_sendrecv_bufs
#define Print_sendrecv_alloc_stats

#endif // FEWER_MALLOCS

#define Op_sendbuf_alloc(Op, type, size, buf)   \
  Op_xbuf_alloc(Op, sendbuf, type, size, buf)

#define Op_recvbuf_alloc(Op, type, size, buf)   \
  Op_xbuf_alloc(Op, recvbuf, type, size, buf)


  rsd_node * g_cur_op, * g_prev_op;
#ifdef REPLAY_HISTO
INLINE void set_cur_op (rsd_node *op) {
  if(member_rank(my_rank, op->rlist)){
    g_prev_op = g_cur_op;
    g_cur_op = op;
  }
}
#else
INLINE void set_cur_op (rsd_node * op)
{
  g_prev_op = g_cur_op;
  g_cur_op = op;
}
#endif

#ifdef FEATURE_TIMING_HISTO
INLINE int prsd_sleep (int sleep_duration)
{
  struct timespec tptr, dummy;
  tot_sleep_time += sleep_duration;
  // enable printf only for debugging on small number of nodes.
  tptr.tv_sec = sleep_duration/1000000;
  tptr.tv_nsec = 1000*(sleep_duration - tptr.tv_sec*1000000);
  //printf ("Comp.:(%d) op:=%d Sleep for %d.%09d s\n", my_rank, op->data.op, 
  //    tptr.tv_sec, tptr.tv_nsec);
  //printf("Comp.:(%d) Sleep for %d.%09d s\n", my_rank, tptr.tv_sec, tptr.tv_nsec);
  dummy.tv_nsec = dummy.tv_sec = 0;
  if(busy_wait (&tptr, &dummy) < 0)
    perror("busy_wait");

  return 0;
}


int leftover = 0;
INLINE void simulate_compute_time(rsd_node *op, int i) {
  int sleep_duration = -1;
  //#if 0 
  sleep_duration = get_sleep_time (op, TIMING_PRE, g_prev_op, i);
  /*if(op->data.op == umpi_MPI_Send) {
    sprintf(debugstr,"Sleep:Send:seq:%d, s:%d", op->data.seq_num, sleep_duration );
    write_to_file(my_rank, debugstr);
  }*/
  //printf("Comp.:(%d) op is %d get_sleep_time: sleep_duration %d\n",my_rank, op->data.op, sleep_duration);
  //sleep_duration -= get_compute_time ();
  //if(my_rank == 0 )
    //printf("Comp.:(%d) op is %d get_sleep_time: comp time %d\n",my_rank, op->data.op, comp_time);
  sleep_duration += leftover;
  leftover = 0;
  if (sleep_duration > 0)
  {
    //printf("Comp.:(%d) op is %d sleep_duration %d\n",my_rank, op->data.op, sleep_duration);
    prsd_sleep (sleep_duration);
  }
  else
  {
    //if(my_rank == 0 )
      //printf("sleep duration is less than zero %d\n", sleep_duration);
    leftover = sleep_duration;
    /*
      if (sleep_duration>10000)
      fprintf (stderr, "(%d) compute op:=%d Overshot by %d.%d ms \n", my_rank,
      op->data.op, (float)sleep_duration/1000, sleep_duration%1000);
    */
  }
  //#endif
}

INLINE int get_sleep_time (rsd_node * op, int i, rsd_node * prev, int j) {
  int sleep_duration = 0;
  void * h = 0;

  if (op->data.time_info[i].time_list)
  {
    if (op->data.replay_prev_op != prev)
    {
      if (!prev || timing_list_size (op->data.time_info[i].time_list) == 1) {
        h = timing_list_begin (op->data.time_info[i].time_list);
      } else  {
        h = timing_find_match (op->data.time_info[i].time_list, &prev->signature);
        if(!h)
          h = timing_list_begin(op->data.time_info[i].time_list);
    //printf("Rank: %d: before replay op %d %d\n", get_my_rank(), op->data.op, op->data.seq_num);
      }

      //assert(h);

      op->data.replay_prev_op = prev;
      op->data.replay_histo = h;

    }
    else
    {
      h = op->data.replay_histo;
    }
    if (h)
    {
      double min, max, avg;
      int min_node, max_node;
      histogram_min_max (h, &min, &max, &min_node, &max_node);
      if (histogram_has_other (h, my_rank))
      {
        if (my_rank == max_node)
          sleep_duration = max;
        else if (my_rank == min_node)
          sleep_duration = min;
        else
          sleep_duration = (int) histogram_value_from_vector(h, timing_rvector);//histogram_value (h, my_rank+1, -1);//my_rank, -1);//0, -1);//

        if (sleep_duration == -1)
        {
          output_rsd_node (stderr, op);
        }
      }
      else
      {
        //sleep_duration = (int) histogram_value (h, my_rank+1, j);
        histogram_average(h, &avg);
        sleep_duration = (int) histogram_value_from_vector(h, timing_rvector);//avg;//(int) histogram_value (h, 0, -1); //min;
      }
    }
  }
  else
  {
    sleep_duration = op->data.time_info[i].time;
  }
  if (sleep_duration < 0)
  {
    fprintf (stderr, "[%s:%d] Invalid sleep_duration\n", __FILE__, __LINE__);
    exit (0);
  }

  return sleep_duration;
}

#endif // FEATURE_TIMING_HISTO

void update_op_buf_size (rsd_node * node)
{
#ifdef FEWER_MALLOCS
  int size = 0;

  if (node && node->data.datatype != -1 && node->data.datatype != 0)
  {
    MPI_Datatype datatype = index_to_type (replay_ptr, node->data.datatype);
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS)
    {
      fprintf(stderr, "error: MPI_Type_extent (%d,%p) failed - probably invalid" \
              "type\n", node->data.datatype, datatype);
      fflush (stderr);
      exit(0);
    }
    assert (size);
    size *= node->data.count;
  }
  switch (node->data.op)
  {
  case umpi_MPI_Send:
    Sendsendbuf_size = (Sendsendbuf_size > size) ?
      Sendsendbuf_size : size;
    break;
  case umpi_MPI_Recv:
    Recvrecvbuf_size = (Recvrecvbuf_size > size) ?
      Recvrecvbuf_size : size;
    break;
  default:
    break;
  }
#endif // FEWER_MALLOCS
}

#ifdef REPLAY_HISTO
/*
 * Callback function for traver_rlist.
 *
 * Check if my_rank will receive a message.
 * If yes, record the sender rank in an array.
 */
void rlist_visit(void *data, int rank) {
  int source = -1;
  rank_array_t *send_list = (rank_array_t *)data;
  source = (send_list->dest + rank) % my_size;
  if (source == my_rank) { //If I receive, then store sender
    send_list->rank_arr[send_list->curr_ptr] = rank;
    send_list->curr_ptr++; 
  }
}

/*
 * Initialize rank_array_t and call traverse_rlist with
 * the call back function to collect sender information.
 */
void get_source_ranks(int *rlist, int dest, rank_array_t *send_list) {
  int total_members = 0;
  struct timeval tv;
  long long int start, end;
  gettimeofday(&tv, NULL);
  start = tv.tv_sec * 1000000 + tv.tv_usec;
  total_members = member_count(rlist);
  send_list->rank_arr = (int *) malloc(total_members * sizeof(int) );
  send_list->curr_ptr = 0;
  send_list->dest = dest;
  traverse_rlist ( rlist, (void *)send_list, rlist_visit );
  gettimeofday(&tv, NULL);
  end = tv.tv_sec * 1000000 + tv.tv_usec;
  rl_parse_time += (end-start);
}

void visitor_fill_rank(void *data, int rank) {
  rank_array_t *dlist = (rank_array_t *)data;
  dlist->rank_arr[dlist->curr_ptr] = rank;
  dlist->curr_ptr++;
}

int get_hist_dest(int h_val, rank_array_t *dest_list, int *rlist ) {
  int dest_arr_sz = member_count(rlist);
  dest_list->rank_arr = (int *)malloc(dest_arr_sz * sizeof(int));
  dest_list->curr_ptr = 0;
  traverse_rlist (rlist, (void *)dest_list, visitor_fill_rank);
  for(int i=1; i < dest_list->curr_ptr; i++) {
    if( (h_val >= dest_list->rank_arr[i-1] && h_val <= dest_list->rank_arr[i])) {
      if(h_val <= (dest_list->rank_arr[i-1] + dest_list->rank_arr[i])/2)
        return dest_list->rank_arr[i-1];
      else
        return dest_list->rank_arr[i];
    }
  }
  exit(1);
}

#endif

void write_to_file(int rank, char *str) {
  FILE *fp;
  char filename[20];
  sprintf(filename, "debug-%d", rank);
  fp = fopen(filename, "a");
  fprintf(fp, "%s\n", str);
  fclose(fp);
}

int exec_prsd_call(rsd_node *op, int rank, int depth, int iter)
{
  int *sendelem, *recvelem ;
  int *sdisp, *rdisp;
  MPI_Op mpi_op;
  MPI_Group group, group_out;
  MPI_Comm comm, comm_out, peer_comm;
  MPI_Status status, status2, *statuses;
  MPI_Request request, request2, *requests;
  MPI_Datatype newtype = 0, datatype = 0, recvtype = 0, * datatypes = 0;
  void *sendbuf = 0, *recvbuf = 0;
  MPI_Aint size = 0;
  int cnt_total, comm_size, ret = MPI_ERR_OTHER, oc, cur_oc, *indices;
  int src = -1, dest = -1;
  int sendmax = -1, recvmax = -1, si = -1, ri = -1;
  MPI_Info info;
  MPI_File file;
  int disp = -1;
    char debugstr[100];
 #ifdef REPLAY_HISTO
  rank_array_t send_list;
  rank_array_t dest_list;
 #endif
  set_cur_op (op);
#ifdef PRSD_DEBUG
  if (!replay_dbg_dump) {
    char buf[20];
    sprintf (buf, "debug/%d", rank);
    replay_dbg_dump = fopen (buf, "w");
  }
  output_rsd_node (replay_dbg_dump, op);
#endif // PRSD_DEBUG
  //if(get_my_rank() == 0)
    //printf("Rank: %d: before replay op %d %d\n", get_my_rank(), op->data.op, op->data.seq_num);
  switch(op->data.op) {
  case umpi_MPI_Abort: 
    unimplemented_error("MPI_Abort");
    break;

  case umpi_MPI_Address: 
    unimplemented_error("MPI_Address");
    break;

  case umpi_MPI_Allgather:
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    recvtype = index_to_type(replay_ptr, op->data.recvtype);
    /* allocate recv buf */
    if(MPI_Type_extent(recvtype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    MPI_Comm_size (comm, &comm_size);
    Op_recvbuf_alloc(Allgather, void, op->data.recvcount * size * comm_size, recvbuf);

    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Allgather, void, op->data.count * size, sendbuf);

    if(!sendbuf || !recvbuf) {
      perror("malloc MPI_Allgather buffers");
      exit(0);
    }

    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Allgather(sendbuf, op->data.count, datatype, recvbuf, op->data.recvcount, recvtype, comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Allgather did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    FREE(sendbuf);
    FREE(recvbuf);
    break;

  case umpi_MPI_Allgatherv:
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* allocate recv buf */
    if(!op->data.array1 || !op->data.array2) {
      fprintf(stderr, "error: RECVCOUNTS or DISPLS array improperly allocated\n");
      exit(0);
    }

    cnt_total = 0;
    for(int i = 0; i < op->data.array_size1; i++)
      cnt_total += op->data.array1[i];

    recvtype = index_to_type (replay_ptr, op->data.recvtype);
    if(MPI_Type_extent(recvtype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_recvbuf_alloc(Allgatherv, void, size * cnt_total, recvbuf);

    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Allgatherv, void, op->data.count * size, sendbuf);

    if(!sendbuf || !recvbuf) {
      perror("malloc MPI_Allgatherv buffers");
      exit(0);
    }

    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Allgatherv(sendbuf, op->data.count,
                         datatype, recvbuf,
                         op->data.array1, op->data.array2, recvtype, comm);
    set_compute_time ();
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Allgatherv did not return MPI_SUCCESS\n");
      exit(0);
    }

    FREE(sendbuf);
    FREE(recvbuf);
    break;

  case umpi_MPI_Allreduce: 
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    if(!op->data.comm || comm == MPI_COMM_NULL)
      comm = MPI_COMM_WORLD;

    datatype = index_to_type (replay_ptr, op->data.datatype);
    if(!op->data.datatype)
      datatype = MPI_INT;


    /* allocate send & recv bufs */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Allreduce, void, op->data.count * size, sendbuf);
    Op_recvbuf_alloc(Allreduce, void, op->data.count * size, recvbuf);
    if(!sendbuf || !recvbuf) {
      perror("malloc MPI_Allreduce buffers");
      exit(0);
    }

    /* MPI call */
    mpi_op = index_to_op (replay_ptr, op->data.mpi_op);
    if(mpi_op == MPI_OP_NULL)
      mpi_op = MPI_SUM;
    simulate_compute_time (op, iter);
    //sprintf(debugstr,"Rank:%d: Allreduce: seq: %d, count: %d", my_rank, op->data.seq_num, op->data.count);
    //write_to_file(my_rank, debugstr);
    ret = MPI_Allreduce(sendbuf, recvbuf, op->data.count, datatype, mpi_op, comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_AllReduce did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    FREE(sendbuf);
    FREE(recvbuf);
    break;

  case umpi_MPI_Alltoall:
    /* get communicator */
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    MPI_Comm_size(comm, &comm_size);

    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Alltoall, void, op->data.count * size * comm_size, sendbuf);

    recvtype = index_to_type (replay_ptr, op->data.recvtype);
    /* allocate recv buf */
    if(MPI_Type_extent(recvtype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_recvbuf_alloc(Alltoall, void, op->data.recvcount * size * comm_size, recvbuf);

    if(!sendbuf || !recvbuf) {
      perror("malloc MPI_Alltoall buffers");
      exit(0);
    }

    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Alltoall(sendbuf, op->data.count, datatype, recvbuf, op->data.recvcount, recvtype, comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Alltoall did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    FREE(sendbuf);
    FREE(recvbuf);
    break;

  case umpi_MPI_Alltoallv: 
    /* get communicator */
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    MPI_Comm_size(comm, &comm_size);

    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    sendelem = (int*) malloc(sizeof(int)*comm_size);
    recvelem = (int*) malloc(sizeof(int)*comm_size);
    sdisp = (int*) malloc (sizeof(int) * comm_size);
    rdisp = (int*) malloc (sizeof(int) * comm_size);
    for (int i=0;i< comm_size;i++)
    {
      /*count+offset*/
      sendelem[i] = op->data.array3[i];
      recvelem[i] = op->data.array1[i];

      sdisp[i] = op->data.array2[i];
      rdisp[i] = op->data.array4[i];

      si = sendelem[i] + sdisp[i];
      ri = recvelem[i] + rdisp[i];

      sendmax = (sendmax > si) ? sendmax : si;
      recvmax = (recvmax > ri) ? recvmax : ri;
    }
    Op_sendbuf_alloc(Alltoallv, void, sendmax * size * comm_size, sendbuf);
    if(!sendbuf)
      fprintf(stderr,"error: malloc(sendbuf)!\n");

    recvtype = index_to_type (replay_ptr, op->data.recvtype);
    /* allocate recv buf */
    if(MPI_Type_extent(recvtype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_recvbuf_alloc(Alltoallv, void, recvmax * size, recvbuf);
    if(!recvbuf)
      fprintf(stderr,"error: malloc(recvbuf)\n");
    if(!sendbuf || !recvbuf) {
      perror("malloc MPI_Alltoallv buffers");
      exit(0);
    }

    simulate_compute_time (op, iter);
    ret = MPI_Alltoallv(sendbuf, sendelem, sdisp,
                        datatype, recvbuf, recvelem, rdisp,
                        recvtype, comm);
    set_compute_time (); 
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Alltoallv did not return MPI_SUCCESS\n");
      exit(0);
    }
    TRACE (("Before ending by %d!!\n",rank));
    free (sdisp);
    free (rdisp);
    FREE(sendbuf);
    FREE(recvbuf);
    free(sendelem);
    free(recvelem);
    break;

  case umpi_MPI_Attr_delete: 
    unimplemented_error("MPI_Attr_delete");
    break;

  case umpi_MPI_Attr_get: 
    unimplemented_error("MPI_Attr_get");
    break;

  case umpi_MPI_Attr_put: 
    unimplemented_error("MPI_Attr_put");
    break;

  case umpi_MPI_Barrier: 

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    if(!op->data.comm || comm == MPI_COMM_NULL)
      comm = MPI_COMM_WORLD;


    /* MPI call */
    simulate_compute_time (op, iter);
    //sprintf(debugstr,"Rank:%d: Barrier: seq: %d", my_rank, op->data.seq_num);
    //write_to_file(my_rank, debugstr);
    ret = MPI_Barrier(comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Barrier did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    break;

  case umpi_MPI_Bcast: 
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    if(!op->data.comm || comm == MPI_COMM_NULL)
      comm = MPI_COMM_WORLD;

    datatype = index_to_type (replay_ptr, op->data.datatype);
    if(!op->data.datatype)
      datatype = MPI_INT;

    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Bcast, void, op->data.count * size, sendbuf);
    if(!sendbuf) {
      perror("malloc MPI_Bcast buffer");
      exit(0);
    }

    if(op->hist_param) {
      op->data.root = (int) histogram_value_from_vector(op->hist_param, param_rvector);//histogram_value(op->hist_param, 0, -1);
      //if(get_my_rank() == 0 || get_my_rank() == 1)
        //printf("Rank %d: Root: %d\n", get_my_rank(), op->data.root);
    }

    /* MPI call */
    simulate_compute_time (op, iter);
    //sprintf(debugstr,"Rank:%d: Bcast: seq: %d, root: %d", my_rank, op->data.seq_num, op->data.root);
    //write_to_file(my_rank, debugstr);
    ret = MPI_Bcast(sendbuf, op->data.count, datatype, op->data.root, comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Bcast did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    FREE(sendbuf);
    break;

  case umpi_MPI_Bsend:
    unimplemented_error("MPI_Bsend");
    break;

  case umpi_MPI_Bsend_init:
    unimplemented_error("MPI_Bsend_init");
    break;

  case umpi_MPI_Buffer_attach:
    unimplemented_error("MPI_Buffer_attach");
    break;

  case umpi_MPI_Buffer_detach:
    unimplemented_error("MPI_Buffer_detach");
    break;

  case umpi_MPI_Cancel:
    unimplemented_error("MPI_Cancel");
    break;

  case umpi_MPI_Cart_coords: // informational only
    break;

  case umpi_MPI_Cart_create: 
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Cart_create(comm, op->data.count, op->data.array1, op->data.array2, op->data.reorder, &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Cart_create did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_comm_entry(replay_ptr, comm_out);
    break;

  case umpi_MPI_Cart_get: // informational only
    break;

  case umpi_MPI_Cart_map:
    unimplemented_error("MPI_Cart_map");
    break;

  case umpi_MPI_Cart_rank: // informational only
    break;

  case umpi_MPI_Cart_shift:  // informational only
    break;

  case umpi_MPI_Cart_sub: 
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Cart_sub(comm, op->data.array1, &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Cart_sub did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_comm_entry(replay_ptr, comm_out);
    break;

  case umpi_MPI_Cartdim_get:
    unimplemented_error("MPI_Cartdim_get");
    break;

  case umpi_MPI_Comm_compare:
    unimplemented_error("MPI_Comm_compare");
    break;

  case umpi_MPI_Comm_create: 
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* get group */

    group = (MPI_Group)index_to_group (replay_ptr,
                                      op->data.group);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Comm_create(comm, group, &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Comm_create did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_comm_entry(replay_ptr, comm_out);
    break;

  case umpi_MPI_Comm_dup:
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Comm_dup(comm, &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Comm_dup did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_comm_entry(replay_ptr, comm_out);
    break;

  case umpi_MPI_Comm_free:
    /* get communicator */
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    remove_comm_entry(replay_ptr, comm);

    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Comm_free(&comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Comm_free did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    break;

  case umpi_MPI_Comm_get_name:  // informational only
    break;

  case umpi_MPI_Comm_group:
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Comm_group(comm, &group);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Comm_group did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_group_entry(replay_ptr, group);

    break;

  case umpi_MPI_Comm_rank:  // informational only
    break;

  case umpi_MPI_Comm_remote_group:
    unimplemented_error("MPI_Comm_remote_group");
    break;

  case umpi_MPI_Comm_remote_size:
    unimplemented_error("MPI_Comm_remote_size");
    break;

  case umpi_MPI_Comm_set_name:
    unimplemented_error("MPI_Comm_set_name");
    break;

  case umpi_MPI_Comm_size: // informational only
    break;

  case umpi_MPI_Comm_split:
   #ifndef REPLAY_HISTO
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Comm_split(comm, op->data.color,
                         DECODE_KEY (op->data.key), &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Comm_split did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_comm_entry(replay_ptr, comm_out);
   #endif
    break;    

  case umpi_MPI_Comm_test_inter: 
    unimplemented_error("MPI_Comm_test_inter");
    break;

  /*#ifndef FEATURE_FILE_IO
  case umpi_MPI_Dims_create: 
    unimplemented_error("MPI_Dims_create");
    break;
  #endif*/

  case umpi_MPI_Errhandler_create: 
    unimplemented_warning("MPI_Errhandler_create");
    break;

  case umpi_MPI_Errhandler_free: 
    unimplemented_warning("MPI_Errhandler_free");
    break;

  case umpi_MPI_Errhandler_get: 
    unimplemented_warning("MPI_Errhandler_get");
    break;

  case umpi_MPI_Errhandler_set: 
    unimplemented_warning("MPI_Errhandler_set");
    break;

  case umpi_MPI_Error_class: 
    unimplemented_warning("MPI_Errhandler_class");
    break;

  case umpi_MPI_Error_string:  // informational only
    break;

  case umpi_MPI_Finalize:  // build into replay app
    finalize_req(replay_req);
    finalize_mem(REQ_LIST_SIZE);
    finalize_ptr (replay_ptr);
    printf("Rank: %d: Total time in get_source_ranks is %lld, tot sleep time %lld\n", get_my_rank(), rl_parse_time, tot_sleep_time);
    printf("Rank: %d: Time spent in no work %lld\n",get_my_rank(),time_no_work);
    printf("Rank: %d: Total recvs replayed %lld\n",get_my_rank(),total_recv);
    break;

  case umpi_MPI_Finalized:  // informational only
    break;

  case umpi_MPI_Gather: 
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    /* recv buff for root only */
    if(rank == op->data.root) {
      recvtype = index_to_type (replay_ptr, op->data.recvtype);
      /* allocate recv buf */
      if(MPI_Type_extent(recvtype, &size) != MPI_SUCCESS) {
        fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
        exit(0);
      }
      MPI_Comm_size(comm, &comm_size);
      Op_recvbuf_alloc(Gather, void, op->data.recvcount * size * comm_size, recvbuf);
    }
    else
      recvbuf = NULL;

    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Gather, void, op->data.count * size, sendbuf);

    if(!sendbuf || (!recvbuf && rank == op->data.root)) {
      perror("malloc MPI_Gather buffers");
      exit(0);
    }

    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Gather(sendbuf, op->data.count, datatype, recvbuf, op->data.recvcount, recvtype, op->data.root, comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Gather did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    FREE(sendbuf);
    FREE(recvbuf);
    break;

  case umpi_MPI_Gatherv:
    unimplemented_error("MPI_Gatherv");
    break;

  case umpi_MPI_Get_count:
    unimplemented_error("MPI_Get_count");
    break;

  case umpi_MPI_Get_elements:
    unimplemented_error("MPI_Get_elements");
    break;

  case umpi_MPI_Get_processor_name:
    unimplemented_error("MPI_Get_processor_name");
    break;

  case umpi_MPI_Get_version:
    unimplemented_error("MPI_Get_version");
    break;

  case umpi_MPI_Graph_create: 
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);

    simulate_compute_time (op, iter);
    ret = MPI_Graph_create (comm,
                            op->data.array_size1, op->data.array1,
                            op->data.array2, op->data.reorder,
                            &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Graph_create did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_comm_entry(replay_ptr, comm_out);
    break;

  case umpi_MPI_Graph_get:
    unimplemented_error("MPI_Graph_get");
    break;

  case umpi_MPI_Graph_map:
    unimplemented_error("MPI_Graph_map");
    break;

  case umpi_MPI_Graph_neighbors:
    unimplemented_error("MPI_Graph_neighbors");
    break;

  case umpi_MPI_Graph_neighbors_count:
    unimplemented_error("MPI_Graph_neighbors_count");
    break;

  case umpi_MPI_Graphdims_get:
    unimplemented_error("MPI_Graphdims_get");
    break;

  case umpi_MPI_Group_compare:
    unimplemented_error("MPI_Group_compare");
    break;

  case umpi_MPI_Group_difference:
    unimplemented_error("MPI_Group_difference");
    break;

  case umpi_MPI_Group_excl:
    unimplemented_error("MPI_Group_excl");
    break;

  case umpi_MPI_Group_free: 
    group = (MPI_Group)index_to_group (replay_ptr,
                                      op->data.group);

    simulate_compute_time (op, iter);
    ret = MPI_Group_free (&group);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Group_free did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    remove_group_entry(replay_ptr, group);
    break;

  case umpi_MPI_Group_incl:
    /* get group */
    if(op->data.array1 == NULL) {
      fprintf(stderr, "error: no group members identified\n");
      exit(0);
    }

    group = (MPI_Group)index_to_group (replay_ptr,
                                      op->data.group);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Group_incl(group,
                         op->data.count, (int*) op->data.array1, &group_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Group_incl did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_group_entry(replay_ptr, group_out);
    break;

  case umpi_MPI_Group_intersection:
    unimplemented_error("MPI_Group_intersection");
    break;

  case umpi_MPI_Group_range_excl:
    unimplemented_error("MPI_Group_range_excl");
    break;

  case umpi_MPI_Group_range_incl:
     /* get group */
    if(op->data.array1 == NULL) {
      fprintf(stderr, "error: no group members identified\n");
      exit(0);
    }

    group = (MPI_Group)index_to_group (replay_ptr,
                                      op->data.group);
    /* MPI call */
    simulate_compute_time (op, iter);
    ret = MPI_Group_range_incl(group,
                         op->data.count, (int*) op->data.array1, &group_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Group_incl did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    add_group_entry(replay_ptr, group_out);
    break;

  case umpi_MPI_Group_rank:
    unimplemented_error("MPI_Group_rank");
    break;

  case umpi_MPI_Group_size:
    unimplemented_error("MPI_Group_size");
    break;

  case umpi_MPI_Group_translate_ranks:
    unimplemented_error("MPI_Group_translate_ranks");
    break;

  case umpi_MPI_Group_union:
    unimplemented_error("MPI_Group_union");
    break;

  case umpi_MPI_Handle2int:
    unimplemented_error("MPI_Handle2int");
    break;

  case umpi_MPI_Ibsend:
    unimplemented_error("MPI_Ibsend");
    break;

  case umpi_MPI_Info_c2f:
    unimplemented_error("MPI_Info_c2f");
    break;

  case umpi_MPI_Info_create:
    unimplemented_error("MPI_Info_create");
    break;

  case umpi_MPI_Info_delete:
    unimplemented_error("MPI_Info_delete");
    break;

  case umpi_MPI_Info_dup:
    unimplemented_error("MPI_Info_dup");
    break;

  case umpi_MPI_Info_f2c:
    unimplemented_error("MPI_Info_f2c");
    break;

  case umpi_MPI_Info_free:
    unimplemented_error("MPI_Info_free");
    break;

  case umpi_MPI_Info_get:
    unimplemented_error("MPI_Info_get");
    break;

  case umpi_MPI_Info_get_nkeys:
    unimplemented_error("MPI_Info_get_nkeys");
    break;

  case umpi_MPI_Info_get_nthkey:
    unimplemented_error("MPI_Info_get_nthkey");
    break;

  case umpi_MPI_Info_get_valuelen:
    unimplemented_error("MPI_Info_get_valuelen");
    break;

  case umpi_MPI_Info_set:
    unimplemented_error("MPI_Info_set");
    break;

  case umpi_MPI_Init:
  
    RECORDED_MPI_REQUEST_NULL = -1;
    init_req(&replay_req, REQ_LIST_SIZE);
    init_mem(REQ_LIST_SIZE);
    replay_ptr = init_ptr (PTR_HANDLER_SZ);
    rl_parse_time = 0;
#ifndef LINK_WITH_RECORD
    MPI_Barrier(MPI_COMM_WORLD); 
#endif

    set_compute_time ();
#ifndef LINK_WITH_RECORD
    MPI_Pcontrol (1);
#endif
    return 0;
    break;

  case umpi_MPI_Initialized:
    unimplemented_error("MPI_Initialized");
    break;

  case umpi_MPI_Int2handle:
    unimplemented_error("MPI_Int2handle");
    break;

  case umpi_MPI_Intercomm_create: 
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    peer_comm = (MPI_Comm)index_to_comm (replay_ptr,
                                         op->data.peer_comm);
    simulate_compute_time (op, iter);
    ret = MPI_Intercomm_create (comm, op->data.local_leader,
                                peer_comm, op->data.remote_leader,
                                op->data.tag, &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Intercomm_create did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_comm_entry(replay_ptr, comm_out);
    break;

  case umpi_MPI_Intercomm_merge: 
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    simulate_compute_time (op, iter);
    ret = MPI_Intercomm_merge (comm, op->data.high,
                               &comm_out);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Intercomm_merge did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_comm_entry(replay_ptr, comm_out);
    break;


  case umpi_MPI_Iprobe:
    unimplemented_error("MPI_Iprobe");
    break;

  case umpi_MPI_Recv_init: 
  case umpi_MPI_Irecv: 
   #ifndef REPLAY_HISTO
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    recvbuf = malloc(op->data.count * size);
    if(!recvbuf) {
      perror("malloc MPI_Irecv buffer");
      exit(0);
    }
    add_mem_entry(recvbuf);

    /* MPI call */
    MPI_Comm_size (comm, &comm_size);
    src = op->data.source_vector ? vector_get_next (op->data.source_vector) : op->data.source;
    src = DECODE_SOURCE (src);
    //sprintf(debugstr,"Rank:%d: IRecv: seq: %d, count: %d, sender: %d", my_rank, op->data.seq_num, op->data.count, src );
    //write_to_file(my_rank, debugstr);
    simulate_compute_time (op, iter);
    if (op->data.op == umpi_MPI_Irecv)
    {
      ret = MPI_Irecv (recvbuf, op->data.count, datatype, src,
                       op->data.tag, comm, &request);
    }
    else if (op->data.op == umpi_MPI_Recv_init)
    {
      ret = MPI_Recv_init (recvbuf, op->data.count, datatype, src,
                           op->data.tag, comm, &request);
    }
    set_compute_time (); 
    add_request_entry (replay_req, request);

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Irecv did not return MPI_SUCCESS\n");
      exit(0);
    }
    //cannot free this right now. mem_handler frees this.
    //free(recvbuf);
   #else
    simulate_compute_time(op, iter);
    set_compute_time();
   #endif
    break;

  case umpi_MPI_Irsend:
    unimplemented_error("MPI_Irsend");
    break;

  case umpi_MPI_Send_init:
  case umpi_MPI_Isend:
   #ifndef REPLAY_HISTO
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    sendbuf = malloc(op->data.count * size);
    if(!sendbuf) {
      perror("malloc MPI_Isend buffer");
      exit(0);
    }
    add_mem_entry(sendbuf);

    MPI_Comm_size(comm, &comm_size);
    /* MPI call */
    dest = op->data.dest_vector ? vector_get_next (op->data.dest_vector) : op->data.dest;
    dest = DECODE_DEST (dest);
    simulate_compute_time (op, iter);
    if (op->data.op == umpi_MPI_Isend)
    {
      ret = MPI_Isend (sendbuf, op->data.count, datatype, dest,
                       op->data.tag, comm, &request);
    }
    else if (op->data.op == umpi_MPI_Send_init)
    {
      ret = MPI_Send_init (sendbuf, op->data.count, datatype, dest,
                           op->data.tag, comm, &request);
    }
    set_compute_time (); 
    add_request_entry(replay_req, request);

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Isend did not return MPI_SUCCESS\n");
      exit(0);
    }

    //free(sendbuf);
   #else
    if(op->hist_param && op->hist_op == DEST) {
      dest =  (int) histogram_value_from_vector(op->hist_param, param_rvector);//histogram_value(op->hist_param, 0, -1 );
      //sprintf(debugstr,"ISend: seq: %d, dest: %d", op->data.seq_num, dest );
      //write_to_file(my_rank, debugstr);
    }
    else {
      dest = op->data.dest_vector ? vector_get_next (op->data.dest_vector) : op->data.dest;
    }
    //if(!op->match)
      //break; //I dont participate. Came to here to pick a random number
    if(op->hist_param)
      dest = get_hist_dest(dest, &dest_list, op->graph.rank_list);

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    if(op->data.comm == 0 || comm == MPI_COMM_NULL)
      comm = MPI_COMM_WORLD;

    struct timeval ttv;
    long long int sstart, send;

    int is_sender = 0;
    if(member_rank(my_rank, op->rlist) ) {
      is_sender = 1;
    }
    datatype = index_to_type (replay_ptr, op->data.datatype);
    if(op->data.datatype == 0)
      datatype = MPI_INT;
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Send, void, op->data.count * size, sendbuf);
    if(!sendbuf) {
      perror("malloc MPI_Send buffer");
      exit(0);
    }
    gettimeofday(&ttv, NULL);
    sstart = ttv.tv_sec * 1000000 + ttv.tv_usec;
    /* MPI call */

    gettimeofday(&ttv, NULL);
    send = ttv.tv_sec * 1000000 + ttv.tv_usec;
    time_no_work += (send - sstart);

    //else { //I am the receiver
      //Find out who are sending to me
      //printf("Inside Recv\n");
      get_source_ranks (op->rlist, dest, &send_list);
      //simulate_compute_time (op, iter);
      if(send_list.curr_ptr)
        requests = (MPI_Request *) malloc(sizeof(MPI_Request) * send_list.curr_ptr);
      int num_recvs = 0;
      for(int iter = 0; iter < send_list.curr_ptr; iter++ ) {
        //ret = MPI_Recv (sendbuf, op->data.count, datatype,
        //              send_list.rank_arr[iter], op->data.tag, comm, &status);
        //sprintf(debugstr,"Rank:%d: IRecv: seq: %d, count: %d, sender: %d", my_rank, op->data.seq_num, op->data.count, send_list.rank_arr[iter] );
        //write_to_file(my_rank, debugstr);
        if(is_sender && send_list.rank_arr[iter] == my_rank)
          continue;

        ret = MPI_Irecv (sendbuf, op->data.count, datatype,
                      send_list.rank_arr[iter], op->data.tag, comm, &requests[iter]);
        num_recvs++;
        if(ret != MPI_SUCCESS) {
          fprintf(stderr, "error: MPI_Recv did not return MPI_SUCCESS\n");
          exit(0);
        }
        //if(sent) {
        //  MPI_Wait(&request, &status);
        //  sent = 0;
        //}
        //MPI_Wait(&request2, &status2);
      }
    //}

    int sent = 0;
    //Check if I am the sender
    if(member_rank(my_rank, op->rlist) ) {
      //printf("Inside Send\n");
      simulate_compute_time (op, iter);
      dest = DECODE_DEST (dest);
      if(dest != my_rank) { //Avoid sending myself
        //ret = MPI_Send(sendbuf, op->data.count, datatype, dest,
        //           op->data.tag, comm);
        //sprintf(debugstr,"Rank:%d: ISend: seq: %d, count: %d, dest: %d", my_rank, op->data.seq_num, op->data.count, dest);
        //write_to_file(my_rank, debugstr);
        ret = MPI_Isend (sendbuf, op->data.count, datatype, dest,
                       op->data.tag, comm, &request);
        sent = 1;
        if(ret != MPI_SUCCESS) {
          fprintf(stderr, "error: MPI_Send did not return MPI_SUCCESS\n");
          exit(0);
        }
        MPI_Wait(&request, &status);
      }
      //set_compute_time ();
    }
    //if(send_list.curr_ptr) {
    if(num_recvs){ 
      total_recv += num_recvs; 
      statuses = (MPI_Status *) malloc(sizeof(MPI_Status) * num_recvs);//send_list.curr_ptr);
      MPI_Waitall(send_list.curr_ptr, requests, statuses);
    }
    
    set_compute_time ();

    //Calculate time wasted by not participating
    /*if(!is_sender && !num_recvs) {
      gettimeofday(&ttv, NULL);
      send = ttv.tv_sec * 1000000 + ttv.tv_usec;
      time_no_work += (send - sstart);
    }*/

/*    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Send did not return MPI_SUCCESS\n");
      exit(0);
    }*/

    FREE(sendbuf);   
   #endif
    break;

  case umpi_MPI_Issend:
    unimplemented_error("MPI_Issend");
    break;

  case umpi_MPI_Keyval_create:
    unimplemented_error("MPI_Keyval_create");
    break;

  case umpi_MPI_Keyval_free:
    unimplemented_error("MPI_Keyval_free");
    break;

  case umpi_MPI_Op_create:
    unimplemented_error("MPI_Op_create");
    break;

  case umpi_MPI_Op_free:
    unimplemented_error("MPI_Op_free");
    break;

  case umpi_MPI_Pack:
    unimplemented_error("MPI_Pack");
    break;

  case umpi_MPI_Pack_size:
    unimplemented_error("MPI_Pack_size");
    break;

  case umpi_MPI_Pcontrol:
    //unimplemented_error("MPI_Pcontrol");
    break;

  case umpi_MPI_Probe:
    unimplemented_error("MPI_Probe");
    break;

  case umpi_MPI_Recv: 
    /* get communicator */
   #ifndef REPLAY_HISTO

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    datatype = index_to_type (replay_ptr, op->data.datatype);
    //FIXME: fix this
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_recvbuf_alloc(Recv, void, op->data.count * size, recvbuf);
    if(!recvbuf) {
      perror("malloc MPI_Recv buffer");
      exit(0);
    }

    /* MPI call */
    src = op->data.source_vector ? vector_get_next (op->data.source_vector) : op->data.source;
    simulate_compute_time (op, iter);
    src = DECODE_SOURCE (src);
    ret = MPI_Recv(recvbuf, op->data.count, datatype, src, op->data.tag, comm, &status);
    set_compute_time ();

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Recv did not return MPI_SUCCESS\n");
      exit(0);
    }

    FREE(recvbuf);
   #else
    simulate_compute_time(op, iter);
    set_compute_time();
   #endif
    break;


  case umpi_MPI_Reduce: 
    /* get communicator */

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Reduce, void, op->data.count * size, sendbuf);

    /* recv buff for root only */
    if(rank == op->data.root) {
      /* allocate recv buf */
      Op_recvbuf_alloc(Reduce, void, op->data.count * size, recvbuf);
    }
    else
      recvbuf = NULL;

    if(!sendbuf || (!recvbuf && rank == op->data.root)) {
      perror("malloc MPI_Reduce buffers");
      exit(0);
    }

    /* MPI call */
    mpi_op = index_to_op (replay_ptr, op->data.mpi_op);
    simulate_compute_time (op, iter);
    ret = MPI_Reduce(sendbuf, recvbuf, op->data.count, datatype, mpi_op, op->data.root, comm);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Reduce did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    FREE(sendbuf);
    FREE(recvbuf);
    break;
    
  case umpi_MPI_Reduce_scatter: 
    unimplemented_error("MPI_Reduce_scatter");
    break;

  case umpi_MPI_Request_c2f: 
    unimplemented_error("MPI_Request_c2f");
    break;
    
  case umpi_MPI_Request_free: 
    {
      int i = (op->data.request_vector ? vector_get_next(op->data.request_vector) : op->data.request);
      lookup_request(replay_req, i, RECORDED_MPI_REQUEST_NULL, &request);
      simulate_compute_time (op, iter);
      ret = MPI_Request_free(&request);
      set_compute_time ();
      if(ret != MPI_SUCCESS) {
        fprintf(stderr, "error: MPI_Request_free did not return MPI_SUCCESS\n");
        exit(0);
      }
      free_mem_entry(i);
    }
    break;

  case umpi_MPI_Rsend:
    unimplemented_error("MPI_Rsend");
    break;

  case umpi_MPI_Rsend_init:
    unimplemented_error("MPI_Rsend_init");
    break;

  case umpi_MPI_Scan:
    unimplemented_error("MPI_Scan");
    break;

  case umpi_MPI_Scatter:
    unimplemented_error("MPI_Scatter");
    break;

  case umpi_MPI_Scatterv:
    unimplemented_error("MPI_Scatterv");
    break;

  case umpi_MPI_Send: 
    /* get communicator */
   #ifndef REPLAY_HISTO
    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    datatype = index_to_type (replay_ptr, op->data.datatype);
    //FIXME: fix this
    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Send, void, op->data.count * size, sendbuf);
    if(!sendbuf) {
      perror("malloc MPI_Send buffer");
      exit(0);
    }
    /* MPI call */

    dest = op->data.dest_vector ? vector_get_next (op->data.dest_vector) : op->data.dest;
    simulate_compute_time (op, iter);
    dest = DECODE_DEST (dest);
    //sprintf(debugstr,"Rank:%d: Send: seq: %d, count: %d, dest: %d", my_rank, op->data.seq_num, op->data.count, dest );
    //write_to_file(my_rank, debugstr);
    ret = MPI_Send(sendbuf, op->data.count, datatype, dest,
                   op->data.tag, comm);
    set_compute_time ();

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Send did not return MPI_SUCCESS\n");
      exit(0);
    }

    FREE(sendbuf);
   #else
    if(op->hist_param && op->hist_op == DEST) {
      dest =  (int) histogram_value_from_vector(op->hist_param, param_rvector);//histogram_value(op->hist_param, 0, -1 );
    }
    else {
      dest = op->data.dest_vector ? vector_get_next (op->data.dest_vector) : op->data.dest;
    }
    /*if(op->match != 1) {
      break; //I dont participate. Came to here to pick a random number
    }*/
    if(op->hist_param)
      dest = get_hist_dest(dest, &dest_list, op->graph.rank_list);
    //  sprintf(debugstr,"Send: seq: %d, dest: %d", op->data.seq_num, dest );
      //write_to_file(my_rank, debugstr);

    comm = (MPI_Comm)index_to_comm (replay_ptr, op->data.comm);
    datatype = index_to_type (replay_ptr, op->data.datatype);

    struct timeval tv;
    long long int start, end;

    int is_Sender = 0;
    if(member_rank(my_rank, op->rlist)) {
      is_Sender = 1;
    }

    /* allocate send buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Send, void, op->data.count * size, sendbuf);
    if(!sendbuf) {
      perror("malloc MPI_Send buffer");
      exit(0);
    }
    gettimeofday(&tv, NULL);
    start = tv.tv_sec * 1000000 + tv.tv_usec;
    /* MPI call */
    /*if(op->hist_param && op->hist_op == DEST) {
      dest = histogram_value(op->hist_param, 0, -1 );
    }
    else {
      dest = op->data.dest_vector ? vector_get_next (op->data.dest_vector) : op->data.dest;
    }*/
    //gettimeofday(&tv, NULL);
    //end = tv.tv_sec * 1000000 + tv.tv_usec;
    //time_no_work += (end - start);
    get_source_ranks(op->rlist, dest, &send_list);
    if(send_list.curr_ptr)
      requests = (MPI_Request *) malloc(sizeof(MPI_Request) * send_list.curr_ptr);
    int num_Recvs = 0;
    for(int iter = 0; iter < send_list.curr_ptr; iter++ ) {
      if(is_Sender && send_list.rank_arr[iter] == my_rank)
        continue;

      //sprintf(debugstr,"Befr:IR:seq:%d, s:%d", op->data.seq_num, send_list.rank_arr[iter] );
      //write_to_file(my_rank, debugstr);
      ret = MPI_Irecv (sendbuf, op->data.count, datatype,
           send_list.rank_arr[iter], op->data.tag, comm, &requests[iter]);
      num_Recvs++;
      if(ret != MPI_SUCCESS) {
        fprintf(stderr, "error: MPI_Recv did not return MPI_SUCCESS\n");
        exit(0);
      }
    }
    gettimeofday(&tv, NULL);
    end = tv.tv_sec * 1000000 + tv.tv_usec;
    time_no_work += (end - start);
    //sprintf(debugstr,"Recv Proc:Send:seq:%d, s:%d", op->data.seq_num, (end-start) );
    //write_to_file(my_rank, debugstr);
     
    //Check if I am the sender
    if(member_rank(my_rank, op->rlist) ) {
      simulate_compute_time (op, iter);
      dest = DECODE_DEST (dest);
      if(dest != my_rank) { //Avoid sending myself
        //sprintf(debugstr,"Befr:Se:seq:%d, d:%d", op->data.seq_num, dest );
        //write_to_file(my_rank, debugstr);
        ret = MPI_Send(sendbuf, op->data.count, datatype, dest,
                   op->data.tag, comm); //, &request);
        //sprintf(debugstr,"Succ:Se:seq:%d, d:%d", op->data.seq_num, dest );
        //write_to_file(my_rank, debugstr);
        if(ret != MPI_SUCCESS) {
          fprintf(stderr, "error: MPI_Send did not return MPI_SUCCESS\n");
          exit(0);
        }
      }
      //set_compute_time ();
    }
    if(num_Recvs) {
      total_recv += num_Recvs;
      statuses = (MPI_Status *) malloc(sizeof(MPI_Status) * num_Recvs);
      MPI_Waitall(send_list.curr_ptr, requests, statuses);
      //sprintf(debugstr,"Succ:IR:seq:%d, s:%d", op->data.seq_num, send_list.rank_arr[iter] );
      //write_to_file(my_rank, debugstr);
    }

    set_compute_time ();

    /*if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Send did not return MPI_SUCCESS\n");
      exit(0);
    }*/

    FREE(sendbuf);   
   #endif
    break;


  case umpi_MPI_Sendrecv:
    unimplemented_error("MPI_Sendrecv");
    break;

  case umpi_MPI_Sendrecv_replace:
    unimplemented_error("MPI_Sendrecv_replace");
    break;

  case umpi_MPI_Ssend:
    unimplemented_error("MPI_Ssend");
    break;

  case umpi_MPI_Ssend_init:
    unimplemented_error("MPI_Ssend_init");
    break;

  case umpi_MPI_Start:
    {
      int i = (op->data.request_vector ? vector_get_next (op->data.request_vector) : op->data.request);
      lookup_request( replay_req, i, RECORDED_MPI_REQUEST_NULL, &request);
      simulate_compute_time (op, iter);
      ret = MPI_Start (&request);
      set_compute_time ();
      if(ret != MPI_SUCCESS) {
        fprintf(stderr, "error: MPI_Start did not return MPI_SUCCESS\n");
        exit(0);
      }
    }
    break;

  case umpi_MPI_Startall: 
    requests = (MPI_Request *) malloc(sizeof(MPI_Request) * op->data.array_size1);
    lookup_requestlist (replay_req, op->data.array1, op->data.array_size1, 
                        RECORDED_MPI_REQUEST_NULL, requests);
    simulate_compute_time (op, iter);
    ret = MPI_Startall (op->data.array_size1, requests);
    free (requests);
    set_compute_time (); 

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_commit did not return MPI_SUCCESS\n");
      exit(0);
    }
    break;

  case umpi_MPI_Status_c2f:
    unimplemented_error("MPI_Status_c2f");
    break;

  case umpi_MPI_Status_f2c:
    unimplemented_error("MPI_Status_f2c");
    break;

  case umpi_MPI_Test:
    unimplemented_error("MPI_Test");
    break;

  case umpi_MPI_Test_cancelled:
    unimplemented_error("MPI_Test_cancelled");
    break;

  case umpi_MPI_Testall:
    unimplemented_error("MPI_Testall");
    break;

  case umpi_MPI_Testany:
    unimplemented_error("MPI_Testany");
    break;

  case umpi_MPI_Testsome:
    unimplemented_error("MPI_Testsome");
    break;

  case umpi_MPI_Topo_test:
    unimplemented_error("MPI_Topo_test");
    break;

  case umpi_MPI_Type_commit:
  #ifndef REPLAY_HISTO
  #ifdef FEATURE_FILE_IO
    newtype = index_to_type (replay_ptr, op->data.newtype);
    printf("type_commit: newtype index %d, newtype %d, %x, %lld\n",op->data.newtype, newtype, newtype, newtype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_commit(&newtype);
    printf("type_commit: after commit newtype index %d, newtype %d, %x, %lld\n",op->data.newtype, newtype, newtype, newtype);

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_commit did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
  #endif
  #endif
    break;

  case umpi_MPI_Type_contiguous:
    datatype = index_to_type (replay_ptr, op->data.oldtype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_contiguous (op->data.count, datatype, &newtype);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_contiguous did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_type_entry(replay_ptr, newtype);
    break;

  case umpi_MPI_Type_count:
    unimplemented_error("MPI_Type_count");
    break;

  case umpi_MPI_Type_create_darray:
    #ifndef FEATURE_FILE_IO
    unimplemented_error("MPI_Type_create_darray");
    #else
    datatype = index_to_type(replay_ptr, op->data.datatype);
    printf("Rank %d: darray: old datatype index is %d value is %d, newtype %d, rank %d\n",my_rank, op->data.datatype, datatype, newtype, op->data.rank);
    simulate_compute_time (op, iter);
    ret = MPI_Type_create_darray(op->data.comm_size, my_rank, op->data.count, op->data.array1, 
              op->data.array2, op->data.array3, op->data.array4, op->data.order, datatype, &newtype );
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_create_darray did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    printf("darray: new type created is %d\n",newtype);
    add_type_entry(replay_ptr, newtype);
    printf("darray: newtype index is %d\n",type_to_index(replay_ptr, newtype));
    #endif
    break;

  case umpi_MPI_File_open:
    comm = index_to_comm(replay_ptr, op->data.comm);
    simulate_compute_time (op, iter);
    ret = MPI_File_open(comm, op->data.filename, op->data.amode, info, &file);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_File_open did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_file_handle_entry(replay_ptr, file);
    break;

  case umpi_MPI_File_close:
    file = index_to_file_handle(replay_ptr, op->data.fh);
    simulate_compute_time(op, iter);
    ret = MPI_File_close(&file);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_File_close did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time();
    remove_file_handle_entry(replay_ptr, file);
    break;

  case umpi_MPI_File_set_view:
    file = index_to_file_handle(replay_ptr, op->data.fh);
    datatype = index_to_type(replay_ptr, op->data.etype);
    newtype = index_to_type(replay_ptr, op->data.filetype);
    //disp = op->data.disp_vector ? vector_get_next (op->data.disp_vector) : op->data.disp;
    disp = 0;
    simulate_compute_time (op, iter);
    printf("Rank %d: set_view: disp %d,  datatype %d, newtype %d, newtype index %d, datarep %s\n",rank, disp, datatype, newtype, op->data.filetype, op->data.datarep);
    ret = MPI_File_set_view(file, disp, datatype, newtype, op->data.datarep, info);
    printf("Rank %d: set_view: after setview call\n",rank);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_File_set_view did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time();
    add_type_entry(replay_ptr, newtype);
    break;

  case umpi_MPI_File_write_at:
    file = index_to_file_handle(replay_ptr, op->data.fh);
    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate write buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    Op_sendbuf_alloc(Send, void, op->data.count * size, sendbuf);
    if(!sendbuf) {
      perror("malloc MPI_Send buffer");
      exit(0);
    }

    simulate_compute_time (op, iter);
    ret = MPI_File_write_at(file, op->data.offset, sendbuf, op->data.count, datatype, &status);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_File_write_at did not return MPI_SUCCESS\n");
      exit(0);
    }
    op->data.offset = (op->data.offset < ((op->node+1)*op->data.offsetinfo.totalcount))? (op->data.offset+op->data.offsetinfo.stride):0;
    set_compute_time();
    FREE(sendbuf);
    break;

  case umpi_MPI_File_write_all:
    file = index_to_file_handle(replay_ptr, op->data.fh);
    datatype = index_to_type (replay_ptr, op->data.datatype);
    /* allocate write buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    printf("Rank %d: write_all: extent size is %d\n", rank, size);
    Op_sendbuf_alloc(Send, void, op->data.count * size, sendbuf);
    if(!sendbuf) {
      perror("malloc MPI_Send buffer");
      exit(0);
    }
    memset(sendbuf, rank+65, op->data.count * size);
    simulate_compute_time (op, iter);
    ret = MPI_File_write_all(file, sendbuf, op->data.count, datatype, &status);
    MPI_Get_count(&status, datatype, &oc);
    printf("Rank %d: write_all: get count returned %d\n", rank, oc);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_File_write_at did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time();
    FREE(sendbuf);
    break;

  case umpi_MPI_File_read_all:
    file = index_to_file_handle(replay_ptr, op->data.fh);
    datatype = index_to_type(replay_ptr, op->data.datatype);
    /* allocate write buf */
    if(MPI_Type_extent(datatype, &size) != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_extent failed - probably invalid type\n");
      exit(0);
    }
    printf("Rank %d: read_all: extent size is %d\n", rank, size);
    Op_recvbuf_alloc(Recv, void, op->data.count * size, recvbuf);
    if(!recvbuf) {
      perror("malloc MPI_Send buffer");
      exit(0);
    }
    simulate_compute_time (op, iter);
    ret = MPI_File_read_all(file, recvbuf, op->data.count, datatype, &status);
    MPI_Get_count(&status, datatype, &oc);
    printf("Rank %d: read_all: get count returned %d\n", rank, oc);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_File_read_all did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time();
    FREE(recvbuf);
    break;

  case umpi_MPI_File_seek:
    file = index_to_file_handle(replay_ptr, op->data.fh);
    simulate_compute_time (op, iter);
    ret = MPI_File_seek(file, op->data.offset, op->data.whence);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_File_seek did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time();
    break;

  case umpi_MPI_Type_create_indexed_block:
    unimplemented_error("MPI_Type_create_indexed_block");
    break;

  case umpi_MPI_Type_create_subarray:
    unimplemented_error("MPI_Type_create_subarray");
    break;

  case umpi_MPI_Type_extent:
    unimplemented_error("MPI_Type_extent");
    break;

  case umpi_MPI_Type_free:
#if 0
    newtype = index_to_type (replay_ptr, op->data.newtype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_free(&newtype);

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_free did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    remove_type_entry(replay_ptr, newtype);
#endif
    break;

  case umpi_MPI_Type_get_contents:
    unimplemented_error("MPI_Type_get_contents");
    break;

  case umpi_MPI_Type_get_envelope:
    unimplemented_error("MPI_Type_get_envelope");
    break;

  case umpi_MPI_Type_hindexed: 
    datatype = index_to_type (replay_ptr, op->data.datatype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_hindexed (op->data.count,
                             op->data.array1,
                             op->data.array_of_displacements,
                             datatype, &newtype);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_hindexed did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_type_entry(replay_ptr, newtype);

    break;

  case umpi_MPI_Type_hvector:
    datatype = index_to_type (replay_ptr, op->data.datatype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_hvector(op->data.count,
                           op->data.blocklen, op->data.stride,
                           datatype, &newtype);

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_hvector did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_type_entry(replay_ptr, newtype);

    break;

  case umpi_MPI_Type_indexed: 
#if 0
    datatype = index_to_type (replay_ptr, op->data.datatype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_indexed (op->data.count,
                            op->data.array1, op->data.array2, datatype,
                            &newtype);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_indexed did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_type_entry(replay_ptr, newtype);
#endif
    break;

  case umpi_MPI_Type_lb:
    unimplemented_error("MPI_Type_lb");
    break;

  case umpi_MPI_Type_size:
    unimplemented_error("MPI_Type_size");
    break;

  case umpi_MPI_Type_struct: 
    datatypes = (MPI_Datatype*) malloc
      (op->data.count);
    OP_ARRAY_GET_TYPES(datatypes, op->data.array3, op->data.count, replay_ptr);
    simulate_compute_time (op, iter);
    ret = MPI_Type_struct (op->data.count,
                           op->data.array1,
                           op->data.array_of_displacements,
                           datatypes, &newtype); 
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_struct did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    free (datatypes);
    add_type_entry(replay_ptr, newtype);

    break;

  case umpi_MPI_Type_ub:
    unimplemented_error("MPI_Type_ub");
    break;

  case umpi_MPI_Type_vector: 
   #ifndef REPLAY_HISTO
    datatype = index_to_type (replay_ptr, op->data.datatype);
    simulate_compute_time (op, iter);
    ret = MPI_Type_vector (op->data.count,
                           op->data.blocklen, op->data.stride, datatype,
                           &newtype);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_vector did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();
    add_type_entry(replay_ptr, newtype);
   #endif
    break;

  case umpi_MPI_Unpack:
    unimplemented_error("MPI_Unpack");
    break;

  case umpi_MPI_Wait:
    {
   #ifndef REPLAY_HISTO
      int i = op->data.request_vector ? vector_get_next (op->data.request_vector) : op->data.request;
      lookup_request( replay_req, i, RECORDED_MPI_REQUEST_NULL, &request);
      simulate_compute_time (op, iter);
      ret = MPI_Wait(&request, &status);
      if(ret != MPI_SUCCESS) {
        fprintf(stderr, "error: MPI_Wait did not return MPI_SUCCESS\n");
        exit(0);
      }
      set_compute_time ();

      if (request == MPI_REQUEST_NULL)
        free_mem_entry(i);
   #else
     simulate_compute_time(op, iter);
     set_compute_time();
   #endif
    }
    break;

  case umpi_MPI_Waitall:
   #ifndef REPLAY_HISTO
    requests = (MPI_Request *) malloc(sizeof(MPI_Request) * op->data.array_size1);
    statuses = (MPI_Status *) malloc(sizeof(MPI_Status) * op->data.array_size1);
    lookup_requestlist( replay_req, op->data.array1, op->data.array_size1, RECORDED_MPI_REQUEST_NULL, requests);
    simulate_compute_time (op, iter);
    ret = MPI_Waitall(op->data.array_size1, requests, statuses);

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Waitall did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();

    for (int i = 0; i < op->data.array_size1; i++) {
      if (requests[i] == MPI_REQUEST_NULL) {
        free_mem_entry (i);
      }
    }
    free(requests);
    free(statuses);
   #else
     simulate_compute_time(op, iter);
     set_compute_time();
   #endif
    break;

  case umpi_MPI_Waitany: 
    requests = (MPI_Request *) malloc(sizeof(MPI_Request) * op->data.array_size1);
    lookup_requestlist( replay_req, op->data.array1, op->data.array_size1, RECORDED_MPI_REQUEST_NULL, requests);
    simulate_compute_time (op, iter);

    int i;
    ret = MPI_Waitany (op->data.array_size1, requests, &i, &status);
    set_compute_time ();

    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Waitall did not return MPI_SUCCESS\n");
      exit(0);
    }

    if (requests[i] == MPI_REQUEST_NULL)
      free_mem_entry(i);
    free(requests);
    break;


  case umpi_MPI_Waitsome:
    oc = 0;
    requests = (MPI_Request *) malloc(sizeof(MPI_Request) * op->data.array_size1);
    statuses = (MPI_Status *) malloc(sizeof(MPI_Status) * op->data.array_size1);
    indices = (int *) malloc(sizeof(int) * op->data.array_size1);

    lookup_requestlist_waitsome( replay_req, op->data.array1, op->data.array_size1, RECORDED_MPI_REQUEST_NULL, requests);
    simulate_compute_time (op, iter);
    while(oc < op->data.outcount) {
      ret = MPI_Waitsome(op->data.array_size1, requests, &cur_oc, indices, statuses);
      if(ret != MPI_SUCCESS) {
        fprintf(stderr, "error: MPI_Waitsome did not return MPI_SUCCESS\n");
        exit(0);
      }

      oc += cur_oc;

      /* free mem of completed ops */
      for(i = 0; i < cur_oc; i++)
        if (requests[op->data.array1[indices[i]]] == MPI_REQUEST_NULL)
          free_mem_entry(op->data.array1[indices[i]]);
    }
    //FIXME: this is slightly inaccurate as we don't account for the time spent
    //in free_mem_entry etc
    set_compute_time ();

    free(requests);
    free(statuses);
    free(indices);
    break;

  case umpi_MPI_Wtick:  // informational only
    break;

  case umpi_MPI_Wtime:  // informational only
    break;

  case umpi_MPI_Dims_create:

    /*simulate_compute_time (op, iter);
    ret = MPI_Dims_create (op->data.comm_size,
                           op->data.count, op->data.stride, datatype,
                           &newtype);
    if(ret != MPI_SUCCESS) {
      fprintf(stderr, "error: MPI_Type_vector did not return MPI_SUCCESS\n");
      exit(0);
    }
    set_compute_time ();*/
    break;

  default:
    fprintf(stderr, "error: undefined operation number specified (%d)\n", op->data.op);
    //exit(0);
  }

  //if(get_my_rank() == 0)
    //printf("Rank: %d: after replay op %d %d\n", get_my_rank(), op->data.op, op->data.seq_num);
  return 0;
}

#ifdef FEATURE_TIMING_HISTO

unsigned long long cTime;
INLINE int set_compute_time ()
{
  cTime = (unsigned long long int)(get_cur_time_ns ()/1000);
}

INLINE long int get_compute_time ()
{
  unsigned long long t;

  t = (unsigned long long int)(get_cur_time_ns ()/1000);
  t = t - cTime;
  if(t < 0) {
    fprintf(stderr, "-ve time diff\n");
    t = 0;
  }
  return t;
  /*else {
    fprintf(stderr, "-ve time diff\n");
    return 0;
  }*/
  //return t - cTime;
}
#endif //FEATURE_TIMING_HISTO

/* EOF */
