#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include "prsd_traverse.h"
#include "rsd_reader.h"
#include "op_exec.h"
#include "trace.h"
#include "utils.h"
#include "mpi_util.h"
#include "RandomVector.h"

//#define DEBUG 0
#ifdef LINK_WITH_RECORD
#define MPI_Pcontrol(a)
#endif

 static long long int rank_time;

 static long long int total_count;

 static long long int time_in_other_calls;

 static long tot_my_recv;

 RandomVector *timing_rvector;
 RandomVector *param_rvector;

int check_collective(rsd_node * node) {
  switch(node->data.op) {
    case umpi_MPI_Allgather:
    case umpi_MPI_Allgatherv:
    case umpi_MPI_Allreduce:
    case umpi_MPI_Alltoall:
    case umpi_MPI_Alltoallv:
    case umpi_MPI_Barrier:
    case umpi_MPI_Bcast:
    case umpi_MPI_Gather:
    case umpi_MPI_Gatherv:
    case umpi_MPI_Reduce:
    case umpi_MPI_Reduce_scatter:
    case umpi_MPI_Scatter:
    case umpi_MPI_Scatterv:
    case umpi_MPI_File_open:
    case umpi_MPI_File_write_all:
    case umpi_MPI_File_read_all:
    case umpi_MPI_File_close:
      return TRUE;
    default:
      return FALSE;
    
  } 
}
/**
 * Visitor routine for executing replay operations.  This is called
 * by the traversal routines to translate ops to actual MPI function 
 * calls.
 */
static int replay_visitor(rsd_node *node, int rank, int depth, int iter, int
    last) {
 #ifdef REPLAY_HISTO
  long long int start, end;
  int is_member=0;
  struct timeval tv;
  int ret;
  gettimeofday(&tv, NULL);
  start = tv.tv_sec * 1000000 + tv.tv_usec;

  is_member = member_rank(rank, node->rlist);

  gettimeofday(&tv, NULL);
  end = tv.tv_sec * 1000000 + tv.tv_usec;
  rank_time += (end-start);
  //if(node->data.op == umpi_MPI_Finalize)
    //printf("Total time in rank traversing %lld\n", rank_time);

  if(is_member || node->data.op == umpi_MPI_Send
          || node->data.op == umpi_MPI_Isend)
    total_count ++;
  if(node->data.op == umpi_MPI_Finalize) {
    printf("Rank: %d: Total op count %lld\n",get_my_rank(), total_count);
    printf("Rank: %d: Time spent in other calls %lld\n", get_my_rank(), time_in_other_calls);
    printf("Rank: %d: Total my recvs %ld\n", get_my_rank(), tot_my_recv);
  }  

  if(is_member || node->data.op == umpi_MPI_Send
          || node->data.op == umpi_MPI_Isend) {//|| check_collective(node))//node->data.op == umpi_MPI_Allreduce)
    //Total count of my recvs
    if(node->data.op == umpi_MPI_Irecv || node->data.op == umpi_MPI_Recv)
      tot_my_recv++;

    //Calculate time spent in other calls
    if(!is_member) {
      gettimeofday(&tv, NULL);
      start = tv.tv_sec * 1000000 + tv.tv_usec;
    }    

    ret = exec_prsd_call(node, rank, depth, iter);

    if(!is_member) {
      gettimeofday(&tv, NULL);
      end = tv.tv_sec * 1000000 + tv.tv_usec;
      time_in_other_calls += (end - start);
    }    

    return ret;
  }
  else
    return 0;
 #else
  total_count ++;
  if(node->data.op == umpi_MPI_Finalize)
    printf("Rank: %d: Total count %lld\n",get_my_rank(), total_count);
  return exec_prsd_call(node, rank, depth, iter);
 #endif
}


int main(int argc, char *argv[])
{
  int rank = -1, len = -1, whole, frac;
  int retval = -1;
  int per_node_trace = FALSE;
  struct timeval t_start, t_stop;
  struct timezone tz;
  rsd_queue *prsd_queue;
  char * dir = 0;
  char format = 'x';

  init_rand_vector(&param_rvector, time(NULL), 1000000);

  MPI_Init(&argc, &argv);
  //disable profiling at init
  MPI_Pcontrol (0);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &my_size);
  my_rank = rank;
  init_rand_vector(&timing_rvector, rank+2, 1000000 );

  /* determine input type */
  if(argc < 2) {
    fprintf(stderr, "usage: %s -b|-a|-p <trace dir>\n", argv[0]);
    MPI_Finalize ();
    return 0;
  }
  while (1)
  {
    retval = getopt (argc, argv, "p:X");

    if (retval == -1)
      break;
    switch (retval)
    {
      case 'p':
        format = 'p';
        dir = optarg;
        break;
      case 'X':
        per_node_trace = TRUE;
        break;
    }
  }
  if (format == -1)
  {
    fprintf(stderr, "error: must specify the input file's format w/ -p\n");
    MPI_Finalize ();
    return 0;
  }

  /* read in op trace for each node */
  char filename[4096];
  sprintf (filename, "%s/%d", dir, (per_node_trace ? rank : 0));
 #ifdef REPLAY_HISTO
  prsd_queue = read_rsd_file(filename);
 #else
  prsd_queue = read_rsd_file_rank(filename, rank);
 #endif
  len = rsd_get_length(prsd_queue);

  /* playback ops in queue */
  gettimeofday(&t_start, &tz);

//#ifdef DEBUG
  /*char outfile[100];
  sprintf(outfile, "trace-%d", rank);
  FILE * ff = fopen(outfile, "w"); 
  //if (rank == 0)  
  output_rsd_queue (ff, prsd_queue);
  */
//  else output_rsd_queue(stderr, prsd_queue);
//  fprintf (stdout, "output rsd queue\n");
//  fflush (stdout);
//#else
  replay_queue(prsd_queue->head, rank, replay_visitor);
//#endif
  // finish profiling
  MPI_Pcontrol (0);

  //printf("Total time in rank traversing %lld\n", rank_time);
  /* stop timer */
  gettimeofday(&t_stop, &tz);
  whole = t_stop.tv_sec - t_start.tv_sec;
  frac = t_stop.tv_usec - t_start.tv_usec;
  if(frac < 0) {
    frac *= -1;
    whole --;
  }
  if(rank == 0)
    printf("execution time:  %d.%d\n", whole, frac);

  MPI_Finalize();
  return 0;
}
/* EOF */
