#include "stackwalk/h/walker.h"
#include "stackwalk/h/procstate.h"
#include "stackwalk/h/frame.h"
#include "stackwalk/h/swk_errors.h"

#include <unistd.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <vector>


using namespace Dyninst;
using namespace Dyninst::Stackwalker;
using namespace std;


void walk_stack(Walker *walker);
void parse_args(std::vector<Walker *> &walkers, int argc, char *argv[]);
void usage();

/**
 * An example and test showing how to build a stackwalker object, how to 
 * handle the debugging aspects of the stackwalker, and finally how to collect
 * a stackwalk.
 *
 * This tool takes a collection of processess on the command line that may include
 * attached processes, created processes, or the current process.  Once every 
 * five seconds, or when input is available on stdin, a stackwalk is taken.
 **/
int main(int argc, char *argv[])
{
  std::vector<Walker *> walkers;
  parse_args(walkers, argc, argv);
  if (!walkers.size())
     usage();


  /**
   * The notification FD is used by third party stackwalkers (debuggers) to tell
   * the user tool when control needs to be given to StackwalkerAPI to handle a 
   * debug event.  If data is available on the notificationFD then the user
   * should call handleDebugEvent.
   **/
  int notification_fd = -1;
  for (unsigned i=0; i<walkers.size(); i++)
  {
     ProcDebug *pdebug = dynamic_cast<ProcDebug *>(walkers[i]->getProcessState());
     if (pdebug)
     {
        //We using at least one debugger walker.  Use the notification fd.
        notification_fd = ProcDebug::getNotificationFD();
        break;
     }
  }

  struct timeval timeout;
  timeout.tv_sec = 5;
  timeout.tv_usec = 0;
  
  while (walkers.size())
  {
    //Set up the select call.  Wait for input on stdin or the notificationFD.
    int max = 1;
    fd_set readfds, writefds, exceptfds;
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_ZERO(&exceptfds);
    FD_SET(0, &readfds);
    if (notification_fd != -1) {
       FD_SET(notification_fd, &readfds);
       max = notification_fd + 1;
    }
       
    /**
     * Wait for one of:
     *  5 seconds                     - Then do a stackwalk
     *  input on stdin                - Then do a stackwalk
     *  input on the notificationFD   - Call handleDebugEvent to let the
     *                                  debugger process any events.
     **/
    int result = select(max, &readfds, &writefds, &exceptfds, &timeout);
    if (result == -1 && errno != EINTR)
    {
      perror("select failure");
      return 0;
    }

    if (notification_fd != -1 && FD_ISSET(notification_fd, &readfds))
    {
       //Handle the event.  Should check for process exit here.
       bool result;
       result = ProcDebug::handleDebugEvent(false); //false = non-blocking
       if (!result)
       {
          fprintf(stderr, "Error handling debug events: %s\n",
                  getLastErrorMsg());
       }
       //TODO: Remove terminated walker processes
       continue;
    }
    if (FD_ISSET(0, &readfds))
    {
      //Input was available on stdin.  If it's 'q', then quit.
      char c;
      int result_i;
      result_i = read(0, &c, 1);
      if (result_i == -1) {
         perror("Couldn't read from 0");
         return -1;
      }
      if (c == 'q')
         break;
    }
    if (result >= 0)
    {
      //Either the five seconds timed out (result == 0) or we have input
      // on stdin (result == 1).  Perform a stackwalk.
      for (unsigned i=0; i<walkers.size(); i++) 
      {
         printf("\n%d:\n", walkers[i]->getProcessState()->getProcessId());
         walk_stack(walkers[i]);
      }
      timeout.tv_sec = 5;
    }
  }  
  return 0;
}

void usage()
{
  fprintf(stderr, "usage: test_basic [ -self | -attach <pid> | -create " \
          "<exec> <args> ]\n\n");
  fprintf(stderr, "You may specify multiple processes at once e.g:\n");
  fprintf(stderr, "\t test_basic -self -create foo arg1 -attach 5000 -attach 5002\n");
  exit(-1);
}

void parse_args(std::vector<Walker *> &walkers, int argc, char *argv[])
{
  for (int i=1; i<argc; i++)
  {
    Walker *walker;
    if (strcmp(argv[i], "-attach") == 0 && i+1 < argc) {
       //Third party stack walker, attached to a pid
      walker = Walker::newWalker(atoi(argv[++i]));
    }
    else if (strcmp(argv[i], "-self") == 0) {
       //First party stack walker
      walker = Walker::newWalker();
    }
    else if (strcmp(argv[i], "-create") == 0 && i+1 < argc) {
       //Thread party stackwalker, creates an executable
       string exec_name(argv[++i]);
       vector<string> args;
       while ((i < argc) && (*(argv[i]) != '-')) 
       {
          args.push_back(argv[i]);
          i++;
       }
       walker = Walker::newWalker(exec_name, args);
    }
    else {
       usage();
    }
    if (!walker)
    {
      fprintf(stderr, "Couldn't creating walker: %s\n", 
              getLastErrorMsg());
      continue;
    }
    walkers.push_back(walker);
  }
}

/**
 * Walk the stack for the given walker object.  Each walker object is
 * associated with one process.
 **/
void walk_stack(Walker *walker)
{
  std::vector<Frame> swalk;

  walker->walkStack(swalk);
  for (unsigned i=0; i<swalk.size(); i++)
  {
     std::string name;
     swalk[i].getName(name);
     printf("\t%s @ 0x%x\n", name.c_str(), (unsigned) swalk[i].getRA());
  }
}
